unit ovAccuracerDBAdapter;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, ovTypes, ovModel, ovComponents, ACRMain;

type
  TovAccuracerDBAdapter = class(TovCustomDatabaseAdapter)
  private
    FDatabase: TACRDatabase;
    FBlobCompressionAlgorithm: TCompressionAlgorithm;
    FBlobCompressionLevel: Integer;
    procedure SetDatabase(const Value: TACRDatabase);
    procedure CheckDatabase;
    procedure FillFieldDefs(ATable: TACRTable; ANewFieldsDefs: TFieldDefs);
    procedure FillIndexDefs(ATable: TACRTable; ANewIndexDefs: TIndexDefs);
    procedure RestructureTable(ATable: TACRTable; AFieldsDefs: TFieldDefs;
      AnIndexDefs: TIndexDefs);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CloneTable(SourceTable: TDataSet; SourceTableFieldNames: TStrings;
      TableName: String; FieldsDefs: TFieldDefs; IndexDefs: TIndexDefs); override;
    function GetLastKeyValue(ATableName, IdFieldName: String): Integer; override;
  public
    procedure SetupConnection(ConnectionParams: TStrings); override;
    constructor Create(AOwner: TComponent); override;
    procedure CloseDatabase; override;
    procedure CommitTransaction; override;
    procedure CreateTable(ATableName: String; AFieldsDefs: TFieldDefs; AnIndexDefs:
      TIndexDefs); override;
    function EditConnection(ConnectionParams: TStrings): Boolean; override;
    procedure ExecQuery(SQLText: String); override;
    procedure GenerateDatabaseName;
    function InTransaction: Boolean; override;
    procedure OpenDatabase; override;
    function OpenQuery(QueryText: String; MasterDataSource: TDataSource;
      LiveDataSet: Boolean): TDataSet; override;
    function OpenTable(ATableName: String; AReadOnly: Boolean = False; AFilter:
      String = ''): TDataSet; override;
    procedure RollbackTransaction; override;
    procedure StartTransaction; override;
    function TableExists(ATableName: String): Boolean; override;
    procedure DeleteTable(ATableName: String);
  published
    property Database: TACRDatabase read FDatabase write SetDatabase;
    property BlobCompressionAlgorithm: TCompressionAlgorithm read
      FBlobCompressionAlgorithm write FBlobCompressionAlgorithm default caNone;
    property BlobCompressionLevel: Integer read FBlobCompressionLevel write 
      FBlobCompressionLevel;
  end;

  ERestructureTableError = class (Exception)
  end;

procedure Register;

implementation

uses ovConsts, ovObjects, ovAccuracerDBParams, ovUtils;

{$R *.dcr}

procedure Register;
begin
  RegisterComponents(cpObjectViewAddOn, [TovAccuracerDBAdapter]);
end;

{
**************************** TovAccuracerDBAdapter ******************************
}
{<<Category>>: Undefined}
constructor TovAccuracerDBAdapter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  with QuerySyntaxParams do begin
    LeftTableDelimiter := '[';
    RightTableDelimiter := ']';
    LeftFieldDelimiter := '[';
    RightFieldDelimiter := ']';
  end;
end;

procedure TovAccuracerDBAdapter.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = Database) and (Operation = opRemove)
    then Database := nil;
end;

{***************************************************************************** }
{                                                                    }
{ **************************************************************************** }

procedure TovAccuracerDBAdapter.SetDatabase(const Value: TACRDatabase);
begin
  if FDatabase <> Value then
  begin
    FDatabase := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovAccuracerDBAdapter.OpenDatabase;
begin
  CheckDatabase;
  Database.Open;
end;

procedure TovAccuracerDBAdapter.CloseDatabase;
begin
  if Assigned(Database) then Database.Close;
end;

procedure TovAccuracerDBAdapter.CheckDatabase;
begin
  if Not Assigned(Database) then
    raise Exception.Create('   Database : ' + Name);
  if Database.DatabaseName = '' then
    raise Exception.Create('      : ' + Name);
end;

{***************************************************************************** }
{                                                                    }
{ **************************************************************************** }

procedure TovAccuracerDBAdapter.StartTransaction;
begin
  CheckDatabase;
  Database.StartTransaction
end;

procedure TovAccuracerDBAdapter.CommitTransaction;
begin
  CheckDatabase;
  Database.Commit;
end;

procedure TovAccuracerDBAdapter.RollbackTransaction;
begin
  CheckDatabase;
  Database.Rollback
end;

function TovAccuracerDBAdapter.InTransaction: Boolean;
begin
  CheckDatabase;
  Result := Database.InTransaction
end;

{***************************************************************************** }
{                                                               }
{ **************************************************************************** }

function TovAccuracerDBAdapter.OpenTable(ATableName: String; AReadOnly: Boolean
  = False; AFilter: String = ''): TDataSet;
begin
  CheckDatabase;

  Result := TACRTable.Create(nil);
  try
    with TACRTable(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      TableName := ATableName;
      ReadOnly := AReadOnly;
      Filter := AFilter;
      if AFilter <> '' then Filtered := True;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

function TovAccuracerDBAdapter.TableExists(ATableName: String): Boolean;
var
  Table: TACRTable;
begin
  Table := TACRTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    TableName := ATableName;
    Result := Exists;
  finally
    Free
  end;
end;

procedure TovAccuracerDBAdapter.CreateTable(ATableName: String; AFieldsDefs:
  TFieldDefs; AnIndexDefs: TIndexDefs);
var
  Table: TACRTable;
begin
  CheckDatabase;

  Table := TACRTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    TableName := ATableName;
    FieldDefs.Assign(AFieldsDefs);
    if AnIndexDefs <> nil then IndexDefs.Assign(AnIndexDefs);
    if Not Exists then
      CreateTable
    else
      Self.RestructureTable(Table, AFieldsDefs, AnIndexDefs)
  finally
    Free
  end;
end;

procedure TovAccuracerDBAdapter.RestructureTable(ATable: TACRTable; AFieldsDefs:
  TFieldDefs; AnIndexDefs: TIndexDefs);
var
  Log: String;
begin
  ATable.Open;
  ATable.Close;
  { fill in restructure properties }
  FillFieldDefs(ATable, AFieldsDefs);
  FillIndexDefs(ATable, AnIndexDefs);
  if Not ATable.RestructureTable(Log) then
    raise ERestructureTableError.Create(Log);
end;

procedure TovAccuracerDBAdapter.FillFieldDefs(ATable: TACRTable; ANewFieldsDefs:
  TFieldDefs);
var
  I: Integer;
//  FieldDef: TABSAdvFieldDef;
begin
//   with ATable do
//   begin
//     //   
//     RestructureFieldDefs.Assign(AdvFieldDefs);
//
//     //       
//     for I := 0 to ANewFieldsDefs.Count-1 do
//     begin
//       FieldDef := RestructureFieldDefs.Find(ANewFieldsDefs[I].Name);
//       if Not Assigned(FieldDef) then
//         FieldDef := RestructureFieldDefs.AddFieldDef;
//       with FieldDef do //      
//       begin
//         Name := ANewFieldsDefs[I].Name;
//         DataType := FieldTypeToABSAdvFieldType(ANewFieldsDefs[I].DataType);
//         Size := ANewFieldsDefs[I].Size;
//         Required := ANewFieldsDefs[I].Required;
//       end
//     end;
//   end
end;

procedure TovAccuracerDBAdapter.FillIndexDefs(ATable: TACRTable; ANewIndexDefs:
  TIndexDefs);
var
  I: Integer;
//  IndexDef: TABSAdvIndexDef;
begin
//   with ATable do
//   begin
//     //   
//     RestructureIndexDefs.Assign(AdvIndexDefs, True);
//
//     //       
//     for I := 0 to ANewIndexDefs.Count-1 do
//     begin
//       IndexDef := RestructureIndexDefs.Find(ANewIndexDefs[I].Name);
//       if Not Assigned(IndexDef) then
//         IndexDef := RestructureIndexDefs.AddIndexDef;
//       with IndexDef do //      
//       begin
//         Name := ANewIndexDefs[I].Name;
//         Options := ANewIndexDefs[I].Options;
//         Fields := ANewIndexDefs[I].Fields;
//       end
//     end;
//   end
end;

procedure TovAccuracerDBAdapter.DeleteTable(ATableName: String);
begin
  CheckDatabase;
  with TACRTable.Create(nil) do
  try
    DatabaseName := Self.Database.DatabaseName;
    TableName := ATableName;
    if Exists then DeleteTable;
  finally
    Free
  end;
end;

{{
    TableName,        
  SourceTable,      .
}
procedure TovAccuracerDBAdapter.CloneTable(SourceTable: TDataSet;
  SourceTableFieldNames: TStrings; TableName: String; FieldsDefs: TFieldDefs;
  IndexDefs: TIndexDefs);
var
  DestTable: TDataSet;
begin
  DeleteTable(TableName);

  //   ,       
  FieldsDefs[0].DataType := ftInteger;
  CreateTable(TableName, FieldsDefs, IndexDefs);

  //        
  DestTable := OpenTable(TableName);
  try
    CopyData(SourceTable, SourceTableFieldNames, DestTable);

    FieldsDefs[0].DataType := ftAutoInc;
    RestructureTable(DestTable as TACRTable, FieldsDefs, IndexDefs);
  finally
    DestTable.Free;
  end;
end;

{***************************************************************************** }
{                                                            }
{ **************************************************************************** }

procedure TovAccuracerDBAdapter.ExecQuery(SQLText: String);
begin
  CheckDatabase;

  with TACRQuery.Create(nil) do
  try
    DatabaseName := Self.Database.DatabaseName;
    SQL.Text := SQLText;
    ExecSQL;
  finally
    Free
  end;
end;

function TovAccuracerDBAdapter.OpenQuery(QueryText: String; MasterDataSource: 
  TDataSource; LiveDataSet: Boolean): TDataSet;
begin
  CheckDatabase;

  Result := TACRQuery.Create(nil);
  try
    with TACRQuery(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      SQL.Text := QueryText;
      if MasterDataSource <> nil then
        DataSource := MasterDataSource;
      RequestLive := LiveDataSet;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

function TovAccuracerDBAdapter.GetLastKeyValue(ATableName, IdFieldName:
  String): Integer;
begin
  CheckDatabase;
  with TACRTable.Create(nil) do
  try
    DatabaseName := Self.Database.DatabaseName;
    TableName := ATableName;
    Open;
    Result := LastAutoincValue(ObjectIdField);
  finally
    Free
  end;
end;

{***************************************************************************** }
{                                                                              }
{ **************************************************************************** }

procedure TovAccuracerDBAdapter.GenerateDatabaseName;
var
  DatabaseCount: Integer;
begin
  DatabaseCount := 1;
  while True do
  try
    Inc(DatabaseCount);
    Database.DatabaseName := 'Database_' + IntToStr(DatabaseCount);
    Break;
  except
    //     DatabaseName   
  end;
end;

procedure TovAccuracerDBAdapter.SetupConnection(ConnectionParams: TStrings);
begin
  Database := TACRDatabase.Create(Self);
  if ConnectionParams.Values['DatabaseName'] = ''
    then GenerateDatabaseName
    else Database.DatabaseName := ConnectionParams.Values['DatabaseName'];

  Database.DatabaseFileName := ConnectionParams.Values['DatabaseFileName'];

  if ConnectionParams.Values[SCompressionAlgorithm] <> '' then
  begin
    BlobCompressionLevel :=
      StrToInt(ConnectionParams.Values[SCompressionLevel]);
    BlobCompressionAlgorithm :=
      TCompressionAlgorithm(StrToInt(ConnectionParams.Values[SCompressionAlgorithm]));
  end;

  if (Database.DatabaseFileName <> '') and Not FileExists(Database.DatabaseFileName) then begin
    ForceDirectories(ExtractFilePath(Database.DatabaseFileName));
//    Database.Password := ConnectionParams.Values['Password'];
//    if ConnectionParams.Values[SCryptoAlgorithm] <> '' then
//      Database.CryptoAlgorithm := TABSCryptoAlgorithm(StrToInt(ConnectionParams.Values[SCryptoAlgorithm]));
    Database.CreateDatabase;
  end
end;

function TovAccuracerDBAdapter.EditConnection(ConnectionParams: TStrings):
  Boolean;
begin
  Result := ShowAdapterParams(ConnectionParams)
end;

initialization
  TovAccuracerDBAdapter.RegisterAdapter(TovAccuracerDBAdapter);
finalization
  TovAccuracerDBAdapter.UnRegisterAdapter(TovAccuracerDBAdapter);
end.
