unit ovBDEAdapter;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, DBTables, ovTypes, ovModel, ovComponents;

type
  TovBDEAdapter = class(TovCustomDatabaseAdapter)
  private
    FDatabase: TDatabase;
    procedure SetDatabase(Value: TDatabase);
    procedure CheckDatabase;
    procedure DatabaseDisconnectHandler(Sender: TObject);
    procedure GenerateDatabaseName(Database: TDatabase);
  protected
    procedure CloseDatabase; override;
    procedure CommitTransaction; override;
    procedure CreateTable(ATableName: String; AFieldsDefs: TFieldDefs; 
      AnIndexDefs: TIndexDefs); override;
    procedure ExecQuery(SQLText: String); override;
    procedure OpenDatabase; override;
    function OpenQuery(QueryText: String; MasterDataSource: TDataSource; 
      LiveDataSet: Boolean): TDataSet; override;
    function OpenTable(ATableName: String; AReadOnly: Boolean; AFilter: String
      = ''): TDataSet; override;
    procedure RollbackTransaction; override;
    procedure StartTransaction; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function TableExists(ATableName: String): Boolean; override;
    function InTransaction: Boolean; override;
  public
    procedure SetupConnection(ConnectionParams: TStrings); override;
    function EditConnection(ConnectionParams: TStrings): Boolean; override;
    constructor Create(AOwner: TComponent); override;
  published
    property Database: TDatabase read FDatabase write SetDatabase;
  end;

procedure Register;

implementation

{$R *.dcr}

uses ovConsts, ovObjects, ovBDEParams;

procedure Register;
begin
  RegisterComponents(cpObjectViewAddOn, [TovBDEAdapter]);
end;

var DatabaseCount: Integer = 0;

{
******************************* TovBDEAdapter *******************************
}
constructor TovBDEAdapter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  with QuerySyntaxParams do begin
    LeftTableDelimiter := '';
    RightTableDelimiter := '';
    LeftFieldDelimiter := '';
    RightFieldDelimiter := '';
  end;
end;

procedure TovBDEAdapter.CloseDatabase;
begin
  if Assigned(Database) then Database.Close;
end;

procedure TovBDEAdapter.StartTransaction;
begin
  CheckDatabase;
  //      TransIsolation  tiDirtyRead
  if Not Database.IsSQLBased
    then Database.TransIsolation := tiDirtyRead;
  Database.StartTransaction
end;

procedure TovBDEAdapter.RollbackTransaction;
begin
  CheckDatabase;
  Database.Rollback
end;

procedure TovBDEAdapter.CommitTransaction;
begin
  CheckDatabase;
  Database.Commit;
end;

function TovBDEAdapter.InTransaction: Boolean;
begin
  CheckDatabase;
  Result := Database.InTransaction
end;

procedure TovBDEAdapter.CreateTable(ATableName: String; AFieldsDefs:
  TFieldDefs; AnIndexDefs: TIndexDefs);
var
  Table: TTable;
begin
  CheckDatabase;

  Table := TTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    TableName := ATableName;
    FieldDefs.Assign(AFieldsDefs);
    if AnIndexDefs <> nil
      then IndexDefs.Assign(AnIndexDefs);
    if Not Exists then CreateTable
  finally
    Free
  end;
end;

procedure TovBDEAdapter.ExecQuery(SQLText: String);
begin
  CheckDatabase;

  with TQuery.Create(nil) do
  try
    DatabaseName := Self.Database.DatabaseName;
    SQL.Text := SQLText;
    ExecSQL;
  finally
    Free
  end;
end;

procedure TovBDEAdapter.OpenDatabase;
begin
  CheckDatabase;
  Database.Open;
end;

function TovBDEAdapter.OpenQuery(QueryText: String; MasterDataSource: 
  TDataSource; LiveDataSet: Boolean): TDataSet;
begin
  CheckDatabase;

  Result := TQuery.Create(nil);
  try
    with TQuery(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      SQL.Text := QueryText;
      if MasterDataSource <> nil then
        DataSource := MasterDataSource;
      RequestLive := LiveDataSet;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

function TovBDEAdapter.OpenTable(ATableName: String; AReadOnly: Boolean; AFilter: String
      = ''):
  TDataSet;
begin
  CheckDatabase;

  Result := TTable.Create(nil);
  try
    with TTable(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      TableName := ATableName;
      ReadOnly := AReadOnly;
      Filter := AFilter;
      if AFilter <> '' then Filtered := True;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

procedure TovBDEAdapter.SetDatabase(Value: TDatabase);
begin
  if FDatabase <> Value then
  begin
    if Assigned(FDatabase) then FDatabase.AfterDisconnect := nil;
    FDatabase := Value;
    if Value <> nil then
    begin
      Value.FreeNotification(Self);
      Value.AfterDisconnect := DatabaseDisconnectHandler;
    end;
  end;
end;

{{
       
}
procedure TovBDEAdapter.DatabaseDisconnectHandler(Sender: TObject);
begin
  //       
  CloseDatabaseNotification
end;

procedure TovBDEAdapter.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = Database) and (Operation = opRemove)
    then Database := nil;
end;

procedure TovBDEAdapter.CheckDatabase;
begin
  if Not Assigned(Database) then
    raise Exception.Create('   Database  ' + Name);
end;

function TovBDEAdapter.TableExists(ATableName: String): Boolean;
var
  Table: TTable;
begin
  Table := TTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    TableName := ATableName;
    Result := Exists;
  finally
    Free
  end;
end;

procedure TovBDEAdapter.GenerateDatabaseName(Database: TDatabase);
var
  DatabaseCount: Integer;
begin
  DatabaseCount := 1;
  while True do
  try
    Inc(DatabaseCount);
    Database.DatabaseName := 'Database_' + IntToStr(DatabaseCount);
    Break;
  except
    //     DatabaseName   
  end;
end;

procedure TovBDEAdapter.SetupConnection(ConnectionParams: TStrings);
begin
  Database := TDatabase.Create(Self);
  if ConnectionParams.Values['DatabaseName'] = ''
    then GenerateDatabaseName(Database)
    else Database.DatabaseName := ConnectionParams.Values['DatabaseName'];

  with ConnectionParams do
  begin
    if Values['AliasName'] <> '' then
      Database.AliasName := Values['AliasName'];
    if Values['DriverName'] <> '' then
      Database.DriverName := Values['DriverName'];

    Database.Params := ConnectionParams;
    Database.Params.Values['DatabaseName'] := '';
    Database.Params.Values['AliasName'] := '';
    Database.Params.Values['DriverName'] := '';
    if Values['PATH'] <> '' then ForceDirectories(Values['PATH']);
  end
end;

function TovBDEAdapter.EditConnection(ConnectionParams: TStrings): Boolean;
begin
  Result := ShowAdapterParams(ConnectionParams)
end;



initialization
  TovBDEAdapter.RegisterAdapter(TovBDEAdapter);
finalization
  TovBDEAdapter.UnregisterAdapter(TovBDEAdapter);
end.
