unit ovDBISAM4Adapter;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, DBISAMTb, ovTypes, ovModel, ovComponents;

type
  TovDBISAM4Adapter = class(TovCustomDatabaseAdapter)
  private
    FDatabase: TDBISAMDatabase;
    FLocaleId: Integer;
    FSuppressBackups: Boolean;
    procedure RestructureTable(Table: TDBISAMTable; NewFieldsDefs: TFieldDefs;
      NewIndexDefs: TIndexDefs);
    procedure SetDatabase(Value: TDBISAMDatabase);
    procedure CheckDatabase;
    procedure GenerateDatabaseName(Database: TDBISAMDatabase);
  protected
    procedure CloseDatabase; override;
    procedure CommitTransaction; override;
    procedure CreateTable(ATableName: String; AFieldsDefs: TFieldDefs; 
      AnIndexDefs: TIndexDefs); override;
    procedure ExecQuery(SQLText: String); override;
    procedure OpenDatabase; override;
    function OpenQuery(QueryText: String; MasterDataSource: TDataSource;
      LiveDataSet: Boolean): TDataSet; override;
    function OpenTable(ATableName: String; AReadOnly: Boolean = False; AFilter:
       String = ''): TDataSet; override;
    procedure RollbackTransaction; override;
    procedure StartTransaction; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function TableExists(ATableName: String): Boolean; override;
    function InTransaction: Boolean; override;
    function ValueToSQLString(const Value: Variant; FieldType: TFieldType;
       FieldSize: Integer): String; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetupConnection(ConnectionParams: TStrings); override;
    function EditConnection(ConnectionParams: TStrings): Boolean; override;
  published
    property Database: TDBISAMDatabase read FDatabase write SetDatabase;
    property LocaleId: Integer read FLocaleId write FLocaleId;
    property SuppressBackups: Boolean read FSuppressBackups write FSuppressBackups
       default False;
  end;

procedure Register;

implementation

uses DBISAMCn, ovConsts, ovObjects, TypInfo, ovIObjects, ovDBISAMParams;

{$R ovDBISAMAdapter.dcr}

procedure Register;
begin
  RegisterComponents(cpObjectViewAddOn, [TovDBISAM4Adapter]);
end;

var DatabaseCount: Integer = 0;

{
******************************* TovDBISAM4Adapter *******************************
}

constructor TovDBISAM4Adapter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  with QuerySyntaxParams do begin
    LeftTableDelimiter := '"';
    RightTableDelimiter := '"';
    LeftFieldDelimiter := '"';
    RightFieldDelimiter := '"';
    FilterLeftFieldDelimiter := '"';
    FilterRightFieldDelimiter := '"';
  end;
end;

procedure TovDBISAM4Adapter.SetDatabase(Value: TDBISAMDatabase);
begin
  if FDatabase <> Value then
  begin
    FDatabase := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovDBISAM4Adapter.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = Database) and (Operation = opRemove)
    then Database := nil;
end;

procedure TovDBISAM4Adapter.CheckDatabase;
begin
  if Not Assigned(Database) then
    raise Exception.Create('   Database  ' + Name);
  if Database.DatabaseName = '' then
    raise Exception.Create('       ' + Name);
end;

procedure TovDBISAM4Adapter.OpenDatabase;
begin
  CheckDatabase;
  Database.Open;
end;

procedure TovDBISAM4Adapter.CloseDatabase;
begin
  if Assigned(Database) then Database.Close;
end;

procedure TovDBISAM4Adapter.StartTransaction;
begin
  CheckDatabase;
  Database.StartTransaction
end;

procedure TovDBISAM4Adapter.CommitTransaction;
begin
  CheckDatabase;
  Database.Commit;
end;

procedure TovDBISAM4Adapter.RollbackTransaction;
begin
  CheckDatabase;
  Database.Rollback
end;

function TovDBISAM4Adapter.InTransaction: Boolean;
begin
  CheckDatabase;
  Result := Database.InTransaction
end;

function TovDBISAM4Adapter.TableExists(ATableName: String): Boolean;
var
  Table: TDBISAMTable;
begin
  Table := TDBISAMTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    SessionName := Self.Database.SessionName;
    TableName := ATableName;
    Result := Exists;
  finally
    Free
  end;
end;

function TovDBISAM4Adapter.OpenTable(ATableName: String; AReadOnly: Boolean =
   False; AFilter: String = ''): TDataSet;
begin
  CheckDatabase;

  Result := TDBISAMTable.Create(nil);
  try
    with TDBISAMTable(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      SessionName := Self.Database.SessionName;
      TableName := ATableName;
      ReadOnly := AReadOnly;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

procedure TovDBISAM4Adapter.CreateTable(ATableName: String; AFieldsDefs:
  TFieldDefs; AnIndexDefs: TIndexDefs);
var
  Table: TDBISAMTable;
begin
  CheckDatabase;

  Table := TDBISAMTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    SessionName := Self.Database.SessionName;
    TableName := ATableName;
    FieldDefs.Assign(AFieldsDefs);
    if AnIndexDefs <> nil
      then IndexDefs.Assign(AnIndexDefs);
    if Not Exists
      then CreateTable(Self.LocaleId)
      else Self.RestructureTable(Table, AFieldsDefs, AnIndexDefs);
  finally
    Free
  end;
end;

{{
   .
}
procedure TovDBISAM4Adapter.RestructureTable(Table: TDBISAMTable; NewFieldsDefs:
  TFieldDefs; NewIndexDefs: TIndexDefs);
var
  I, J, N: Integer;
begin
  with Table do begin

    FieldDefs.Update; //    
    IndexDefs.Update; //    

    N := FieldDefs.Count;
    for I := 0 to NewFieldsDefs.Count - 1 do
    with NewFieldsDefs[I] do begin
      J := FieldDefs.IndexOf(Name);
      if J < 0 then begin //   
        Inc(N);
        FieldDefs.Add(N, Name, DataType, Size, Required);
      end else begin      //    
        FieldDefs[J].DataType := DataType;
        FieldDefs[J].Size := Size;
        FieldDefs[J].Required := Required;
      end
    end;

    //   
    IndexDefs.Clear;
    for I := 0 to NewIndexDefs.Count - 1 do
      with NewIndexDefs[I] do
        IndexDefs.Add(Name, Fields, Options);

    //   
    AlterTable(LocaleID, UserMajorVersion, UserMinorVersion, Encrypted,
      Password, Description, IndexPageSize, BlobBlockSize, LastAutoIncValue,
      TextIndexFields, TextIndexStopWords, TextIndexSpaceChars, TextIndexIncludeChars, SuppressBackups);
  end
end;

procedure TovDBISAM4Adapter.ExecQuery(SQLText: String);
begin
  CheckDatabase;

  with TDBISAMQuery.Create(nil) do
  try
    DatabaseName := Self.Database.DatabaseName;
    SessionName := Self.Database.SessionName;
    SQL.Text := SQLText;
    ExecSQL;
  finally
    Free
  end;
end;

function TovDBISAM4Adapter.OpenQuery(QueryText: String; MasterDataSource: 
  TDataSource; LiveDataSet: Boolean): TDataSet;
begin
  CheckDatabase;

  Result := TDBISAMQuery.Create(nil);
  try
    with TDBISAMQuery(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      SessionName := Self.Database.SessionName;
      SQL.Text := QueryText;
      DataSource := MasterDataSource;
      RequestLive := LiveDataSet;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

procedure TovDBISAM4Adapter.GenerateDatabaseName(Database: TDBISAMDatabase);
var
  DatabaseCount: Integer;
begin
  DatabaseCount := 1;
  while True do
  try
    Inc(DatabaseCount);
    Database.DatabaseName := 'Database_' + IntToStr(DatabaseCount);
    Break;
  except
    //     DatabaseName   
  end;
end;

procedure TovDBISAM4Adapter.SetupConnection(ConnectionParams: TStrings);
var
  Session: TDBISAMSession;
begin
  Session := TDBISAMSession.Create(Self);
  if ConnectionParams.Values['SessionName'] = ''
    then Session.AutoSessionName := True
    else Session.SessionName := ConnectionParams.Values['SessionName'];

  Database := TDBISAMDatabase.Create(Self);
  Database.SessionName := Session.SessionName;
  if ConnectionParams.Values['DatabaseName'] = ''
    then GenerateDatabaseName(Database)
    else Database.DatabaseName := ConnectionParams.Values['DatabaseName'];

  if UpperCase(ConnectionParams.Values['SessionType']) <> 'REMOTE' then
    Database.Directory := ConnectionParams.Values['Directory']
  else begin
  (*
    Database.RemoteDatabase := ConnectionParams.Values['RemoteDatabase'];

    Session.SessionType := stRemote;
    with ConnectionParams do
    begin
      Session.RemoteHost := Values['RemoteHost'];
      Session.RemoteAddress := Values['RemoteAddress'];
      if Values['RemotePort'] <> '' then
        Session.RemotePort := StrToInt(Values['RemotePort']);
      Session.RemoteService := Values['RemoteService'];
      if Values['RemoteType'] <> '' then
        Session.RemoteType := TRemoteType(GetEnumValue(TypeInfo(TRemoteType), Values['RemoteType']));
    end
  *)  
  end
end;

function TovDBISAM4Adapter.EditConnection(ConnectionParams: TStrings): Boolean;
begin
  Result := EditAdapterParams(ConnectionParams)
end;

var
  FormatSettings: TFormatSettings;

function TovDBISAM4Adapter.ValueToSQLString(const Value: Variant; FieldType:
  TFieldType; FieldSize: Integer): String;
const
  DateFormat = 'yyyy-mm-dd';
  TimeFormat = 'hh":"nn":"ss"."z';
begin
  case FieldType of
    ftDate:
      Result := '''' + FormatDateTime(DateFormat, Value)+ '''';
    ftFloat:
      Result := FloatToStr(Value, FormatSettings);
    ftTime:
      Result := '''' + FormatDateTime(TimeFormat, Value)+ '''';
    ftDateTime:
      Result := '''' + FormatDateTime(DateFormat+' '+TimeFormat, Value)+ '''';
    else Result := inherited ValueToSQLString(Value, FieldType, FieldSize)
  end
end;

initialization
  GetLocaleFormatSettings(LOCALE_USER_DEFAULT, FormatSettings);
  FormatSettings.DecimalSeparator := '.';

  TovDBISAM4Adapter.RegisterAdapter(TovDBISAM4Adapter);
finalization
  TovDBISAM4Adapter.UnregisterAdapter(TovDBISAM4Adapter);
end.
