unit ovDBISAMAdapter;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, DBISAMTb, ovTypes, ovModel, ovComponents;

type
  TovDBISAMAdapter = class (TovCustomDatabaseAdapter)
  private
    FDatabase: TDBISAMDatabase;
    FLanguageId: Integer;
    procedure RestructureTable(Table: TDBISAMTable; NewFieldsDefs: TFieldDefs; 
      NewIndexDefs: TIndexDefs);
    procedure SetDatabase(Value: TDBISAMDatabase);
    procedure CheckDatabase;
    procedure GenerateDatabaseName(Database: TDBISAMDatabase);
  protected
    procedure CloseDatabase; override;
    procedure CommitTransaction; override;
    procedure CreateTable(ATableName: String; AFieldsDefs: TFieldDefs; 
      AnIndexDefs: TIndexDefs); override;
    procedure ExecQuery(SQLText: String); override;
    procedure OpenDatabase; override;
    function OpenQuery(QueryText: String; MasterDataSource: TDataSource;
      LiveDataSet: Boolean): TDataSet; override;
    function OpenTable(ATableName: String; AReadOnly: Boolean; AFilter: String):
      TDataSet; override;
    procedure RollbackTransaction; override;
    procedure StartTransaction; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function TableExists(ATableName: String): Boolean; override;
    function InTransaction: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetupConnection(ConnectionParams: TStrings); override;
    function EditConnection(ConnectionParams: TStrings): Boolean; override;
  published
    property Database: TDBISAMDatabase read FDatabase write SetDatabase;
    property LanguageId: Integer read FLanguageId write FLanguageId;
  end;

procedure Register;

implementation

uses DBISAMCn, ovConsts, ovObjects, TypInfo, ovIObjects, ovDBISAMParams;

{$R *.dcr}

procedure Register;
begin
  RegisterComponents(cpObjectViewAddOn, [TovDBISAMAdapter]);
end;

var DatabaseCount: Integer = 0;

{
******************************* TovDBISAMAdapter *******************************
}
constructor TovDBISAMAdapter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  with QuerySyntaxParams do begin
    LeftTableDelimiter := '"';
    RightTableDelimiter := '"';
    LeftFieldDelimiter := '"';
    RightFieldDelimiter := '"';
  end;

  FLanguageId := LGID_RUSSIAN;
end;

procedure TovDBISAMAdapter.CloseDatabase;
begin
  if Assigned(Database) then Database.Close;
end;

procedure TovDBISAMAdapter.CommitTransaction;
begin
  CheckDatabase;
  Database.Commit;
end;

procedure TovDBISAMAdapter.CreateTable(ATableName: String; AFieldsDefs:
  TFieldDefs; AnIndexDefs: TIndexDefs);
var
  Table: TDBISAMTable;
begin
  CheckDatabase;

  Table := TDBISAMTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    SessionName := Self.Database.SessionName;
    TableName := ATableName;
    FieldDefs.Assign(AFieldsDefs);
    if AnIndexDefs <> nil
      then IndexDefs.Assign(AnIndexDefs);
    if Not Exists
      then CreateTable
      else Self.RestructureTable(Table, AFieldsDefs, AnIndexDefs);
  finally
    Free
  end;
end;

procedure TovDBISAMAdapter.ExecQuery(SQLText: String);
begin
  CheckDatabase;

  with TDBISAMQuery.Create(nil) do
  try
    DatabaseName := Self.Database.DatabaseName;
    SessionName := Self.Database.SessionName;
    SQL.Text := SQLText;
    ExecSQL;
  finally
    Free
  end;
end;

procedure TovDBISAMAdapter.OpenDatabase;
begin
  CheckDatabase;
  Database.Open;
end;

function TovDBISAMAdapter.OpenQuery(QueryText: String; MasterDataSource: 
  TDataSource; LiveDataSet: Boolean): TDataSet;
begin
  CheckDatabase;

  Result := TDBISAMQuery.Create(nil);
  try
    with TDBISAMQuery(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      SessionName := Self.Database.SessionName;
      SQL.Text := QueryText;
      DataSource := MasterDataSource;
      RequestLive := LiveDataSet;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

function TovDBISAMAdapter.OpenTable(ATableName: String; AReadOnly: Boolean;
  AFilter: String): TDataSet;
begin
  CheckDatabase;

  Result := TDBISAMTable.Create(nil);
  try
    with TDBISAMTable(Result) do begin
      DatabaseName := Self.Database.DatabaseName;
      SessionName := Self.Database.SessionName;
      TableName := ATableName;
      ReadOnly := AReadOnly;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

procedure TovDBISAMAdapter.RestructureTable(Table: TDBISAMTable; NewFieldsDefs: 
  TFieldDefs; NewIndexDefs: TIndexDefs);
var
  I, J, N: Integer;
begin
  with Table do begin

    //    
    RestructureFieldDefs.Update; //    
    N := RestructureFieldDefs.Count;
    for I := 0 to NewFieldsDefs.Count - 1 do
    with NewFieldsDefs[I] do begin
      J := RestructureFieldDefs.IndexOf(Name);
      if J < 0 then begin //   
        Inc(N);
        RestructureFieldDefs.Add(N, Name, DataType, Size, Required);
      end else begin      //    
        RestructureFieldDefs[J].DataType := DataType;
        RestructureFieldDefs[J].Size := Size;
        RestructureFieldDefs[J].Required := Required;
      end
    end;

    //   
    RestructureIndexDefs.Clear;
    for I := 0 to NewIndexDefs.Count - 1 do
      with NewIndexDefs[I] do
        RestructureIndexDefs.Add(Name, Fields, Options);

    //   
    RestructureTable(Self.LanguageID);
  end
end;

procedure TovDBISAMAdapter.RollbackTransaction;
begin
  CheckDatabase;
  Database.Rollback
end;

procedure TovDBISAMAdapter.SetDatabase(Value: TDBISAMDatabase);
begin
  if FDatabase <> Value then
  begin
    FDatabase := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovDBISAMAdapter.StartTransaction;
begin           
  CheckDatabase;
  Database.StartTransaction
end;

function TovDBISAMAdapter.InTransaction: Boolean;
begin
  CheckDatabase;
  Result := Database.InTransaction
end;

procedure TovDBISAMAdapter.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = Database) and (Operation = opRemove)
    then Database := nil;
end;

procedure TovDBISAMAdapter.CheckDatabase;
begin
  if Not Assigned(Database) then
    raise Exception.Create('   Database  ' + Name);
  if Database.DatabaseName = '' then
    raise Exception.Create('       ' + Name);
end;

function TovDBISAMAdapter.TableExists(ATableName: String): Boolean;
var
  Table: TDBISAMTable;
begin
  Table := TDBISAMTable.Create(nil);
  with Table do
  try
    DatabaseName := Self.Database.DatabaseName;
    SessionName := Self.Database.SessionName;
    TableName := ATableName;
    Result := Exists;
  finally
    Free
  end;
end;

procedure TovDBISAMAdapter.SetupConnection(ConnectionParams: TStrings);
var
  Session: TDBISAMSession;
begin
  Session := TDBISAMSession.Create(Self);
  if ConnectionParams.Values['SessionName'] = ''
    then Session.AutoSessionName := True
    else Session.SessionName := ConnectionParams.Values['SessionName'];

  Database := TDBISAMDatabase.Create(Self);
  Database.SessionName := Session.SessionName;
  if ConnectionParams.Values['DatabaseName'] = ''
    then GenerateDatabaseName(Database)
    else Database.DatabaseName := ConnectionParams.Values['DatabaseName'];
    
  if UpperCase(ConnectionParams.Values['SessionType']) <> 'REMOTE' then
    Database.Directory := ConnectionParams.Values['Directory']
  else begin
    Database.RemoteDatabase := ConnectionParams.Values['RemoteDatabase'];

    Session.SessionType := stRemote;
    with ConnectionParams do
    begin
      Session.RemoteHost := Values['RemoteHost'];
      Session.RemoteAddress := Values['RemoteAddress'];
      if Values['RemotePort'] <> '' then
        Session.RemotePort := StrToInt(Values['RemotePort']);
      Session.RemoteService := Values['RemoteService'];
      if Values['RemoteType'] <> '' then
        Session.RemoteType := TRemoteType(GetEnumValue(TypeInfo(TRemoteType), Values['RemoteType']));
    end
  end
end;

procedure TovDBISAMAdapter.GenerateDatabaseName(Database: TDBISAMDatabase);
var
  DatabaseCount: Integer;
begin
  DatabaseCount := 1;
  while True do
  try
    Inc(DatabaseCount);
    Database.DatabaseName := 'Database_' + IntToStr(DatabaseCount);
    Break;
  except
    //     DatabaseName   
  end;
end;

function TovDBISAMAdapter.EditConnection(ConnectionParams: TStrings): Boolean;
begin
  Result := ShowAdapterParams(ConnectionParams)
end;

initialization
  TovDBISAMAdapter.RegisterAdapter(TovDBISAMAdapter);
finalization
  TovDBISAMAdapter.UnRegisterAdapter(TovDBISAMAdapter);
end.
