unit ovNexusDB1Adapter;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, ovTypes, ovModel, ovComponents,
  nxsdServerEngine, nxsdTypes, nxdbBase, nxdb, nxsdDataDictionary;

type
  TovNexusDB1Adapter = class(TovCustomDatabaseAdapter)
  private
    FNewDatabaseBaseDirectory: String;
    FDatabase: TnxDatabase;
    FModule: TDataModule;
    procedure RestructureTable(Table: TnxTable; NewFieldsDefs: TFieldDefs;
      NewIndexDefs: TIndexDefs);
    procedure SetDatabase(Value: TnxDatabase);
    procedure CheckDatabase;
    procedure DatabaseStateChangedHandler(Sender: TObject);
  protected
    procedure CloneTable(SourceTable: TDataSet; SourceTableFieldNames: TStrings;
      ATableName: String; FieldsDefs: TFieldDefs; IndexDefs: TIndexDefs); override;
    procedure CloseDatabase; override;
    procedure CommitTransaction; override;
    procedure CreateTable(ATableName: String; AFieldsDefs: TFieldDefs;
      AnIndexDefs: TIndexDefs); override;
    procedure ExecQuery(SQLText: String); override;
    procedure OpenDatabase; override;
    function OpenQuery(QueryText: String; MasterDataSource: TDataSource;
      LiveDataSet: Boolean): TDataSet; override;
    function OpenTable(ATableName: String; AReadOnly: Boolean = False; AFilter:
      String = ''): TDataSet; override;
    procedure RollbackTransaction; override;
    procedure StartTransaction; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function TableExists(ATableName: String): Boolean; override;
    function InTransaction: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure CreateDatabase(DatabaseName: String; ConnectionParams: TStrings);
      override;
    procedure SetupConnection(ConnectionParams: TStrings); override;
    function EditConnection(ConnectionParams: TStrings): Boolean; override;
  published
    property NewDatabaseBaseDirectory: String read FNewDatabaseBaseDirectory write
      FNewDatabaseBaseDirectory;
    property Database: TnxDatabase read FDatabase write SetDatabase;
  end;

procedure Register;

implementation

uses ovConsts, ovObjects, TypInfo, ovIObjects,
     nxllComponent, ovNexusDB1Params, ovNexusDB1Local, ovNexusDB1Remote,
     ovUtils;

{$R *.dcr}

procedure Register;
begin
  RegisterComponents(cpObjectViewAddOn, [TovNexusDB1Adapter]);
end;

const
  nxtUnknown = nxtByte;

  FieldTypeMap: array [ftUnknown..ftFMTBcd] of TnxFieldType =
   (nxtUnknown, nxtNullString, nxtInt16, nxtInt32, nxtWord16, nxtBoolean,
    nxtDouble, nxtCurrency, nxtUnknown, nxtDate, nxtTime, nxtDateTime, nxtByteArray,
    nxtByteArray, nxtAutoInc, nxtBlob, nxtBlobMemo, nxtBlobGraphic, nxtBlobMemo,
    nxtBlob, nxtBlob, nxtUnknown, nxtUnknown, nxtChar, nxtWideString,
    nxtInt64, nxtUnknown, nxtUnknown, nxtUnknown, nxtUnknown, nxtBlob, nxtBlob,
    nxtUnknown, nxtUnknown, nxtUnknown, nxtUnknown, nxtDateTime, nxtUnknown);

{
******************************* TovNexusDB1Adapter *****************************
}
constructor TovNexusDB1Adapter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  with QuerySyntaxParams do begin
    LeftTableDelimiter := '"';
    RightTableDelimiter := '"';
    LeftFieldDelimiter := '"';
    RightFieldDelimiter := '"';
    FilterLeftFieldDelimiter := '[';
    FilterRightFieldDelimiter := ']';
  end;
  FFormatSettings.DecimalSeparator := '.';
end;

procedure TovNexusDB1Adapter.CloseDatabase;
begin
  if Assigned(Database) and (Database.State = nxsStarted) then Database.Close;
end;

procedure TovNexusDB1Adapter.CommitTransaction;
begin
  CheckDatabase;
  Database.Commit;
end;

procedure TovNexusDB1Adapter.CreateTable(ATableName: String; AFieldsDefs:
  TFieldDefs; AnIndexDefs: TIndexDefs);
var
  Table: TnxTable;
begin
  CheckDatabase;

  Table := TnxTable.Create(nil);
  with Table do
  try
    Database := Self.Database;
    Session := Self.Database.Session;
    TableName := ATableName;
    FieldDefs.Assign(AFieldsDefs);
    if AnIndexDefs <> nil
      then IndexDefs.Assign(AnIndexDefs);
    if Not Exists
      then CreateTable
      else Self.RestructureTable(Table, AFieldsDefs, AnIndexDefs);
  finally
    Free
  end;
end;

procedure TovNexusDB1Adapter.ExecQuery(SQLText: String);
begin
  CheckDatabase;

  with TnxQuery.Create(nil) do
  try
    Database := Self.Database;
    Session := Self.Database.Session;
    SQL.Text := SQLText;
    ExecSQL;
  finally
    Free
  end;
end;

procedure TovNexusDB1Adapter.OpenDatabase;
begin
  CheckDatabase;
  Database.Open;
end;

function TovNexusDB1Adapter.OpenQuery(QueryText: String; MasterDataSource: 
  TDataSource; LiveDataSet: Boolean): TDataSet;
begin
  CheckDatabase;

  Result := TnxQuery.Create(nil);
  try
    with TnxQuery(Result) do begin
      Database := Self.Database;
      Session := Self.Database.Session;
      SQL.Text := QueryText;
      DataSource := MasterDataSource;
      RequestLive := LiveDataSet;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

function TovNexusDB1Adapter.OpenTable(ATableName: String; AReadOnly: Boolean = 
  False; AFilter: String = ''): TDataSet;
begin
  CheckDatabase;

  Result := TnxTable.Create(nil);
  try
    with TnxTable(Result) do begin
      Database := Self.Database;
      Session := Self.Database.Session;
      TableName := ATableName;
      ReadOnly := AReadOnly;
      Filter := AFilter;
      if AFilter <> '' then Filtered := True;
    end;
    Result.Open;
  except
    Result.Free; raise;
  end;
end;

procedure TovNexusDB1Adapter.RestructureTable(Table: TnxTable; NewFieldsDefs:
  TFieldDefs; NewIndexDefs: TIndexDefs);
var
  Status: TnxTaskStatus;
  Completed: Boolean;
  I: Integer;
  ADictionary: TnxDataDictionary;
  AFieldMap, OldFields: TStrings;
  ATaskInfo: TnxAbstractTaskInfo;
  OldField: String;

  function FindOldField(NewField: String; var OldField: String): Boolean;
  var
    I: Integer;
  begin
    for I := 0 to OldFields.Count - 1 do
      if AnsiUpperCase(OldFields[I]) = AnsiUpperCase(NewField) then
      begin
        OldField := OldFields[I];
        Result := True;
        Exit;
      end;
    Result := False;
  end;

begin
  AFieldMap := nil;
  OldFields := nil;
  ADictionary := nil;
  ATaskInfo := nil;
  try
    ADictionary := TnxDataDictionary.Create;
    AFieldMap := TStringList.Create;
    OldFields := TStringList.Create;

    with ADictionary do
    begin
      Clear;
      Table.GetFieldNames(OldFields);
      for I := 0 to NewFieldsDefs.Count - 1 do
        with NewFieldsDefs[I] do
        begin
          AddField(Name, '', FieldTypeMap[DataType], Size, Precision, Required);
          if FindOldField(Name, OldField) then
            AFieldMap.Values[Name] := OldField;
        end;
    end;

    //   
    Check(Database.RestructureTable(Table.TableName, ADictionary, AFieldMap, ATaskInfo));
    repeat
      Application.ProcessMessages;
      ATaskInfo.GetStatus(Completed, Status);
    until Completed;
    Check(Status.tsErrorCode);

    //      
    Table.IndexDefs.Update;
    for I := Table.IndexDefs.Count - 1 downto 0 do
      Table.DeleteIndex(Table.IndexDefs[I].Name);
    for I := 0 to NewIndexDefs.Count - 1 do
      Table.AddIndex(NewIndexDefs[I].Name, NewIndexDefs[I].Fields, NewIndexDefs[I].Options);

  finally
    ATaskInfo.Free;
    OldFields.Free;
    AFieldMap.Free;
    ADictionary.Free;
  end;
end;

procedure TovNexusDB1Adapter.RollbackTransaction;
begin
  CheckDatabase;
  Database.Rollback
end;

procedure TovNexusDB1Adapter.SetDatabase(Value: TnxDatabase);
begin
  if FDatabase <> Value then
  begin
    if Assigned(FDatabase) then FDatabase.OnStateChanged := nil;
    FDatabase := Value;
    if Value <> nil then
    begin
      Value.FreeNotification(Self);
      Value.OnStateChanged := DatabaseStateChangedHandler;
    end;
  end;
end;

{{
      
}
procedure TovNexusDB1Adapter.DatabaseStateChangedHandler(Sender: TObject);
begin
  //       
  if Database.State = nxsInactive then CloseDatabaseNotification
end;

procedure TovNexusDB1Adapter.StartTransaction;
begin
  CheckDatabase;
  Database.StartTransaction
end;

function TovNexusDB1Adapter.InTransaction: Boolean;
begin
  CheckDatabase;
  Result := Database.InTransaction
end;

procedure TovNexusDB1Adapter.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = Database) and (Operation = opRemove)
    then Database := nil;
end;

procedure TovNexusDB1Adapter.CheckDatabase;
begin
  if Not Assigned(Database) then
    raise Exception.Create('   Database  ' + Name);
end;

function TovNexusDB1Adapter.TableExists(ATableName: String): Boolean;
var
  Table: TnxTable;
begin
  CheckDatabase;
  Table := TnxTable.Create(nil);
  with Table do
  try
    Database := Self.Database;
    Session := Self.Database.Session;
    TableName := ATableName;
    Result := Exists;
  finally
    Free
  end;
end;

procedure TovNexusDB1Adapter.SetupConnection(ConnectionParams: TStrings);
begin
  if ConnectionParams.Values['DatabaseDirectory'] <> '' then begin
    //  
    FModule := TdmNexusLocalDatabase.Create(Self);
    Database := TdmNexusLocalDatabase(FModule).nxDatabase;
    Database.AliasPath := ConnectionParams.Values['DatabaseDirectory'];
    if NewDatabaseBaseDirectory = '' then
      NewDatabaseBaseDirectory := ConnectionParams.Values['DatabaseDirectory'];
  end else begin
    //   -   Winsock
    FModule := TdmNexusRemoteDatabase.Create(Self);
    Database := TdmNexusRemoteDatabase(FModule).nxDatabase;
    TdmNexusRemoteDatabase(FModule).nxWinsockTransport.ServerNameRuntime := ConnectionParams.Values['ServerName'];
    Database.AliasName := ConnectionParams.Values['DatabaseAliasName'];
    if NewDatabaseBaseDirectory = '' then
      NewDatabaseBaseDirectory := ConnectionParams.Values['NewDatabaseBaseDirectory'];
  end
end;

{{
     .   ConnectionParams 
   .
}
procedure TovNexusDB1Adapter.CreateDatabase(DatabaseName: String;
  ConnectionParams: TStrings);
var
  DBPath: String;
begin
  DBPath := IncludeTrailingPathDelimiter(NewDatabaseBaseDirectory) + DatabaseName;
  ForceDirectories(DBPath);
  ConnectionParams.Values['DatabaseDirectory'] := DBPath;
end;

{{
    ATableName,        
  SourceTable,      .
}
procedure TovNexusDB1Adapter.CloneTable(SourceTable: TDataSet;
  SourceTableFieldNames: TStrings; ATableName: String; FieldsDefs: TFieldDefs;
  IndexDefs: TIndexDefs);
var
  TargetTable: TnxTable;
  RequiredType: TFieldType;
begin
  //  ,    
  TargetTable := TnxTable.Create(nil);
  with TargetTable do
  try
    Database := Self.Database;
    Session := Self.Database.Session;
    TableName := ATableName;
    if Exists then DeleteTable;
  finally
    Free
  end;

  //  AutoInc    
  RequiredType := FieldsDefs[0].DataType;
  FieldsDefs[0].DataType := ftInteger;
  CreateTable(ATableName, FieldsDefs, IndexDefs);

  //         
  TargetTable := OpenTable(ATableName) as TnxTable;
  try
    CopyData(SourceTable, SourceTableFieldNames, ATableName, TargetTable);
    //     
    FieldsDefs[0].DataType := RequiredType;
    TargetTable.Close;
    RestructureTable(TargetTable, FieldsDefs, IndexDefs);
  finally
    TargetTable.Free;
  end;
end;

function TovNexusDB1Adapter.EditConnection(ConnectionParams: TStrings): Boolean;
begin
  Result := ShowAdapterParams(ConnectionParams)
end;

initialization
  TovNexusDB1Adapter.RegisterAdapter(TovNexusDB1Adapter);
finalization
  TovNexusDB1Adapter.UnregisterAdapter(TovNexusDB1Adapter);
end.
