unit ovCxDBTreeListCustomization;

interface

uses SysUtils, Classes, Forms, cxDBTL,
     ovDataSet, ovDataLinks, ovCustomization;

type
  TovCxDBTreeListCustomization = class(TovCustomization)
  private
    FObjectExposer: TovExposer;
    FOnCreateColumns: TNotifyEvent;
    FDataLinks: TovDataLinks;
    FRequiredFields: TStrings;
    FSelectedAttributes: String;
    procedure SetTreeList(Value: TcxDBTreeList);
    function GetTreeList: TcxDBTreeList;
    procedure CheckTreeList;
    procedure ColumnsChanged;
    procedure CreateColumn(FieldName: String; Hidden: Boolean);
    procedure SetDataLinks(const Value: TovDataLinks);
    procedure InitColumnProperties(AColumn: TcxDBTreeListColumn);
    procedure SetObjectExposer(const Value: TovExposer);
    procedure SetRequiredFields(Value: TStrings);
  protected
    procedure CreateColumns(Attributes: TStrings; AExposer: TovExposer);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property ObjectExposer: TovExposer read FObjectExposer write SetObjectExposer;
  public
    procedure LoadSettings; override;
    procedure SaveSettings; override;
    procedure SelectColumns;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property TreeList: TcxDBTreeList read GetTreeList write SetTreeList;
    {{
        ,     (TovExposer),
      TreeList.DataController.DataSet,  -  ,  
     TreeList   DataLinks.
    }
    property DataLinks: TovDataLinks read FDataLinks write SetDataLinks;
    {{
       ,       
    .  ,          ,
         .
    }
    property RequiredFields: TStrings read FRequiredFields write SetRequiredFields;
    {{
            
      /         .
           ,  
        ,       .
    }
    property OnCreateColumns: TNotifyEvent read FOnCreateColumns write
      FOnCreateColumns;
  end;

implementation

uses ovAttributeSelectionDlg, ovUtils, DB, cxCheckBox, ovDataObjects, cxTL;

resourcestring
  SSelectedAttributesRegKeyName = 'SelectedAttributes';
  SColumnFieldsRegKeyName = 'ColumnFields';

{ TovCxDBTreeListCustomization }

constructor TovCxDBTreeListCustomization.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRequiredFields := TStringList.Create;
end;

destructor TovCxDBTreeListCustomization.Destroy;
begin
  FRequiredFields.Free;
  inherited Destroy;
end;

{***************************************************************************** }
{                                                             }
{ **************************************************************************** }

procedure TovCxDBTreeListCustomization.SetTreeList(Value: TcxDBTreeList);
begin
  SubjectComponent := Value
end;

function TovCxDBTreeListCustomization.GetTreeList: TcxDBTreeList;
begin
  Result := SubjectComponent as TcxDBTreeList;
end;

procedure TovCxDBTreeListCustomization.CheckTreeList;
begin
  if Not Assigned(TreeList) then raise Exception.Create('   TreeList :' + Name);
end;

procedure TovCxDBTreeListCustomization.SetDataLinks(const Value: TovDataLinks);
begin
  if FDataLinks <> Value then
  begin
    FDataLinks := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovCxDBTreeListCustomization.SetObjectExposer(const Value:
  TovExposer);
begin
  if FObjectExposer <> Value then
  begin
    FObjectExposer := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovCxDBTreeListCustomization.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = DataLinks) then DataLinks := nil else
  if (Operation = opRemove) and (AComponent = ObjectExposer) then
  begin
    CheckSaveSettings;
    FObjectExposer := nil
  end;
end;

procedure TovCxDBTreeListCustomization.SetRequiredFields(Value: TStrings);
begin
  FRequiredFields.Assign(Value);
end;

{ **************************************************************************** }
{                                                   }
{ **************************************************************************** }

procedure TovCxDBTreeListCustomization.SaveSettings;

  procedure SaveAdditionalSettings(ARegKey: String);
  var
    ColumnFieldNames: TStringList;
    I: Integer;
  begin
    SaveStringToRegistry(ARegKey, SSelectedAttributesRegKeyName, FSelectedAttributes);

    ColumnFieldNames := TStringList.Create;
    try
      with TreeList do
      for I := 0 to ColumnCount - 1 do
        ColumnFieldNames.Values[Columns[I].Name] := Columns[I].DataBinding.FieldName;
      SaveStringToRegistry(ARegKey, SColumnFieldsRegKeyName, ColumnFieldNames.Text);
    finally
      ColumnFieldNames.Free;
    end;
  end;

var
  ARegistryKey: String;
begin
  inherited;
  if Assigned(TreeList) then
  begin
    ARegistryKey := GetFullRegistryKey + '\' + TreeList.Name;
    TreeList.StoreToRegistry(ARegistryKey, True);
    SaveAdditionalSettings(ARegistryKey);
  end;
end;

procedure TovCxDBTreeListCustomization.LoadSettings;

  function LoadAdditionalSettings1(ARegKey: String): Boolean;
  var
    DataLink: TovDataLink;
    DataObject: TovDataObject;
    NewAttributes: String;
  begin
    Result := False;
    ObjectExposer := TreeList.DataController.DataSet as TovExposer;

    FSelectedAttributes := LoadStringFromRegistry(ARegKey, SSelectedAttributesRegKeyName);
    NewAttributes := CombineSubStr(FSelectedAttributes, ',', AttributesFromList(RequiredFields));
    if NewAttributes = '' then Exit;

    if Assigned(DataLinks) then
    begin
      DataLink := DataLinks.FindLink(TreeList);
      if Assigned(DataLink) then
      begin
        DataObject := DataLinks.DataObjects.FindObjectByName(DataLink.DataObject);
        if Assigned(DataObject) then
        begin
          { TODO :  ,     -   , 
             Exposer'     ObjectAttributes }
          if DataObject.DataSet <> nil then
            (DataObject.DataSet as TovExposer).ObjectAttributes := NewAttributes;

          DataObject.ObjectAttributes := NewAttributes;
          DataObject.Open;
          ObjectExposer := DataObject.DataSet as TovExposer;
        end;
      end;
    end else
    begin
      ObjectExposer.ObjectAttributes := NewAttributes;
      ObjectExposer.Open;
    end;
    Result := True;
  end;

  procedure LoadAdditionalSettings2(ARegKey: String);

  var
    ColumnFieldNames: TStringList;
    I: Integer;
    SaveCaption: String;
    SaveWidth: Integer;
  begin
    ColumnFieldNames := TStringList.Create;
    try
      ColumnFieldNames.Text := LoadStringFromRegistry(ARegKey, SColumnFieldsRegKeyName);
      for I := 0 to TreeList.ColumnCount - 1 do
      with TreeList.Columns[I] do
      begin
        SaveCaption := Caption.Text;
        SaveWidth := Width;
        DataBinding.FieldName := ColumnFieldNames.Values[Name];
        Caption.Text := SaveCaption;
        Width := SaveWidth;
      end;
    finally
      ColumnFieldNames.Free;
    end;

    for I := 0 to TreeList.ColumnCount - 1 do
      InitColumnProperties(TreeList.Columns[I]);
  end;

var
  ARegistryKey: String;
begin
  if Assigned(TreeList) then
  begin
    ARegistryKey := GetFullRegistryKey + '\' + TreeList.Name;
    if LoadAdditionalSettings1(ARegistryKey) then
    begin
      TreeList.DeleteAllColumns;
      TreeList.RestoreFromRegistry(ARegistryKey, True, True);
      LoadAdditionalSettings2(ARegistryKey);
      ColumnsChanged;
    end;
  end;

  inherited;
end;

{ **************************************************************************** }
{                                                      }
{ **************************************************************************** }

{{
       , 
          
.  ,   ,   
RequiredFields.        Id,  
TovExposer   dsoExposeIdField.
}
procedure TovCxDBTreeListCustomization.SelectColumns;
var
  AttributeList: TStringList;
  AExposer: TovExposer;
begin
  CheckTreeList;
  AExposer := TreeList.DataController.DataSet as TovExposer;
  with AExposer do
  begin
    if SelectObjectAttributes(Model,
      Model.DomainModel.GetObject(ObjectClassName), FSelectedAttributes,
        ' ', [soExpandRelations])
    then begin
      {  FSelectedAttributes  ,   }
      {  ObjectAttributes       }
      ObjectAttributes := CombineSubStr(FSelectedAttributes, ',', AttributesFromList(RequiredFields));
      { TODO :    }
      Open;
      AttributeList := TStringList.Create;
      try
        AttributesToList(FSelectedAttributes, AttributeList);
        CreateColumns(AttributeList, AExposer);
        ColumnsChanged;
      finally
        AttributeList.Free;
      end;
    end;
  end;
end;

{{
        Attributes. (* ,
     ,    RequiredFields, *) -  
    Id,   TovExposer   dsoExposeIdField.
}
procedure TovCxDBTreeListCustomization.CreateColumns(Attributes: TStrings;
  AExposer: TovExposer);

  function ExistGridViewItem(FieldName: String): Boolean;
  var
    I: Integer;
  begin
    for I := 0 to TreeList.ColumnCount - 1 do
      if TreeList.Columns[I].DataBinding.FieldName = FieldName then
      begin
        Result := True;
        Exit;
      end;
    Result := False;
  end;

var
  I: Integer;
begin
  {    }
  for I := TreeList.ColumnCount - 1 downto 0 do
    if Attributes.IndexOf(TreeList.Columns[I].DataBinding.FieldName) < 0 then
      TreeList.Columns[I].Free;

  {    }
  for I := 0 to Attributes.Count - 1 do
    if Not ExistGridViewItem(Attributes[I]) then
      CreateColumn(Attributes[I], False);

//  for I := 0 to RequiredFields.Count - 1 do
//    if Not ExistGridViewItem(RequiredFields[I]) then
//      CreateColumn(RequiredFields[I], True);

  if dsoExposeIdField in AExposer.Options then
    CreateColumn(AExposer.IdFieldName, True);

  TreeList.ApplyBestFit;
end;

{{
   ,    FieldName  . 
 Hidden  True,     .
}
procedure TovCxDBTreeListCustomization.CreateColumn(FieldName: String; Hidden:
  Boolean);
var
  AColumn: TcxDBTreeListColumn;
begin
  AColumn := TcxDBTreeListColumn(TreeList.CreateColumn);
  GenerateComponentName(AColumn, TreeList.Name + 'Column');
  AColumn.DataBinding.FieldName := FieldName;
  InitColumnProperties(AColumn);
  AColumn.Options.Hidden := Hidden;
  AColumn.Visible := Not Hidden;
end;

procedure TovCxDBTreeListCustomization.InitColumnProperties(AColumn:
  TcxDBTreeListColumn);
begin
  AColumn.Caption.AlignHorz := taCenter;
  if AColumn.DataBinding.Field <> nil then
  case AColumn.DataBinding.Field.DataType of
    ftBoolean:
    begin
      AColumn.PropertiesClass := TcxCheckBoxProperties;
      (AColumn.Properties as TcxCheckBoxProperties).NullStyle := nssUnchecked;
    end;
  end;
end;

procedure TovCxDBTreeListCustomization.ColumnsChanged;
begin
  if Assigned(FOnCreateColumns) then FOnCreateColumns(Self);
end;

end.
