unit ovActions;

interface

uses Windows, Classes, Dialogs, Controls, Forms, ActnList, DBActns,
     ovDataSet, ovDataObjects, ovIObjects;

type
  TActionType = (atCreateObject, atCloneObject, atEditObject,
    atDeleteObject, atSelectObject);

  TEditObjectEvent = procedure (Sender: TObject; AObject: IovObject) of object;

  TovDataSetCustomAction = class(TDataSetAction)
  private
    FActionArgument: String;
    FActionType: TActionType;
    FConfirmDelete: Boolean;
    FConfirmDeleteMsg: String;
    FDeleteAll: Boolean;
    FOnAfterExecute: TNotifyEvent;
    FOnBeforeExecute: TNotifyEvent;
    FOnEditObject: TEditObjectEvent;
    FNewObjectClassName: String;
    FDisabled: Boolean;
    procedure SetDeleteAll(Value: Boolean);
    function IsStoredConfirmDeleteMsg: Boolean;
  protected
    function GetOvDataSet(Target: TObject): TovDataSet; virtual; abstract;
  public
    procedure ExecuteTarget(Target: TObject); override;
    procedure UpdateTarget(Target: TObject); override;
    constructor Create(AOwner: TComponent); override;
  published
    property ActionArgument: String read FActionArgument write FActionArgument;
    property ActionType: TActionType read FActionType write FActionType default 
      atCreateObject;
    property ConfirmDelete: Boolean read FConfirmDelete write FConfirmDelete 
      default True;
    property ConfirmDeleteMsg: String read FConfirmDeleteMsg write 
      FConfirmDeleteMsg stored IsStoredConfirmDeleteMsg;
    property DeleteAll: Boolean read FDeleteAll write SetDeleteAll default False;
    {{
    ,       atCreateObject. 
      ,      ,   
        .
    }
    property NewObjectClassName: String read FNewObjectClassName write
      FNewObjectClassName;
    {{
         Action   
      .     
      Enabled.
    }
    property Disabled: Boolean read FDisabled write FDisabled default False;
    property OnAfterExecute: TNotifyEvent read FOnAfterExecute write
      FOnAfterExecute;
    property OnBeforeExecute: TNotifyEvent read FOnBeforeExecute write
      FOnBeforeExecute;
    property OnEditObject: TEditObjectEvent read FOnEditObject write FOnEditObject;
  end;

  TovDataObjectsAction = class(TovDataSetCustomAction)
  private
    FDataObjectName: String;
    FDataObjects: TovDataObjects;
    procedure SetDataObjectName(const Value: String);
    procedure SetDataObjects(Value: TovDataObjects);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function GetOvDataSet(Target: TObject): TovDataSet; override;
  public
    function HandlesTarget(Target: TObject): Boolean; override;
  published
    property DataObjectName: String read FDataObjectName write SetDataObjectName;
    property DataObjects: TovDataObjects read FDataObjects write SetDataObjects;
  end;

  TovDataSetAction = class(TovDataSetCustomAction)
  private
  protected
    function GetOvDataSet(Target: TObject): TovDataSet; override;
  public
  published
    property DataSource;
  end;

implementation

uses ovStdOperations;

resourcestring
  SDefaultCreateObjectCaption = '   ';
  SDefaultDeleteObjectCaption = '   ';
  SDefaultDeleteAllObjectsCaption = '    ';
  SDefaultConfirmDeleteMsg = '   ?';
  SDefaultConfirmDeleteAllMsg = '   ?';

{ ------------------------ TovCustomDataObjectsAction -------------------------}

constructor TovDataSetCustomAction.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FConfirmDelete := True;
  FConfirmDeleteMsg := SDefaultConfirmDeleteMsg;
end;

function TovDataSetCustomAction.IsStoredConfirmDeleteMsg: Boolean;
begin
  if DeleteAll then
    Result := ConfirmDeleteMsg <> SDefaultConfirmDeleteAllMsg
  else
    Result := ConfirmDeleteMsg <> SDefaultConfirmDeleteMsg;
end;

procedure TovDataSetCustomAction.SetDeleteAll(Value: Boolean);
begin
  if FDeleteAll <> Value then begin
    FDeleteAll := Value;
    if DeleteAll then
    begin
      if (ConfirmDeleteMsg = '') or (ConfirmDeleteMsg = SDefaultConfirmDeleteMsg)
        then ConfirmDeleteMsg := SDefaultConfirmDeleteAllMsg
    end else
    begin
      if (ConfirmDeleteMsg = '') or (ConfirmDeleteMsg = SDefaultConfirmDeleteAllMsg)
        then ConfirmDeleteMsg := SDefaultConfirmDeleteMsg
    end
  end
end;

procedure TovDataSetCustomAction.UpdateTarget(Target: TObject);
var
  DataSet: TovDataSet;
begin
  DataSet := GetOvDataSet(Target);
  if Not Assigned(DataSet) or Disabled then Enabled := False
  else
  case ActionType of
    atCreateObject:
      with DataSet do
        Enabled := Active and
          (Not Assigned(MasterDataSet) or
           Not (MasterDataSet.Bof and MasterDataSet.Eof));

    atEditObject, atCloneObject, atDeleteObject, atSelectObject:
      Enabled := DataSet.IsCurrentObjectAvailable;
  end
end;

procedure TovDataSetCustomAction.ExecuteTarget(Target: TObject);
var
  DataSet: TovDataSet;
  AObject: IovObject;
  AObjectClassName: String;
begin
  if Assigned(FOnBeforeExecute) then FOnBeforeExecute(Self);

  case ActionType of
    atCreateObject:
    begin
      DataSet := GetOvDataSet(Target);
      if Not DataSet.IsMasterDetail then
      begin
        if NewObjectClassName <> '' then
          AObjectClassName := NewObjectClassName
        else
          AObjectClassName := (DataSet as TovExposer).ObjectClassName;
        if NewObject(DataSet.ModelName, AObjectClassName, [ActionArgument], AObject) then
        begin
          (DataSet as TovExposer).FindObject(AObject.ObjectId);
          if Assigned(FOnEditObject) then FOnEditObject(Self, AObject);
        end;
      end else
      begin
        if NewChildObject(DataSet.MasterDataSet.CurrentObject,
          DataSet.MasterRelationName, [ActionArgument], AObject) and
          Assigned(FOnEditObject) then FOnEditObject(Self, AObject);
      end
    end;

    atCloneObject:
      begin
        if NewObjectCopy((GetOvDataSet(Target) as TovDataSet).CurrentObject,
          [ActionArgument], AObject) and
          Assigned(FOnEditObject) then FOnEditObject(Self, AObject);
      end;

    atEditObject:
      begin
        AObject := (GetOvDataSet(Target) as TovDataSet).CurrentObject;
        if EditObject(AObject, [ActionArgument]) and
          Assigned(FOnEditObject) then FOnEditObject(Self, AObject);
      end;

    atDeleteObject:
      if Not ConfirmDelete or (Application.MessageBox(PChar(ConfirmDeleteMsg), '',
        MB_ICONQUESTION+MB_YESNO) = IDYES)
      then
        if DeleteAll
          then GetOvDataSet(Target).DeleteAll
          else GetOvDataSet(Target).Delete;

    atSelectObject:
      begin
        GetOvDataSet(Target).CheckBrowseMode;
        (Owner as TForm).ModalResult := mrOk;
      end;
  end;

  if Assigned(FOnAfterExecute) then FOnAfterExecute(Self);
end;


{ ------------------------------ TovDataObjectsAction -------------------------}

procedure TovDataObjectsAction.SetDataObjectName(const Value: String);
var
  OldValue: String;
begin
  if FDataObjectName <> Value then
  begin
    OldValue := DataObjectName;
    FDataObjectName := Value;
    if (Caption = Name) or (Caption = SDefaultDeleteObjectCaption + OldValue) then
      Caption := SDefaultDeleteObjectCaption + Value;
  end
end;

procedure TovDataObjectsAction.SetDataObjects(Value: TovDataObjects);
begin
  if Value <> FDataObjects then
  begin
    FDataObjects := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

function TovDataObjectsAction.HandlesTarget(Target: TObject): Boolean;
var
  DataObject: TovDataObject;
begin
  if Assigned(DataObjects) then
  begin
    DataObject := DataObjects.FindObjectByName(DataObjectName);
    Result := Assigned(DataObject) and Assigned(DataObject.DataSet) and
      DataObject.DataSet.Active;
  end else
    Result := False
end;

procedure TovDataObjectsAction.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = DataObjects) then DataObjects := nil;
end;

function TovDataObjectsAction.GetOvDataSet(Target: TObject): TovDataSet;
begin
  Result := DataObjects[DataObjectName]
end;

{ -------------------------------- TovDataSetAction ---------------------------}

function TovDataSetAction.GetOvDataSet(Target: TObject): TovDataSet;
begin
  if DataSource = nil then
    Result := nil
  else
    Result := DataSource.DataSet as TovDataSet
end;

end.
