unit ovCustomization;

interface

uses SysUtils, Classes, Forms, ovDataSet;

type
  TovCustomization = class(TComponent)
  private
    FSubjectComponent: TComponent;
    FAutoLoad: Boolean;
    FAutoSave: Boolean;
    FOnLoadSettings: TNotifyEvent;
    FOnSaveSettings: TNotifyEvent;
    FSettingsSaved: Boolean;
    FProfileName: String;
    FRegistryKey: String;
    FSectionName: String;
    FOnLoadSettingsError: TNotifyEvent;
    FOnSaveSettingsError: TNotifyEvent;
    procedure SetSubjectComponent(Value: TComponent);
    function IsStoredProfileName: Boolean;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
    procedure CheckSaveSettings;
    property SubjectComponent: TComponent read FSubjectComponent write
      SetSubjectComponent;
  public
    procedure LoadSettings; dynamic;
    procedure SaveSettings; dynamic;
    destructor Destroy; override;
    constructor Create(AOwner: TComponent); override;
    function GetFullRegistryKey: String;
  published
    property Tag;
    property AutoLoad: Boolean read FAutoLoad write FAutoLoad default True;
    property AutoSave: Boolean read FAutoSave write FAutoSave default True;
    property ProfileName: String read FProfileName write FProfileName stored
      IsStoredProfileName;
    property RegistryKey: String read FRegistryKey write FRegistryKey;
    property SectionName: String read FSectionName write FSectionName;
    property OnLoadSettings: TNotifyEvent read FOnLoadSettings write
      FOnLoadSettings;
    property OnSaveSettings: TNotifyEvent read FOnSaveSettings write
      FOnSaveSettings;
    property OnLoadSettingsError: TNotifyEvent read FOnLoadSettingsError write
      FOnLoadSettingsError;
    property OnSaveSettingsError: TNotifyEvent read FOnSaveSettingsError write
      FOnSaveSettingsError;
  end;

var
  CustomizationBaseRegistryKey: String = 'Software\';

implementation

resourcestring
  SDefaultProfileName = '  ';

{ TovCustomization }

constructor TovCustomization.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAutoLoad := True;
  FAutoSave := True;
  FProfileName := SDefaultProfileName;
end;

destructor TovCustomization.Destroy;
begin
  if Not (csDesigning in ComponentState) then CheckSaveSettings;
  inherited Destroy;
end;

{***************************************************************************** }
{                                                             }
{ **************************************************************************** }

function TovCustomization.IsStoredProfileName: Boolean;
begin
  Result := ProfileName <> SDefaultProfileName;
end;

procedure TovCustomization.SetSubjectComponent(Value: TComponent);
begin
  if FSubjectComponent <> Value then
  begin
    FSubjectComponent := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

{{
   ,    .

 RegistryKey       
.    ,    
CustomizationBaseRegistryKey, SectionName  ProfileName.  
,  CustomizationBaseRegistryKey    
    , SectionName   
  ,  ProfileName -      
      .
}
function TovCustomization.GetFullRegistryKey: String;
begin
  if RegistryKey <> '' then
    Result := RegistryKey +'\'+ProfileName
  else
    Result := CustomizationBaseRegistryKey + SectionName +'\'+ProfileName
end;

{***************************************************************************** }
{                                                   }
{ **************************************************************************** }

procedure TovCustomization.LoadSettings;
begin
  if Assigned(FOnLoadSettings) then FOnLoadSettings(Self);
end;

procedure TovCustomization.SaveSettings;
begin
  if Assigned(FOnSaveSettings) then FOnSaveSettings(Self);
end;

procedure TovCustomization.Loaded;
begin
  inherited;
  if AutoLoad and Not (csDesigning in ComponentState) then
  try
    LoadSettings;
  except
    if Assigned(FOnLoadSettingsError) then FOnLoadSettingsError(Self);
  end;
end;

procedure TovCustomization.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);

  //         
  //      SubjectComponent
  if Not (csDesigning in ComponentState) and (csDestroying in ComponentState) and
    (Operation <> opInsert) //            
    then CheckSaveSettings;

  if Operation = opRemove then
    if AComponent = SubjectComponent then SubjectComponent := nil
end;

procedure TovCustomization.CheckSaveSettings;
begin
  if AutoSave and Not FSettingsSaved then
  try
    SaveSettings;
    FSettingsSaved := True;
  except
    if Assigned(FOnSaveSettingsError) then FOnSaveSettingsError(Self);
  end;
end;

end.
