unit ovDBBackup;

interface

uses Classes, ovComponents, ovDBConnections, Contnrs;

type
  TovDatabaseBackupInfo = class
  private
    FBackupName: String;
    FBackupDescription: String;
    FBackupDate: TDateTime;
    FConnectionName: String;
    FConnectionParams: TStrings;
  public
    constructor Create;
    destructor Destroy; override;
    property BackupName: String read FBackupName;
    property BackupDescription: String read FBackupDescription;
    property BackupDate: TDateTime read FBackupDate;
    property ConnectionName: String read FConnectionName;
    property ConnectionParams: TStrings read FConnectionParams;
  end;

  TovModelDatabaseBackup = class(TovModelClient)
  private
    FBackupInfoFileName: String;
    function DatabaseConnections: TovDatabaseConnections;
    procedure SaveBackupInfo(ConnectionName, BackupName: String; ConnectionParams:
      TStrings; BackupDescription: String; BackupDate: TDateTime);
    function GetBackupCount(ConnectionName: String): Integer;
  protected
  public
    procedure BackupModelDatabase(ConnectionName, BackupName, BackupDescription:
      String);
    procedure GetBackupConnectionParams(AConnectionName, ABackupName: String;
      AConnectionParams: TStrings);
    procedure GetBackupList(ConnectionName: String; BackupList: TObjectList);
    procedure RestoreModelDatabase(ConnectionName, BackupName: String);
    function ShowBackupDialog(ConnectionName: String = ''): Boolean;
    function ShowRestoreDialog(ConnectionName: String = ''): Boolean;
  published
    property BackupInfoFileName: String read FBackupInfoFileName write
      FBackupInfoFileName;
  end;

implementation

uses Windows, Forms, SysUtils, ovBackupDatabaseDlg, ovRestoreDatabaseDlg;

{ TovDatabaseBackupInfo }

constructor TovDatabaseBackupInfo.Create;
begin
  inherited Create;
  FConnectionParams := TStringList.Create;
end;

destructor TovDatabaseBackupInfo.Destroy;
begin
  FreeAndNil(FConnectionParams);
  inherited Destroy;
end;

{ TovModelDatabaseBackup }

function TovModelDatabaseBackup.DatabaseConnections: TovDatabaseConnections;
begin
  ModelNeeded;
  if Not Assigned(Model.DatabaseConnection) then
    raise Exception.Create('   DatabaseConnection  ' + Model.Name);
  Result := Model.DatabaseConnection as TovDatabaseConnections;
end;

{***************************************************************************** }
{                                                    }
{ **************************************************************************** }

procedure TovModelDatabaseBackup.BackupModelDatabase(ConnectionName,
  BackupName, BackupDescription: String);
var
  SourceDatabaseAdapter: TovCustomDatabaseAdapter;
  TargetDatabaseAdapter: TovCustomDatabaseAdapter;
  BackupConnectionParams: TStrings;
begin
  BackupConnectionParams := TStringList.Create;
  try

    SourceDatabaseAdapter := nil;
    TargetDatabaseAdapter := nil;
    try
      SourceDatabaseAdapter := DatabaseConnections.OpenConnection(ConnectionName, Model);
      SourceDatabaseAdapter.CreateDatabase('DatabaseBackup' +
        IntToStr(GetBackupCount(ConnectionName)+1),
        BackupConnectionParams);
      TargetDatabaseAdapter := DatabaseConnections.OpenConnection(ConnectionName, Model, BackupConnectionParams);

      Model.CopyDatabase(SourceDatabaseAdapter, TargetDatabaseAdapter);

      SaveBackupInfo(ConnectionName, BackupName, BackupConnectionParams,
        BackupDescription, Now {  })

    finally
      TargetDatabaseAdapter.Free;
      SourceDatabaseAdapter.Free;
    end;

  finally
    BackupConnectionParams.Free;
  end;
end;

procedure TovModelDatabaseBackup.RestoreModelDatabase(ConnectionName,
  BackupName: String);
var
  SourceDatabaseAdapter: TovCustomDatabaseAdapter;
  TargetDatabaseAdapter: TovCustomDatabaseAdapter;
  BackupConnectionParams: TStrings;
begin
  BackupConnectionParams := TStringList.Create;
  try

    SourceDatabaseAdapter := nil;
    TargetDatabaseAdapter := nil;
    try
      GetBackupConnectionParams(ConnectionName, BackupName, BackupConnectionParams);
      SourceDatabaseAdapter := DatabaseConnections.OpenConnection(ConnectionName, Model, BackupConnectionParams);
      TargetDatabaseAdapter := DatabaseConnections.OpenConnection(ConnectionName, Model);

      Model.CopyDatabase(SourceDatabaseAdapter, TargetDatabaseAdapter);

    finally
      TargetDatabaseAdapter.Free;
      SourceDatabaseAdapter.Free;
    end;

  finally
    BackupConnectionParams.Free;
  end;
end;

{{
         
 .     ,    
  .
}
function TovModelDatabaseBackup.ShowBackupDialog(ConnectionName: String = ''):
  Boolean;
var
  BackupName, BackupDescription: String;
begin
  Result := False;
  if ConnectionName = '' then
    ConnectionName := DatabaseConnections.DefaultConnection.ConnectionName;
  if ShowDatabaseBackupDlg(BackupName, BackupDescription) then
  begin
    BackupModelDatabase(ConnectionName, BackupName, BackupDescription);
    Application.MessageBox('    ',
      ' ', MB_ICONINFORMATION or MB_OK);
    Result := True;
  end
end;

{{
        
    .     ,
       .
}
function TovModelDatabaseBackup.ShowRestoreDialog(ConnectionName: String = ''):
  Boolean;
var
  BackupName: String;
begin
  Result := False;
  if ConnectionName = '' then
    ConnectionName := DatabaseConnections.DefaultConnection.ConnectionName;
  if ShowDatabaseRestoreDlg(Self, ConnectionName, BackupName) then
  begin
    RestoreModelDatabase(ConnectionName, BackupName);
    Application.MessageBox('    ',
      ' ', MB_ICONINFORMATION or MB_OK);
    Result := True;
  end
end;

{***************************************************************************** }
{                          }
{ **************************************************************************** }

{{
  BackupInfoFileName      
ConnectionName.
}
procedure TovModelDatabaseBackup.SaveBackupInfo(ConnectionName, BackupName: String;
  ConnectionParams: TStrings; BackupDescription: String; BackupDate: TDateTime);

  procedure WriteParameter(var AFile: TextFile; AParamName, AParamValue: String);
  begin
    Writeln(AFile, AParamName + '=' + AParamValue);
  end;

  procedure WriteBackupInfo(var AFile: TextFile);
  var
    I: Integer;
  begin
    Writeln(AFile, '[Backup]');
    WriteParameter(AFile, 'ConnectionName', ConnectionName);
    WriteParameter(AFile, 'BackupName', BackupName);
    WriteParameter(AFile, 'BackupDate', DateTimeToStr(BackupDate));
    WriteParameter(AFile, 'BackupDescription', BackupDescription);
    Writeln(AFile, '[Connection]');
    for I := 0 to ConnectionParams.Count - 1 do
      Writeln(AFile, ConnectionParams[I]);
    Writeln(AFile);
  end;

var
  AFile: TextFile;
begin
  AssignFile(AFile, BackupInfoFileName);
  if FileExists(BackupInfoFileName) then Append(AFile) else Rewrite(AFile);
  try
     WriteBackupInfo(AFile);
  finally
    CloseFile(AFile)
  end;
end;

{{
  BackupInfoFileName      
ConnectionName.
}
procedure TovModelDatabaseBackup.GetBackupList(ConnectionName: String;
  BackupList: TObjectList);

  {{
     AFile  ,    
      AParamName.    
  }
  function ReadParameter(var AFile: TextFile; AParamName: String): String;
  var
    Line: String;
    P: Integer;
  begin
    ReadLn(AFile, Line);
    P := Pos('=', Line);
    if P = 0 then raise Exception.Create('   "="');
    if Copy(Line, 1, P-1) = AParamName then
      Result := Copy(Line, P+1, Length(Line))
    else
      raise Exception.Create('     ' + AParamName)
  end;

  function ReadBackupInfo(var AFile: TextFile): TovDatabaseBackupInfo;

    procedure ReadBackupProperties;
    var
      Line: String;
    begin
      while Not Eof(AFile) do
      begin
        ReadLn(AFile, Line);
        if Line = '[Backup]' then Break;
      end;
      if Line <> '[Backup]' then
        raise Exception.Create('  ');

      Result.FConnectionName := ReadParameter(AFile, 'ConnectionName');
      Result.FBackupName := ReadParameter(AFile, 'BackupName');
      Result.FBackupDate := StrToDateTime(ReadParameter(AFile, 'BackupDate'));
      Result.FBackupDescription := ReadParameter(AFile, 'BackupDescription');
      while Not Eof(AFile) do
      begin
        ReadLn(AFile, Line);
        if Line = '[Connection]' then Break;
        Result.FBackupDescription := Result.FBackupDescription + #13#10 + Line;
      end;
      if Line <> '[Connection]' then
        raise Exception.Create('  ');
    end;

    procedure ReadConnectionParams;
    var
      Line: String;
    begin
      with Result.ConnectionParams do
      while Not Eof(AFile) do
      begin
        ReadLn(AFile, Line);
        if Line = '' then Break; //       
        Add(Line);
      end;
    end;

  begin
    Result := TovDatabaseBackupInfo.Create;
    try
      ReadBackupProperties;
      ReadConnectionParams;
    except
      Result.Free; raise
    end;
  end;

var
  AFile: TextFile;
  ABackupInfo: TovDatabaseBackupInfo;
begin
  if FileExists(BackupInfoFileName) then
  begin
    AssignFile(AFile, BackupInfoFileName);
    Reset(AFile);
    try
     while Not Eof(AFile) do
     begin
       ABackupInfo := ReadBackupInfo(AFile);
       if ABackupInfo.ConnectionName = ConnectionName then
         BackupList.Add(ABackupInfo)
       else
         ABackupInfo.Free;
     end;
    finally
      CloseFile(AFile)
    end;
  end;
end;

{{
  BackupInfoFileName     
ConnectionName.
}
function TovModelDatabaseBackup.GetBackupCount(ConnectionName: String): Integer;
var
  BackupList: TObjectList;
begin
  BackupList := TObjectList.Create;
  try
    GetBackupList(ConnectionName, BackupList);
    Result := BackupList.Count;
  finally
    BackupList.Free;
  end;
end;

procedure TovModelDatabaseBackup.GetBackupConnectionParams(AConnectionName,
  ABackupName: String; AConnectionParams: TStrings);
var
  I: Integer;
  BackupList: TObjectList;
begin
  BackupList := TObjectList.Create;
  try
    GetBackupList(AConnectionName, BackupList);
    for I := 0 to BackupList.Count - 1 do
    with TovDatabaseBackupInfo(BackupList[I]) do
      if (ConnectionName = AConnectionName) and (BackupName = ABackupName) then
      begin
        AConnectionParams.Assign(ConnectionParams);
        Exit;
      end;
    raise Exception.CreateFmt('   %s   %s  ',
      [ABackupName, AConnectionName]);
  finally
    BackupList.Free;
  end;
end;

end.
