unit ovDBConnections;

interface

uses
  SysUtils, Classes, ovComponents, Contnrs, IniFiles, ovModel;

type
  TovDatabaseConnection = class;

  TovDatabaseConnectionList = class (TObjectList)
  private
    function GetItems(Index: Integer): TovDatabaseConnection;
    procedure SetDefaultConnection(Connection: TovDatabaseConnection);
  protected
    function Find(ConnectionName: String): TovDatabaseConnection;
  public
    function Add(ConnectionName: String): TovDatabaseConnection;
    property Items[Index: Integer]: TovDatabaseConnection read GetItems; default;
  end;

  TovDatabaseConnection = class
  private
    FAdapterClassName: String;
    FAdapterPackageFileName: String;
    FConnectionName: String;
    FConnectionParams: TStrings;
    FIsDefault: Boolean;
    FOwner: TovDatabaseConnectionList;
    procedure SetConnectionName(const Value: String);
    procedure SetIsDefault(const Value: Boolean);
  public
    constructor Create(AOwner: TovDatabaseConnectionList);
    destructor Destroy; override;
    property AdapterClassName: String read FAdapterClassName write
      FAdapterClassName;
    property AdapterPackageFileName: String read FAdapterPackageFileName write
      FAdapterPackageFileName;
    property ConnectionName: String read FConnectionName write SetConnectionName;
    property ConnectionParams: TStrings read FConnectionParams;
    property IsDefault: Boolean read FIsDefault write SetIsDefault;
    property Owner: TovDatabaseConnectionList read FOwner;
  end;

  TovConnectionStorage = class(TObject)
  private
    FFileName: String;
    function GetConnectionNameFromSectionName(SectionName: String): String;
    function GetDefaultConnectionName: String;
    procedure ReadConnectionFromSection(Ini: TIniFile; SectionName: String;
      AConnection: TovDatabaseConnection);
  public
    function GetConnection(AConnections: TovDatabaseConnectionList; ConnectionName:
      String): TovDatabaseConnection;
    procedure LoadConnections(AConnections: TovDatabaseConnectionList);
    procedure SaveConnections(AConnections: TovDatabaseConnectionList);
    property DefaultConnectionName: String read GetDefaultConnectionName;
  published
    property FileName: String read FFileName write FFileName;
  end;

  TOpenConnectionEvent = procedure(Sender: TObject; ConnectionName: String;
    DatabaseAdapter: TovCustomDatabaseAdapter; AModel: TovModel) of object;
  TCloseConnectionEvent = TNotifyEvent;

  TovDatabaseConnections = class(TovCustomDatabaseConnections)
  private
    FConnections: TovDatabaseConnectionList;
    FConnectionStorage: TovConnectionStorage;
    FOnOpenConnection: TOpenConnectionEvent;
    FActiveDatabaseAdapter: TovCustomDatabaseAdapter;
    FOnCloseConnection: TCloseConnectionEvent;
    function GetConnectionFileName: String;
    function GetDatabaseAdapter(AdapterPackageFileName, AdapterClassName: String):
      TovCustomDatabaseAdapter;
    function GetConnectionCount: Integer;
    function GetConnections(Index: Integer): TovDatabaseConnection;
    function GetDefaultConnection: TovDatabaseConnection;
    procedure SetConnectionFileName(const Value: String);
  protected
    function GetConnection(ConnectionName: String): TovDatabaseConnection;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddConnection(ConnectionName: String): TovDatabaseConnection;
    function AddConnectionFromPackage(ConnectionName, AdapterPackageFileName:
      String): TovDatabaseConnection;
    procedure DeleteConnection(ConnectionName: String);
    procedure LoadAllConnections;
    function OpenConnection(AConnectionName: String; AModel: TovModel;
      AConnectionParams: TStrings = nil; OpenDatabaseImmediately: Boolean = True):
      TovCustomDatabaseAdapter;
    procedure SaveAllConnections;

    //   TovCustomDatabaseConnections
    function OpenDefaultConnection(AModel: TovModel): TovCustomDatabaseAdapter;
      override;
    procedure CloseActiveConnection; override;
  public
    property ConnectionCount: Integer read GetConnectionCount;
    property Connections[Index: Integer]: TovDatabaseConnection read GetConnections;
    property DefaultConnection: TovDatabaseConnection read GetDefaultConnection;
  published
    property ConnectionFileName: String read GetConnectionFileName write
      SetConnectionFileName;
    property ActiveDatabaseAdapter: TovCustomDatabaseAdapter read
      FActiveDatabaseAdapter;
    {{
            .
    }
    property OnOpenConnection: TOpenConnectionEvent read FOnOpenConnection write
      FOnOpenConnection;
    {{
           
    .
    }
    property OnCloseConnection: TCloseConnectionEvent read FOnCloseConnection write
      FOnCloseConnection;
  end;

  EDuplicateConnectionName = class (Exception)
  end;

  ENoDefaultConnection = class (Exception)
  end;

implementation

resourcestring
  SConnectionSectionNamePrefix = 'Connection-';
  idPackageName = 'PackageName';
  idAdapterName = 'AdapterName';
  idIsDefault = 'IsDefault';

{ TovDatabaseConnection }

constructor TovDatabaseConnection.Create(AOwner: TovDatabaseConnectionList);
begin
  inherited Create;
  FOwner := AOwner;
  FConnectionParams := TStringList.Create;
end;

destructor TovDatabaseConnection.Destroy;
begin
  FreeAndNil(FConnectionParams);
  inherited Destroy;
end;

procedure TovDatabaseConnection.SetConnectionName(const Value: String);
begin
  if FConnectionName <> Value then
  begin
    if Owner.Find(Value) <> nil then
      raise EDuplicateConnectionName.Create('    ');
    FConnectionName := Value;
  end;
end;

procedure TovDatabaseConnection.SetIsDefault(const Value: Boolean);
begin
  if (FIsDefault <> Value) and Value then Owner.SetDefaultConnection(Self);
end;

{ TovDatabaseConnectionList }

function TovDatabaseConnectionList.GetItems(Index: Integer):
  TovDatabaseConnection;
begin
  Result := TovDatabaseConnection(inherited Items[Index])
end;

function TovDatabaseConnectionList.Add(ConnectionName: String):
  TovDatabaseConnection;
begin
  if Find(ConnectionName) <> nil then
    raise EDuplicateConnectionName.Create('    ');
  Result := TovDatabaseConnection.Create(Self);
  Result.FConnectionName := ConnectionName;
  inherited Add(Result);
end;

function TovDatabaseConnectionList.Find(ConnectionName: String):
  TovDatabaseConnection;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    if Items[I].ConnectionName = ConnectionName
    then
    begin
      Result := Items[I];
      Exit;
    end;
  Result := nil;
end;

{{
   .      
.
}
procedure TovDatabaseConnectionList.SetDefaultConnection(Connection:
  TovDatabaseConnection);
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    Items[I].FIsDefault := False;
  Connection.FIsDefault := True;
end;

{ TovConnectionStorage }

{{
   AConnections     
 ConnectionName.
}
function TovConnectionStorage.GetConnection(AConnections:
  TovDatabaseConnectionList; ConnectionName: String): TovDatabaseConnection;
var
  Ini: TIniFile;
  SectionName: String;
begin
  Result := AConnections.Add(ConnectionName);
  Ini := TIniFile.Create(FileName);
  try
    SectionName := SConnectionSectionNamePrefix + ConnectionName;
    ReadConnectionFromSection(Ini, SectionName, Result);
  finally
    Ini.Free
  end;
end;

{{
      AConnections.
}
procedure TovConnectionStorage.LoadConnections(AConnections:
  TovDatabaseConnectionList);
var
  Ini: TIniFile;
  Sections: TStrings;
  Connection: TovDatabaseConnection;
  I: Integer;
begin
  Ini := TIniFile.Create(FileName);
  Sections := TStringList.Create;
  try
    Ini.ReadSections(Sections);
    AConnections.Clear;
    for I := 0 to Sections.Count - 1 do
    begin
      Connection := AConnections.Add(GetConnectionNameFromSectionName(Sections[I]));
      ReadConnectionFromSection(Ini, Sections[I], Connection);
    end;
  finally
    Sections.Free;
    Ini.Free
  end;
end;

{{
      AConnections.
}
procedure TovConnectionStorage.SaveConnections(AConnections:
  TovDatabaseConnectionList);
var
  SectionName: String;
  I, J: Integer;
  Ini: TIniFile;
begin
  DeleteFile(FileName); //    ,        
  Ini := TIniFile.Create(FileName);
  try
    for I := 0 to AConnections.Count - 1 do
    with AConnections[I] do
    begin
      SectionName := SConnectionSectionNamePrefix + ConnectionName;
      Ini.WriteString(SectionName, idPackageName, AdapterPackageFileName);
      Ini.WriteString(SectionName, idAdapterName, AdapterClassName);
      if IsDefault then
        Ini.WriteString(SectionName, idIsDefault, 'True');

      with ConnectionParams do
        for J := 0 to Count - 1 do
          Ini.WriteString(SectionName, Names[J], Values[Names[J]]);
    end;
  finally
    Ini.Free
  end;
end;

{{
      .
}
function TovConnectionStorage.GetDefaultConnectionName: String;
var
  Ini: TIniFile;
  Sections: TStrings;
  I: Integer;
begin
  Ini := TIniFile.Create(FileName);
  try
    Sections := TStringList.Create;
    try
      Result := '';
      Ini.ReadSections(Sections);
      for I := 0 to Sections.Count - 1 do
      if Ini.ReadString(Sections[I], idIsDefault, '') = 'True' then
        Result := GetConnectionNameFromSectionName(Sections[I])
    finally
      Sections.Free;
    end;
  finally
    Ini.Free
  end;
end;

{***************************************************************************** }
{                                                         }
{ **************************************************************************** }

function TovConnectionStorage.GetConnectionNameFromSectionName(SectionName:
  String): String;
begin
  Result := Copy(SectionName, Length(SConnectionSectionNamePrefix) + 1,
    Length(SectionName));
end;

procedure TovConnectionStorage.ReadConnectionFromSection(Ini: TIniFile;
  SectionName: String; AConnection: TovDatabaseConnection);
begin
  with AConnection do
  begin
    AdapterPackageFileName := Ini.ReadString(SectionName, idPackageName, '');
    AdapterClassName := Ini.ReadString(SectionName, idAdapterName, '');
    FIsDefault := StrToBool(Ini.ReadString(SectionName, idIsDefault, 'False'));

    Ini.ReadSectionValues(SectionName, ConnectionParams);
    ConnectionParams.Values[idPackageName] := '';
    ConnectionParams.Values[idAdapterName] := '';
    ConnectionParams.Values[idIsDefault] := '';
  end;
end;

{ TovDatabaseConnections }

constructor TovDatabaseConnections.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FConnections := TovDatabaseConnectionList.Create;
  FConnectionStorage := TovConnectionStorage.Create;
end;

destructor TovDatabaseConnections.Destroy;
begin
  FreeAndNil(FConnectionStorage);
  FreeAndNil(FConnections);
  inherited Destroy;
end;

function TovDatabaseConnections.GetConnectionFileName: String;
begin
  Result := FConnectionStorage.FileName;
end;

procedure TovDatabaseConnections.SetConnectionFileName(const Value: String);
begin
  FConnectionStorage.FileName := Value;
end;

{***************************************************************************** }
{                                                   }
{ **************************************************************************** }

function TovDatabaseConnections.GetConnectionCount: Integer;
begin
  Result := FConnections.Count
end;

function TovDatabaseConnections.GetConnections(Index: Integer):
  TovDatabaseConnection;
begin
  Result := FConnections[Index]
end;

{{
     .

:
-       /,  
  ,     
}
function TovDatabaseConnections.GetConnection(ConnectionName: String):
  TovDatabaseConnection;
begin
  Result := FConnections.Find(ConnectionName);
  if Result = nil then
    Result := FConnectionStorage.GetConnection(FConnections, ConnectionName)
end;

{{
       .
}
procedure TovDatabaseConnections.LoadAllConnections;
begin
  FConnectionStorage.LoadConnections(FConnections)
end;

{{
  /   .
}
procedure TovDatabaseConnections.SaveAllConnections;
begin
  FConnectionStorage.SaveConnections(FConnections)
end;

function TovDatabaseConnections.AddConnection(ConnectionName: String):
  TovDatabaseConnection;
begin
  Result := FConnections.Add(ConnectionName);
end;

procedure TovDatabaseConnections.DeleteConnection(ConnectionName: String);
var
  Connection: TovDatabaseConnection;
begin
  Connection := FConnections.Find(ConnectionName);
  if Connection <> nil then FConnections.Remove(Connection);
end;

function TovDatabaseConnections.GetDefaultConnection: TovDatabaseConnection;
var
  ConnectionName: String;
begin
  ConnectionName := FConnectionStorage.DefaultConnectionName;
  if ConnectionName = '' then
    Result := nil
  else
    Result := FConnections.Find(ConnectionName);
end;

{{
     ,   
AdapterPackageFileName.
}
function TovDatabaseConnections.AddConnectionFromPackage(ConnectionName,
  AdapterPackageFileName: String): TovDatabaseConnection;
var
  Handle: HModule;
  Adapters: TStrings;
begin
  Result := nil;
  Adapters := TStringList.Create;
  try
    Handle := LoadPackage(AdapterPackageFileName);
    TovCustomDatabaseAdapter.GetAdaptersFromModule(Handle, Adapters);
    if Adapters.Count = 0 then
      raise Exception.CreateFmt(' %s     ', [ExtractFileName(AdapterPackageFileName)]);

    Result := AddConnection(ConnectionName);
    Result.AdapterPackageFileName := AdapterPackageFileName;
    //       ,  , 
    //        
    Result.AdapterClassName := Adapters[0];
  finally
    Adapters.Free;
  end;
end;

{***************************************************************************** }
{                       /                            }
{ **************************************************************************** }

{{
         ,
     ,    .
     ,   
 ENoDefaultConnection.
}
function TovDatabaseConnections.OpenDefaultConnection(AModel: TovModel):
  TovCustomDatabaseAdapter;
var
  ConnectionName: String;
begin
  if Not Assigned(ActiveDatabaseAdapter) then
  begin
    ConnectionName := FConnectionStorage.DefaultConnectionName;
    if ConnectionName = '' then
      raise ENoDefaultConnection.Create('    ');
    Result := OpenConnection(ConnectionName, AModel);
    FActiveDatabaseAdapter := Result;
  end else
    Result := FActiveDatabaseAdapter;
end;

procedure TovDatabaseConnections.CloseActiveConnection;
begin
  if Assigned(ActiveDatabaseAdapter) then
  begin
    if Assigned(FOnCloseConnection) then FOnCloseConnection(Self);
    FreeAndNil(FActiveDatabaseAdapter);
  end;
end;

{{
     .  
AConnectionParams  ,   ,   
,    .    
   OnOpenConnection.
}
function TovDatabaseConnections.OpenConnection(AConnectionName: String; AModel:
  TovModel; AConnectionParams: TStrings = nil; OpenDatabaseImmediately: Boolean
  = True): TovCustomDatabaseAdapter;
begin
  with GetConnection(AConnectionName) do
  begin
    Result := GetDatabaseAdapter(AdapterPackageFileName, AdapterClassName);
    try
      if AConnectionParams <> nil then
        Result.SetupConnection(AConnectionParams)
      else
        Result.SetupConnection(ConnectionParams);
      if OpenDatabaseImmediately then Result.OpenDatabase;

      if Assigned(FOnOpenConnection) then
        FOnOpenConnection(Self, ConnectionName, Result, AModel);
    except
      Result.Free; raise;
    end;
  end;
end;

function TovDatabaseConnections.GetDatabaseAdapter(AdapterPackageFileName,
  AdapterClassName: String): TovCustomDatabaseAdapter;
begin
  LoadPackage(AdapterPackageFileName);
  Result := TovCustomDatabaseAdapter.GetAdapter(AdapterClassName).Create(nil);
end;

end.

