unit ovDBSynchronization;

interface

uses DB, ovComponents, Classes, ovTypes, ovXMLImportExport, ovIObjects,
  SysUtils, XMLIntf;

type
  TDBId = String;

  {{
            
   .     SaveToFile 
  SaveToStream.       
    LoadFromFile  LoadFromStream.
             
      LookupTableName.
  }
  TovDBSynchronization = class(TovModelClient)
  private
    FLocalDBId: TDBId;
    FRemoteDBId: TDBId;
    FLookupTableName: String;
    FObjects: IovObjectList;
    FAutoGenerateLookupTable: Boolean;
    FOnProgress: TProgressEvent;
    FOnAfterObjectImport: TAfterObjectImportEvent;
    FOnAfterObjectExport: TAfterObjectExportEvent;
    FOnObjectExport: TObjectExportEvent;
    FOnChildObjectImport: TChildObjectImportEvent;
    procedure DoAfterObjectExport(Sender: TObject; AObject: IovObject; AObjectNode:
      IXMLNode);
    procedure DoBeforeObjectImport(Sender: TObject; AObjectClassName: String;
      AObjectNode: IXMLNode; var AObjectId: TOID);
    procedure CheckModel;
    function IsLookupTableNameStored: Boolean;
    procedure DoRelationImport(Sender: TObject; AObject: IovObject;
      ARelation: String; AChildObjectNode: IXMLNode; var Skip: Boolean);
    procedure DoAfterExport(XMLImportExport: TovXMLImportExport);
    procedure DoBeforeImport(XMLImportExport: TovXMLImportExport);
    procedure DoAttributesImport(Sender: TObject; AObject: IovObject;
      AObjectNode: IXMLNode; var Skip: Boolean);
    procedure DoChildObjectImport(Sender: TObject; AObject: IovObject;
      ARelation: String; AChildObjectNode: IXMLNode; var AObjectId: TOID);
  protected
    function GetRemoteObjectId(AClassId: Integer; ALocalObjectId: TOID): TOID;
    procedure UpdateRemoteObjectId(AClassId: Integer; ALocalObjectId,
      ARemoteObjectId: TOID);
    function GetLocalObjectId(AClassId: Integer; ARemoteObjectId: TOID): TOID;
  public
    procedure SaveToFile(FileName: String);
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromFile(FileName: String);
    procedure LoadFromStream(Stream: TStream);
    procedure GenerateLookupTable;
    constructor Create(AOwner: TComponent); override;
    function IsRemoteObject(AObject: IovObject; ADBId: TDBId): Boolean;
    property Objects: IovObjectList read FObjects;
  published
    property LocalDBId: TDBId read FLocalDBId write FLocalDBId;
    property RemoteDBId: TDBId read FRemoteDBId write FRemoteDBId;
    property LookupTableName: String read FLookupTableName write FLookupTableName
      stored IsLookupTableNameStored;
    property AutoGenerateLookupTable: Boolean read FAutoGenerateLookupTable write
      FAutoGenerateLookupTable default True;

    {  }
    property OnAfterObjectExport: TAfterObjectExportEvent read FOnAfterObjectExport
      write FOnAfterObjectExport;
    property OnAfterObjectImport: TAfterObjectImportEvent read FOnAfterObjectImport
      write FOnAfterObjectImport;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
    property OnObjectExport: TObjectExportEvent read FOnObjectExport write
      FOnObjectExport;
    property OnChildObjectImport: TChildObjectImportEvent read FOnChildObjectImport
      write FOnChildObjectImport;
  end;

implementation

uses Variants, ovObjects;

type

  TXMLImportExportFriend = class(TovXMLImportExport)
  end;

resourcestring
  {      }
  DefaultLookupTableName = 'SYNCDB_TABLE';

  {     }
  ltRecordIdFieldName = 'RecID';
  ltDBIdFieldName = 'DatabaseID';
  ltClassIdFieldName = 'ClassID';
  ltLocalObjectIdFieldName = 'LocalOID';
  ltRemoteObjectIdFieldName = 'RemoteOID';

{ TovDBSynchronization }

constructor TovDBSynchronization.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAutoGenerateLookupTable := True;
  LookupTableName := DefaultLookupTableName;
  FObjects := NewObjectList;
end;

function TovDBSynchronization.IsLookupTableNameStored: Boolean;
begin
  Result := LookupTableName <> DefaultLookupTableName;
end;

procedure TovDBSynchronization.CheckModel;
begin
  if LocalDBId = '' then raise Exception.Create('     ');
  if RemoteDBId = '' then raise Exception.Create('     ');

  ModelNeeded;
  Model.Open;

  if AutoGenerateLookupTable and Not Model.DatabaseAdapter.TableExists(LookupTableName) then
    GenerateLookupTable;
end;

{***************************************************************************** }
{                                                                 }
{ **************************************************************************** }

procedure TovDBSynchronization.SaveToFile(FileName: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream)
  finally
    Stream.Free;
  end;
end;

procedure TovDBSynchronization.SaveToStream(Stream: TStream);
var
  XMLExport: TovXMLExport;
begin
  CheckModel;
  XMLExport := TovXMLExport.Create(Self);
  try
    {   }
    XMLExport.RootObjectsTagName := 'Document';
    XMLExport.Model := Model;
    XMLExport.OnAfterObjectExport := DoAfterObjectExport;
    XMLExport.OnObjectExport := OnObjectExport;
    XMLExport.OnProgress := OnProgress;
    XMLExport.Objects.Assign(Self.Objects);

    {  XML-      }
    XMLExport.Execute;
    DoAfterExport(XMLExport);
    XMLExport.SaveToStream(Stream);
  finally
    XMLExport.Free;
  end;
end;

procedure TovDBSynchronization.DoAfterObjectExport(Sender: TObject; AObject:
  IovObject; AObjectNode: IXMLNode);
var
  RemoteId: TOID;
begin
  {   id  (    RemoteDBId) }
  RemoteId := GetRemoteObjectId(AObject.ClassId, AObject.ObjectId);
  {     }
  AObjectNode.Attributes['LocalOID'] := AObject.ObjectId;
  AObjectNode.Attributes['RemoteOID'] := RemoteId;

  if Assigned(FOnAfterObjectExport) then FOnAfterObjectExport(Self, AObject, AObjectNode);
end;

procedure TovDBSynchronization.DoAfterExport(XMLImportExport:
  TovXMLImportExport);
var
  Node: IXMLNode;
begin
  Node := TXMLImportExportFriend(XMLImportExport).XMLDocument.DocumentElement;
  Node.Attributes['SourceDatabase'] := LocalDBId;
  Node.Attributes['TargetDatabase'] := RemoteDBId;
end;

{***************************************************************************** }
{                                                                  }
{ **************************************************************************** }

procedure TovDBSynchronization.LoadFromFile(FileName: String);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream)
  finally
    Stream.Free;
  end;
end;

procedure TovDBSynchronization.LoadFromStream(Stream: TStream);
var
  XMLImport: TovXMLImport;
begin
  CheckModel;
  XMLImport := TovXMLImport.Create(Self);
  try
    {   }
    XMLImport.RootObjectsTagName := 'Document';
    XMLImport.Model := Model;
    XMLImport.OnBeforeObjectImport := DoBeforeObjectImport;
    XMLImport.OnAfterObjectImport := self.OnAfterObjectImport;
    XMLImport.OnAttributesImport := DoAttributesImport;
    XMLImport.OnRelationImport := DoRelationImport;
    XMLImport.OnChildObjectImport := DoChildObjectImport;
    XMLImport.OnProgress := OnProgress;

    {  XML-       }
    XMLImport.LoadFromStream(Stream);
    DoBeforeImport(XMLImport);
    XMLImport.Execute;
  finally
    XMLImport.Free;
  end;
end;

procedure TovDBSynchronization.DoBeforeImport(XMLImportExport:
  TovXMLImportExport);
var
  Node: IXMLNode;
begin
  Node := TXMLImportExportFriend(XMLImportExport).XMLDocument.DocumentElement;
  { TODO :       
         }
  RemoteDBId := Node.Attributes['SourceDatabase'];
  if Node.Attributes['TargetDatabase'] <> LocalDBId then
    raise Exception.Create('        ');
end;

procedure TovDBSynchronization.DoBeforeObjectImport(Sender: TObject;
  AObjectClassName: String; AObjectNode: IXMLNode; var AObjectId: TOID);
var
  LocalId, RemoteId: TOID;
  ClassId: Integer;
begin
  {  LocalId, RemoteId    ;   
    LocalOID      ,   
       RemoteId  ! }
  RemoteId := AObjectNode.Attributes['LocalOID'];
  LocalId := AObjectNode.Attributes['RemoteOID'];
  AObjectNode.AttributeNodes.Delete('LocalOID');
  AObjectNode.AttributeNodes.Delete('RemoteOID');

  ClassId := Model.DomainModel.GetObject(AObjectClassName).Id;
  if LocalId = 0 then
    {          ( 
        LocalId,          }
    LocalId := GetLocalObjectId(ClassId, RemoteId);

  if LocalId = 0 then //        ,    
  begin
    LocalId := Model.ObjectSpace.CreateObject(AObjectClassName).ObjectId;
    UpdateRemoteObjectId(ClassId, LocalId, RemoteId);
  end
  else
  begin
    {      ,     }
    UpdateRemoteObjectId(ClassId, LocalId, RemoteId);
  end;
  AObjectId := LocalId;
end;

{{
   .

        
  () (      DoBeforeObjectImport).
}
procedure TovDBSynchronization.DoAttributesImport(Sender: TObject; AObject: IovObject;
  AObjectNode: IXMLNode; var Skip: Boolean);
var
  RemoteId: TOID;
begin
  {        }
  if AObjectNode.AttributeNodes.FindNode('LocalOID') <> nil then
  begin
    { LocalOID      ,   
         RemoteId }
    RemoteId := AObjectNode.Attributes['LocalOID'];
    AObjectNode.AttributeNodes.Delete('LocalOID');
    AObjectNode.AttributeNodes.Delete('RemoteOID');
    UpdateRemoteObjectId(AObject.ClassId, AObject.ObjectId, RemoteId);
  end;
end;

{{
   .
}
procedure TovDBSynchronization.DoRelationImport(Sender: TObject; AObject: IovObject;
  ARelation: String; AChildObjectNode: IXMLNode; var Skip: Boolean);
begin
  {  ,         }
  AChildObjectNode.AttributeNodes.Delete('LocalOID');
  AChildObjectNode.AttributeNodes.Delete('RemoteOID');
end;

{{
    (  ).
}
procedure TovDBSynchronization.DoChildObjectImport(Sender: TObject; AObject: IovObject;
    ARelation: String; AChildObjectNode: IXMLNode; var AObjectId: TOID);
var
  LocalId, RemoteId: TOID;
begin
  if Assigned(OnChildObjectImport) then
  begin
    OnChildObjectImport(Self, AObject, ARelation, AChildObjectNode, AObjectId);
    if AObjectId < 0 then Exit; {   }
  end;

  if AChildObjectNode.AttributeNodes.FindNode('LocalOID') <> nil then
  begin
    RemoteId := AChildObjectNode.Attributes['LocalOID'];
    LocalId := GetLocalObjectId(AObject.ClassId, RemoteId);
    if LocalId <> 0 then AObjectId := LocalId;
  end
end;

{***************************************************************************** }
{                                           }
{ **************************************************************************** }

{{
     RemoteDBId  
ALocalObjectId,    AClassId.
}
function TovDBSynchronization.GetRemoteObjectId(AClassId: Integer;
  ALocalObjectId: TOID): TOID;
begin
  with Model.DatabaseAdapter.OpenQuery(
    Format('SELECT * FROM %s WHERE %s = ''%s'' and %s = %d and %s = %d',
      [LookupTableName, ltDBIdFieldName, RemoteDBId, ltClassIdFieldName, AClassId, ltLocalObjectIdFieldName, ALocalObjectId])) do
  try
    if Not IsEmpty
      then Result := FieldByName(ltRemoteObjectIdFieldName).AsInteger
      else Result := 0;
  finally
    Free
  end;
end;

{{
         
ARemoteObjectId   RemoteDBId,    AClassId.
}
function TovDBSynchronization.GetLocalObjectId(AClassId: Integer;
  ARemoteObjectId: TOID): TOID;
begin
  with Model.DatabaseAdapter.OpenQuery(
    Format('SELECT * FROM %s WHERE %s = ''%s'' and %s = %d and %s = %d',
      [LookupTableName, ltDBIdFieldName, RemoteDBId, ltClassIdFieldName, AClassId, ltRemoteObjectIdFieldName, ARemoteObjectId])) do
  try
    if Not IsEmpty
      then Result := FieldByName(ltLocalObjectIdFieldName).AsInteger
      else Result := 0;
  finally
    Free
  end;
end;

{{
     RemoteDBId   ALocalObjectId,
   AClassId,    ARemoteObjectId.
}
procedure TovDBSynchronization.UpdateRemoteObjectId(AClassId: Integer;
  ALocalObjectId, ARemoteObjectId: TOID);
begin
  if GetRemoteObjectId(AClassId, ALocalObjectId) <> 0 then
    Model.DatabaseAdapter.ExecQuery(
      Format('UPDATE %s SET %s = %d WHERE %s = ''%s'' and %s = %d and %s = %d',
        [LookupTableName, ltRemoteObjectIdFieldName, ARemoteObjectId,
          ltDBIdFieldName, RemoteDBId, ltClassIdFieldName, AClassId, ltLocalObjectIdFieldName, ALocalObjectId]))
  else
    Model.DatabaseAdapter.ExecQuery(
      Format('INSERT INTO %s (%s, %s, %s, %s) VALUES (''%s'', %d, %d, %d)',
        [LookupTableName,
        ltDBIdFieldName, ltClassIdFieldName, ltLocalObjectIdFieldName, ltRemoteObjectIdFieldName,
        RemoteDBId, AClassId, ALocalObjectId, ARemoteObjectId]))
end;

procedure TovDBSynchronization.GenerateLookupTable;
var
  TableDefs: TTableDefs;
begin
  ModelNeeded;
  Model.Open;

  TableDefs := TTableDefs.Create(LookupTableName);
  try
    with TableDefs do
    begin
      AddField(ltRecordIdFieldName, ftInteger, 0);
      AddField(ltDBIdFieldName, ftString, 30);
      AddField(ltClassIdFieldName, ftInteger, 0);
      AddField(ltLocalObjectIdFieldName, ftInteger, 0);
      AddField(ltRemoteObjectIdFieldName, ftInteger, 0);

      AddIndex('MainIdx', ltDBIdFieldName+';'+ltClassIdFieldName+';'+ltLocalObjectIdFieldName, [ixUnique]);

      Model.DatabaseAdapter.CreateTable(TableName, FieldDefs, IndexDefs);
    end;
  finally
    TableDefs.Free
  end;
end;

{{
 -     .
}
function TovDBSynchronization.IsRemoteObject(AObject: IovObject; ADBId: TDBId):
  Boolean;
begin
  with Model.DatabaseAdapter.OpenQuery(
    Format('SELECT * FROM %s WHERE %s = ''%s'' and %s = %d and %s = %d',
      [LookupTableName,
        ltDBIdFieldName, ADBId, ltClassIdFieldName, AObject.ClassId,
        ltLocalObjectIdFieldName, AObject.ObjectId])) do
  try
    Result := Not IsEmpty and (FieldByName(ltRemoteObjectIdFieldName).AsInteger <> 0)
  finally
    Free
  end;
end;

end.
