unit ovInspector;

interface

uses ExtCtrls, Classes, SysUtils, Controls, Contnrs, Windows, StdCtrls, Forms;

type
  TovCustomInspector = class;
  TInspectorItem = class;

  TItemEditedEvent = procedure(Sender: TObject; Item: TInspectorItem) of object;

  TInspectorItemProperties = class(TPersistent)
  private
    FItem: TInspectorItem;
    FControl: TWinControl;
  protected
    function GetValue: Variant; virtual; abstract;
    procedure SetValue(const Value: Variant); virtual; abstract;
    function GetReadOnly: Boolean; dynamic;
    procedure SetReadOnly(const Value: Boolean); dynamic;
    procedure PlaceControlAt(ALeft, ATop, AWidth, AHeight: Integer); dynamic;
  public
    constructor Create(AItem: TInspectorItem); overload; virtual; abstract;
    destructor Destroy; override;
    constructor Create(AItem: TInspectorItem; AControl: TWinControl); overload;
      virtual;
    property Item: TInspectorItem read FItem;
    property Control: TWinControl read FControl;
    property Value: Variant read GetValue write SetValue;
  end;
  TInspectorItemPropertiesClass = class of TInspectorItemProperties;

  TInspectorItem = class(TCollectionItem)
  private
    FCaption: String;
    FPropertiesClass: String;
    FProperties: TInspectorItemProperties;
    FTag: Integer;
    FOnEnter: TNotifyEvent;
    FOnExit: TNotifyEvent;
    FOnEdited: TNotifyEvent;
    procedure SetReadOnly(const Value: Boolean);
    procedure SetCaption(const Value: String);
    procedure SetPropertiesClass(const Value: String);
    procedure UpdateInspectorItem;
    function GetValue: Variant;
    procedure SetValue(const Value: Variant);
    function GetEnabled: Boolean;
    procedure SetEnabled(const Value: Boolean);
    function GetReadOnly: Boolean;
    procedure ControlEnter(Sender: TObject);
    procedure ControlExit(Sender: TObject);
    procedure ControlKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
  protected
    function GetDisplayName: string; override;
    function GetInspector: TovCustomInspector;
    procedure SetProperties(const Value: TInspectorItemProperties);
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    property Inspector: TovCustomInspector read GetInspector;
    property Value: Variant read GetValue write SetValue;
  published
    property Caption: String read FCaption write SetCaption;
    property PropertiesClass: String read FPropertiesClass write SetPropertiesClass;
    property Properties: TInspectorItemProperties read FProperties write
      SetProperties;
    property Enabled: Boolean read GetEnabled write SetEnabled default True;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property OnEnter: TNotifyEvent read FOnEnter write FOnEnter;
    property OnExit: TNotifyEvent read FOnExit write FOnExit;
    property OnEdited: TNotifyEvent read FOnEdited write FOnEdited;
    property Tag: Integer read FTag write FTag;
  end;

  TInspectorItemCollection = class(TCollection)
  private
    FInspector: TovCustomInspector;
    function GetItems(Index: Integer): TInspectorItem;
    function GetItemByCaption(ItemCaption: String): TInspectorItem;
  protected
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(AInspector: TovCustomInspector);
    property Items[Index: Integer]: TInspectorItem read GetItems; default;
    property ItemByCaption[ItemCaption: String]: TInspectorItem read
      GetItemByCaption;
  end;

  TovCustomInspector = class(TCustomPanel)
  private
    FItems: TInspectorItemCollection;
    FControls: TComponentList;
    FLeftPanel: TPanel;
    FRightPanel: TPanel;
    FRowsHeight: Integer;
    FFocusedItem: TInspectorItem;
    procedure SetItems(const Value: TInspectorItemCollection);
    procedure ClearControls;
    procedure CreateItemControls(Item: TInspectorItem);
    procedure UpdateItemControls(Item: TInspectorItem);
    procedure SetCaptionColumnWidth(const Value: Integer);
    function GetCaptionColumnWidth: Integer;
  protected
    procedure Loaded; override;
    procedure RebuildControls; dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property FocusedItem: TInspectorItem read FFocusedItem write FFocusedItem;
  published
    property Items: TInspectorItemCollection read FItems write SetItems;
    property CaptionColumnWidth: Integer read GetCaptionColumnWidth write
      SetCaptionColumnWidth default 120;
  end;

  TovInspector = class(TovCustomInspector)
  published
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BiDiMode;
    property BorderWidth;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property UseDockManager default True;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FullRepaint;
    property Font;
    property Locked;
    property ParentBiDiMode;
    property ParentBackground;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

function GetPropertiesClass(PropertiesClassName: String): TInspectorItemPropertiesClass;
procedure GetRegisteredProperties(PropertiesClassNames: TStrings);
procedure RegisterPropertiesClass(PropertiesClass: TInspectorItemPropertiesClass);

implementation

type
  TWinControlFriend = class(TWinControl)
  end;

var
  RegisteredProperties: TStringList = nil;

procedure RegisterPropertiesClass(PropertiesClass: TInspectorItemPropertiesClass);
begin
  RegisteredProperties.AddObject(PropertiesClass.ClassName, TObject(PropertiesClass))
end;

function GetPropertiesClass(PropertiesClassName: String): TInspectorItemPropertiesClass;
var
  I: Integer;
begin
  I := RegisteredProperties.IndexOf(PropertiesClassName);
  if I >= 0 then
    Result := TInspectorItemPropertiesClass(RegisteredProperties.Objects[I])
  else
    raise Exception.Create('    ' + PropertiesClassName);
end;

procedure GetRegisteredProperties(PropertiesClassNames: TStrings);
begin
  PropertiesClassNames.Assign(RegisteredProperties);
end;

{ TInspectorItemCollection }

constructor TInspectorItemCollection.Create(AInspector: TovCustomInspector);
begin
  inherited Create(TInspectorItem);
  FInspector := AInspector;
end;

function TInspectorItemCollection.GetItems(Index: Integer): TInspectorItem;
begin
  Result := TInspectorItem(inherited Items[Index]);
end;

procedure TInspectorItemCollection.Update(Item: TCollectionItem);
begin
  if Not (csDestroying in FInspector.ComponentState)
    then FInspector.RebuildControls
end;

function TInspectorItemCollection.GetItemByCaption(ItemCaption: String):
  TInspectorItem;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    if Items[I].Caption = ItemCaption then
    begin
      Result := Items[I];
      Exit;
    end;
  raise Exception.CreateFmt(' "%s"  ', [ItemCaption]);
end;

{ TInspectorItem }

constructor TInspectorItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
end;

destructor TInspectorItem.Destroy;
begin
  FreeAndNil(FProperties);
  inherited Destroy;
end;

function TInspectorItem.GetInspector: TovCustomInspector;
begin
  Result := TInspectorItemCollection(Collection).FInspector;
end;

function TInspectorItem.GetDisplayName: string;
begin
  Result := Caption;
end;

procedure TInspectorItem.UpdateInspectorItem;
begin
  Inspector.UpdateItemControls(Self);
  if Assigned(Properties) then
  begin
    TWinControlFriend(Properties.Control).OnEnter := ControlEnter;
    TWinControlFriend(Properties.Control).OnExit := ControlExit;
    TWinControlFriend(Properties.Control).OnKeyDown := ControlKeyDown;
  end;
end;

{{
      
}
procedure TInspectorItem.SetProperties(const Value: TInspectorItemProperties);
begin
end;

{***************************************************************************** }
{                                                             }
{ **************************************************************************** }

procedure TInspectorItem.SetPropertiesClass(const Value: String);
begin
  if FPropertiesClass <> Value then
  begin
    FreeAndNil(FProperties);
    FProperties := GetPropertiesClass(Value).Create(Self);
    FPropertiesClass := Value;
    UpdateInspectorItem;
  end;
end;

procedure TInspectorItem.SetCaption(const Value: String);
begin
  if FCaption <> Value then
  begin
    FCaption := Value;
    UpdateInspectorItem;
  end;
end;

procedure TInspectorItem.SetReadOnly(const Value: Boolean);
begin
  if Assigned(Properties) then Properties.SetReadOnly(Value)
end;

function TInspectorItem.GetReadOnly: Boolean;
begin
  if Assigned(Properties)
  then
    Result := Properties.GetReadOnly
  else
    Result := True
end;

function TInspectorItem.GetEnabled: Boolean;
begin
  if Assigned(Properties) then
    Result := Properties.Control.Enabled
  else
    Result := False
end;

procedure TInspectorItem.SetEnabled(const Value: Boolean);
begin
  if Assigned(Properties) then Properties.Control.Enabled := Value;
end;

function TInspectorItem.GetValue: Variant;
begin
  if Assigned(Properties) then Result := Properties.Value
end;

procedure TInspectorItem.SetValue(const Value: Variant);
begin
  if Assigned(Properties) then Properties.Value := Value
end;

procedure TInspectorItem.ControlEnter(Sender: TObject);
begin
  Inspector.FFocusedItem := Self;
  if Assigned(FOnEnter) then FOnEnter(Self);
end;

procedure TInspectorItem.ControlExit(Sender: TObject);
begin
  if Assigned(FOnEdited) then FOnEdited(Self);
  if Assigned(FOnExit) then FOnExit(Self);
  Inspector.FFocusedItem := nil;
end;

procedure TInspectorItem.ControlKeyDown(Sender: TObject; var Key: Word; Shift:
  TShiftState);
begin
  if (Key = VK_DOWN) and (Shift = []) then Inspector.SelectNext(Screen.ActiveControl, True, True) else
  if (Key = VK_UP) and (Shift = []) then Inspector.SelectNext(Screen.ActiveControl, False, True) else
  if (Key = VK_RETURN) and Assigned(FOnEdited) then FOnEdited(Self)
  else Exit;
  Key := 0;
end;

{ TovCustomInspector }

constructor TovCustomInspector.Create(AOwner: TComponent);

  procedure CreateBackgroundControls;
  begin
    FLeftPanel := TPanel.Create(Self);
    FLeftPanel.Parent := Self;
    FLeftPanel.ParentColor := True;
    FLeftPanel.BevelOuter := bvNone;
    FLeftPanel.Align := alLeft;
    FLeftPanel.Width := CaptionColumnWidth;

    with TSplitter.Create(Self) do
    begin
      Beveled := True;
      Align := alLeft;
      Width := 2;
      Parent := Self;
      Left := FLeftPanel.BoundsRect.Right;
    end;

    FRightPanel := TPanel.Create(Self);
    FRightPanel.Parent := Self;
    FRightPanel.ParentColor := True;
    FRightPanel.BevelOuter := bvNone;
    FRightPanel.Align := alClient;
  end;

begin
  inherited Create(AOwner);
  FItems := TInspectorItemCollection.Create(Self);
  FControls := TComponentList.Create;
  CreateBackgroundControls;
  CaptionColumnWidth := 120;
end;

destructor TovCustomInspector.Destroy;
begin
  FreeAndNil(FControls);
  FreeAndNil(FItems);
  inherited Destroy;
end;

{{
      
}
procedure TovCustomInspector.SetItems(const Value: TInspectorItemCollection);
begin
end;

procedure TovCustomInspector.Loaded;
begin
  inherited;
  RebuildControls;
end;

procedure TovCustomInspector.RebuildControls;
var
  I: Integer;
begin
  if Not (csLoading in ComponentState) then
  begin
    LockWindowUpdate(Handle);
    try
      ClearControls;
      with Items do
       for I := 0 to Count - 1 do
         CreateItemControls(Items[I])
    finally
      LockWindowUpdate(0);
    end;
  end;
end;

procedure TovCustomInspector.ClearControls;
begin
  FControls.Clear;
  FRowsHeight := 0;
end;

function TovCustomInspector.GetCaptionColumnWidth: Integer;
begin
  Result := FLeftPanel.Width;
end;

procedure TovCustomInspector.SetCaptionColumnWidth(const Value: Integer);
begin
  FLeftPanel.Width := Value;
end;

{***************************************************************************** }
{                       }
{ **************************************************************************** }

{{
       .

:
-   ,     
-    
-         
}
procedure TovCustomInspector.CreateItemControls(Item: TInspectorItem);

  procedure CreateCaption(Item: TInspectorItem);
  var
    ALabel: TLabel;
  begin
    ALabel := TLabel.Create(Self);
    with ALabel do
    begin
      Parent := FLeftPanel;
      Caption := Item.Caption;
      AutoSize := False;
      WordWrap := True;
      Left := 5;
      Top := Item.Properties.Control.Top + 2;
      Width := FLeftPanel.Width - 5;
      Height := Item.Properties.Control.Height - 2;
      Anchors := Anchors + [akRight];
    end;

    FControls.Add(ALabel);
  end;

  procedure CreateSeparator(Item: TInspectorItem);
  var
    ABevel: TBevel;
  begin
    ABevel := TBevel.Create(Self);
    ABevel.Parent := FRightPanel;
    ABevel.Shape := bsBottomLine;

    FControls.Add(ABevel);

    ABevel.Left := 0;
    ABevel.Width := FRightPanel.Width;
    ABevel.Anchors := ABevel.Anchors + [akRight];

    ABevel.Top := Item.Properties.Control.Top - 1;
    ABevel.Height := Item.Properties.Control.Height+4;
    Inc(FRowsHeight, 2);
  end;

begin
  if Assigned(Item.Properties) then
  begin
    Item.Properties.PlaceControlAt(5, FRowsHeight+1, FRightPanel.Width - 6, 0);
    Item.Properties.Control.Anchors := Item.Properties.Control.Anchors + [akRight];
    Inc(FRowsHeight, Item.Properties.Control.Height+1);

    CreateCaption(Item);
    CreateSeparator(Item);
  end;
end;

procedure TovCustomInspector.UpdateItemControls(Item: TInspectorItem);
begin
  if Not (csLoading in ComponentState) then RebuildControls;
end;

{ TInspectorItemProperties }

constructor TInspectorItemProperties.Create(AItem: TInspectorItem; AControl:
  TWinControl);
begin
  FItem := AItem;
  FControl := AControl;
  FControl.Parent := Item.Inspector.FRightPanel;
end;

destructor TInspectorItemProperties.Destroy;
begin
  FreeAndNil(FControl);
  inherited Destroy;
end;

procedure TInspectorItemProperties.PlaceControlAt(ALeft, ATop, AWidth, AHeight:
  Integer);
begin
  Control.SetBounds(ALeft, ATop, AWidth, Control.Height);
end;

function TInspectorItemProperties.GetReadOnly: Boolean;
begin
  Result := Not Control.Enabled
end;

procedure TInspectorItemProperties.SetReadOnly(const Value: Boolean);
begin
  Control.Enabled := Not Value
end;

initialization
  RegisteredProperties := TStringList.Create;
finalization
  FreeAndNil(RegisteredProperties);
end.
