unit ovMemObjectManager;

interface

uses SysUtils, Classes, DB, ovTypes, ovModel, ovMemObjects, ovComponents, ovIObjects,
  Contnrs;

type
  TovMemoryObject = class;

  TovMemoryObjectList = class (TList)
  private
    function GetItems(Index: Integer): TovMemoryObject;
  public
    function Find(AObjectClassName: String; AObjectId: TOID): Integer; overload;
    function Find(const Attributes: array of String; const Values: array of
      Variant): Integer; overload;
    property Items[Index: Integer]: TovMemoryObject read GetItems; default;
  end;

  TovMemoryObjectClass = class of TovMemoryObject;

  TValidateObjectEvent = procedure (Sender: TObject; AMemoryObject: TovMemoryObject) of object;

  TMemoryObjectManagerOption = (moValidateObjectOnSave);
  TMemoryObjectManagerOptions = set of TMemoryObjectManagerOption;

  TovMemoryObjectManager = class(TovModelClient)
  private
    FObjects: TObjectList;
    FObjectClasses: TStringList;
    FDefaultObjectClass: TovMemoryObjectClass;
    FOnValidateObject: TValidateObjectEvent;
    FOptions: TMemoryObjectManagerOptions;
    function GetObjectCount: Integer;
    function GetObjects(Index: Integer): TovMemoryObject;
    function CreateObjectInstance(ADataSet: TDataSet): TovMemoryObject; overload;
  protected
    function GetObjectClass(AObjectClassName: String; AObjectID: TOID):
      TovMemoryObjectClass; virtual;
    procedure ValidateObject(AMemoryObject: TovMemoryObject); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function CreateObjectInstance(AObjectClassName: String; AObjectId: TOID;
      LoadFromDBImmediately: Boolean; AParentObject: TovMemoryObject = nil):
      TovMemoryObject; overload;
    {{
        .
    }
    procedure ReleaseAllObjects;
    procedure RegisterObjectClass(AObjectClassName: String; AObjectClass:
      TovMemoryObjectClass);
    procedure UnregisterObjectClass(AObjectClassName: String);
    function FindObject(AObjectClassName: String; AObjectId: TOID): TovMemoryObject;
    function GenerateObjectID(AObjectClassName: String): TOID;
    procedure FreeObjectInstance(AMemoryObject: TovMemoryObject);
    function CreateNewObject(AObjectClassName: String): TovMemoryObject;
    function GetObjectFromDB(ADBObject: IovObject): TovMemoryObject; overload;
    function GetObjectFromDB(AObjectClassName: String; AObjectId: TOID):
      TovMemoryObject; overload;
    function GetObjectsFromDB(AObjectClassName: String; AFilter: String = ''):
      TovMemoryObjectList;

    {{
       ,   
    CreateObjectInstance.
    }
    property ObjectCount: Integer read GetObjectCount;
    {{
         ,  
     CreateObjectInstance.
    }
    property Objects[Index: Integer]: TovMemoryObject read GetObjects;
    property DefaultObjectClass: TovMemoryObjectClass read FDefaultObjectClass
      write FDefaultObjectClass;
  published
    property Options: TMemoryObjectManagerOptions read FOptions write FOptions
      default [moValidateObjectOnSave];
    property OnValidateObject: TValidateObjectEvent read FOnValidateObject write
      FOnValidateObject;
  end;

  EomObjectValidationError = class;
  EomObjectValidationErrorClass = class of EomObjectValidationError;

  TovMemoryObject = class(TMemoryObject)
  private
    FDBObject: IovObject;
    FObjectManager: TovMemoryObjectManager;
    FModelObject: TModelObject;
    FParent: TovMemoryObject;
    procedure BindWithModelObject(AModelObject: TModelObject);
    procedure SaveObjectToDB(ADBObject: IovObject); overload;
    function IsSameObject(AMemoryObject: TMemoryObject): Boolean; overload;
    function GetObjectSpace: IovObjectSpace;
    class function GetSelectQuery(AModelObject: TModelObject; AObjectId: TOID;
      AFilter: String): String;
    procedure LoadObjectAttributes(ADataSet: TDataSet); overload;
    procedure LoadObjectFromDB(ADataSet: TDataSet); overload;
  protected
    procedure InitAttribute(Attribute: TMemoryObjectAttribute); override;
    function GetObject(AObjectClassName: String; AObjectId: TOID): TMemoryObject;
      override;
    procedure CalculateAttributes; virtual;
    procedure InitRelation(ARelation: TMemoryObjectRelation); override;
    procedure SaveRelation(ARelation: TModelRelation); virtual;
    procedure NewObject; virtual;
    procedure LoadObjectFromDB; overload; virtual;
    procedure SaveObjectToDB; overload; virtual;
    procedure SaveAttribute(Attribute: TModelAttribute); virtual;
  public
    procedure Refresh;
    constructor Create(AObjectManager: TovMemoryObjectManager; const
      AObjectClassName: String; AObjectId: TOID; AParent: TovMemoryObject = nil);
      virtual;
    procedure Update; overload;
    procedure Recalculate;
    function IsSameObject(ADBObject: IovObject): Boolean; overload;
    procedure ValidateError(ExceptionClass: EomObjectValidationErrorClass;
      ErrorMsg, ErrorSource: String); virtual;
    procedure Validate; virtual;
    property ObjectManager: TovMemoryObjectManager read FObjectManager;
    {{
          ,   
     .
    }
    property DBObject: IovObject read FDBObject;
    property ModelObject: TModelObject read FModelObject;
    property Parent: TovMemoryObject read FParent;
    property ObjectSpace: IovObjectSpace read GetObjectSpace;
  end;

  EomException = class (Exception);
  EomObjectNotFoundError = class (EomException);

  {    }

  EomObjectValidationError = class(EomException)
  private
    FErrorObject: TovMemoryObject;
    FErrorSource: String;
  public
    constructor Create(const Msg: string; ASource: TovMemoryObject; const
      AErrorSource: String);
  published
    property ErrorObject: TovMemoryObject read FErrorObject;
    property ErrorSource: String read FErrorSource;
  end;

  EomAttributeValidationError = class(EomObjectValidationError);
  EomRelationValidationError = class(EomObjectValidationError);
  EomRequiredAttributeError = class(EomAttributeValidationError);
  EomRequiredRelationError = class(EomRelationValidationError);
  EomRelationOutOfMaxCountError = class(EomRelationValidationError);
  EomRelationOutOfMinCountError = class(EomRelationValidationError);

implementation

uses Variants, ovConsts;

{ EomObjectValidationError }

constructor EomObjectValidationError.Create(const Msg: string; ASource:
  TovMemoryObject; const AErrorSource: String);
begin
  inherited Create(Msg);
  FErrorObject := ASource;
  FErrorSource := AErrorSource;
end;

{ TovMemoryObjectList }

function TovMemoryObjectList.GetItems(Index: Integer): TovMemoryObject;
begin
  Result := TovMemoryObject(inherited Items[Index]);
end;

{{
       .
}
function TovMemoryObjectList.Find(AObjectClassName: String; AObjectId: TOID):
  Integer;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    if (Items[I].ObjectClassName = AObjectClassName) and (Items[I].ObjectId = AObjectId)  then
    begin
      Result := I;
      Exit;
    end;
  Result := -1;
end;

{{
      (  Attributes) ,
   ,   Values.
}
function TovMemoryObjectList.Find(const Attributes: array of String; const
  Values: array of Variant): Integer;

  {{
   True,   ,   Attributes 
    Index   ,   Values.
  }
  function CheckObject(Index: Integer): Boolean;
  var
    I: Integer;
  begin
    for I := 0 to Length(Attributes) - 1 do
      if Items[Index][Attributes[I]].Value <> Values[I] then
      begin
        Result := False;
        Exit;
      end;
    Result := True;
  end;

var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    if CheckObject(I) then
    begin
      Result := I;
      Exit;
    end;
  Result := -1;
end;

{ TovMemoryObject }

constructor TovMemoryObject.Create(AObjectManager: TovMemoryObjectManager;
  const AObjectClassName: String; AObjectId: TOID; AParent: TovMemoryObject =
  nil);
var
  AModelObject: TModelObject;
begin
  inherited Create(AObjectClassName, AObjectId);
  FObjectManager := AObjectManager;
  FParent := AParent;

  {     ,      }
  FObjectManager.ModelNeeded;
  AModelObject := FObjectManager.DomainModel.GetObject(AObjectClassName);
  BindWithModelObject(AModelObject);
end;

{{
     ,     .
}
procedure TovMemoryObject.BindWithModelObject(AModelObject: TModelObject);
var
  I: Integer;
begin
  FModelObject := AModelObject;

  Clear;

  for I := 0 to AModelObject.AttributeCount - 1 do
    AddAttribute(AModelObject.Attributes[I].Name, Null);

  for I := 0 to AModelObject.RelationCount - 1 do
    AddRelation(AModelObject.Relations[I].Name, AModelObject.Relations[I].TargetObjectName);
end;

function TovMemoryObject.GetObjectSpace: IovObjectSpace;
begin
  Result := ObjectManager.Model.ObjectSpace;
end;

function TovMemoryObject.IsSameObject(ADBObject: IovObject): Boolean;
begin
  Result := Assigned(ADBObject) and
    (ObjectClassName = ADBObject.ClassName) and (ObjectId = ADBObject.ObjectId)
end;

function TovMemoryObject.IsSameObject(AMemoryObject: TMemoryObject): Boolean;
begin
  Result := Assigned(AMemoryObject) and
    (ObjectClassName = AMemoryObject.ObjectClassName) and (ObjectId = AMemoryObject.ObjectId)
end;

{{
        .
}
procedure TovMemoryObject.NewObject;
begin

end;

{{
        
 TMemoryObjectRelation.Objects.    
  !
}
function TovMemoryObject.GetObject(AObjectClassName: String; AObjectId: TOID):
  TMemoryObject;
begin
  Result := ObjectManager.FindObject(AObjectClassName, AObjectId);
  if Not Assigned(Result) then
    Result := ObjectManager.CreateObjectInstance(AObjectClassName, AObjectId, True, Self)
end;

{{
     ( ) 
   (   
  ). ,      
 !!
}
procedure TovMemoryObject.InitAttribute(Attribute: TMemoryObjectAttribute);
begin
  if Assigned(DBObject) then
    Attribute.Value := DBObject[Attribute.Name].Value
  else
    Attribute.Value := Null; {        }
end;

procedure TovMemoryObject.InitRelation(ARelation: TMemoryObjectRelation);
var
  I: Integer;
begin
  if Assigned(DBObject) then
  with FDBObject[ARelation.Name].AsObjects do
    for I := 0 to Count - 1 do
      with Objects[I] do
        { TODO : ClassName           -   }
        ARelation.AddObject(ObjectId, ClassName);
end;

{{
         
  .
}
procedure TovMemoryObject.Refresh;
begin
  {     ( ,   
  InitAttribute             )
         }
  Clear;
  {     ()}
  BindWithModelObject(ModelObject);
  {     }
  LoadObjectFromDB;
end;

procedure TovMemoryObject.Recalculate;
begin
  CalculateAttributes;
end;

{{
         
 .
}
procedure TovMemoryObject.CalculateAttributes;
begin
end;

{{
          
 .           .
}
procedure TovMemoryObject.Validate;
var
  I: Integer;
  Relation: TModelRelation;
  Attribute: TModelAttribute;
  ACount: Integer;
begin
  with ModelObject do
  begin
    //   
    for I := 0 to AttributeCount - 1 do
    begin
      Attribute := Attributes[I];
      if Attribute.Required and Self[Attribute.Name].IsNull then
        ValidateError(EomRequiredAttributeError, Attribute.RequiredMsg, Attribute.Name);
    end;

    //      
    for I := 0 to RelationCount - 1 do
    begin
      Relation := Relations[I];
      if Not Relation.Derived then
      begin
        ACount := RelationByName[Relation.Name].ObjectCount;
  
        if Relation.Required and (ACount = 0) then
          ValidateError(EomRequiredRelationError, Relation.RequiredMsg, Relation.Name);
        if ACount < Relation.TargetMinCount then
          ValidateError(EomRelationOutOfMinCountError,
            Format('   "%s"   ',
              [Relation.Name]), Relation.Name);
        if ACount > Relation.TargetMaxCount then
          ValidateError(EomRelationOutOfMaxCountError,
            Format('     "%s"',
              [Relation.Name]), Relation.Name);
      end;
    end;
  end;
end;

procedure TovMemoryObject.ValidateError(ExceptionClass:
  EomObjectValidationErrorClass; ErrorMsg, ErrorSource: String);
begin
  raise ExceptionClass.Create(ErrorMsg, Self, ErrorSource);
end;

{***************************************************************************** }
{                                                           }
{ **************************************************************************** }

{{
       
    ObjectClassName   ObjectId.
}
procedure TovMemoryObject.LoadObjectFromDB;
var
  DataSet: TDataSet;
begin
  DataSet := ObjectSpace.OpenQuery(GetSelectQuery(ModelObject, ObjectId, ''));
  try
    if Not DataSet.Eof then
      LoadObjectFromDB(DataSet)
    else
      EomObjectNotFoundError.CreateFmt(
        '  "%s"   %d    ', [ModelObject.FullName, ObjectId]);
  finally
    DataSet.Free
  end;
end;

{{
  -        
 .
}
procedure TovMemoryObject.LoadObjectFromDB(ADataSet: TDataSet);
begin
  ObjectManager.ModelNeeded;
  FDBObject := ObjectSpace.GetObject(ObjectClassName, ObjectId);
  LoadObjectAttributes(ADataSet);
  CalculateAttributes;
end;

{{
      ,   
 AObjectId (AObjectId <> 0)   , 
 AFilter,   AModelObject.      
     ,    
     .
}
class function TovMemoryObject.GetSelectQuery(AModelObject: TModelObject;
  AObjectId: TOID; AFilter: String): String;
var
  AttributeList: String;
  I: Integer;
begin
  AttributeList := '';
  with AModelObject do
  for I := 0 to AttributeCount - 1 do
    if Attributes[I].Persistent then
      AttributeList := AttributeList + ', [' + Attributes[I].Name +']';
  Delete(AttributeList, 1, 1);

  if AObjectId <> 0 then
    Result := Format('SELECT [OID], [CID], %s FROM [%s] WHERE [OID] = %d', [AttributeList, AModelObject.FullName, AObjectId])
  else
    if AFilter <> '' then
      Result := Format('SELECT [OID], [CID], %s FROM [%s] WHERE %s', [AttributeList, AModelObject.FullName, AFilter])
    else
      Result := Format('SELECT [OID], [CID], %s FROM [%s]', [AttributeList, AModelObject.FullName, AFilter])
end;

{{
      ,   
   ADataSet,    .
}
procedure TovMemoryObject.LoadObjectAttributes(ADataSet: TDataSet);
var
  I, J: Integer;
  Attribute: TModelAttribute;
  AValue: Variant;
begin
  with ADataSet do
  begin
    J := 2;
    for I := 0 to ModelObject.AttributeCount - 1 do
    begin
      Attribute := ModelObject.Attributes[I];
      if Attribute.Persistent then
      begin
        AValue := ADataSet.Fields[J].Value;
        {    }
        if Attribute.IsBoolean and Not ObjectManager.Model.Database.Adapter.SupportsBooleanType
          then AValue := Not VarIsNull(AValue) and (AValue <> FalseValueReplacement);
        {   }
        AttributeValues[Attribute.Name] := AValue;
        Inc(J);
      end
    end;
  end;
end;

{***************************************************************************** }
{                                                          }
{ **************************************************************************** }

procedure TovMemoryObject.Update;
begin
  ObjectManager.ModelNeeded;
  with ObjectSpace do
  begin
    StartTransaction;
    try
      SaveObjectToDB(nil);
      CommitTransaction;
    except
      RollbackTransaction;
      raise
    end;
  end;
end;

{{
        
 .

     (      
)    .
}
procedure TovMemoryObject.SaveObjectToDB(ADBObject: IovObject);
begin
  with ObjectManager do
    if moValidateObjectOnSave in Options then ValidateObject(Self);

  if ADBObject = nil then
  begin
    if ObjectId < 0 then
    begin
      FDBObject := ObjectSpace.CreateObject(ObjectClassName);
      FObjectId := FDBObject.ObjectId;
    end
    else if FDBObject = nil then
      FDBObject := ObjectSpace.GetObject(ObjectClassName, ObjectId);
  end else
  begin
    FDBObject := ADBObject;
    FObjectId := FDBObject.ObjectId;
  end;
  SaveObjectToDB;
end;

procedure TovMemoryObject.SaveObjectToDB;
var
  I: Integer;
begin
  {         DBObject, 
    SaveAttribute      }
  DBObject.BeginUpdate;
  try
    for I := 0 to ModelObject.AttributeCount - 1 do
      if Not ModelObject.Attributes[I].Calculated then
        SaveAttribute(ModelObject.Attributes[I]);
  finally
    DBObject.EndUpdate;
  end;

  {         DBObject, 
    SaveRelation   ,   }
  for I := 0 to ModelObject.RelationCount - 1 do
    if Not ModelObject.Relations[I].Derived then SaveRelation(ModelObject.Relations[I]);
end;

procedure TovMemoryObject.SaveAttribute(Attribute: TModelAttribute);
begin
  DBObject[Attribute.Name].Value := AttributeValues[Attribute.Name];
end;

{{
       ARelation.Name.
      :

-  ,    ,     
-            
         ( ,  
         ,    
      )
-          
     ,   
}
procedure TovMemoryObject.SaveRelation(ARelation: TModelRelation);
var
  J: Integer;
  AChildObject: IovObject;
  AObjectRelation: TMemoryObjectRelation;
begin
  {       ,      , ..
        }
  if Not IsRelationInitialized(ARelation.Name) then Exit;

  AObjectRelation := RelationByName[ARelation.Name];
  if ARelation.IsAssociation then
  begin
    DBObject[ARelation.Name].Clear;
    with AObjectRelation do
      for J := 0 to ObjectCount - 1 do
      begin
        if ObjectIDs[J] > 0 then {     }
          DBObject[ARelation.Name].Add(ObjectIDs[J])
        else begin
          (Objects[J] as TovMemoryObject).SaveObjectToDB(nil);
          DBObject[ARelation.Name].Add(ObjectIDs[J]) {         }
        end
      end;

  end else
  if ARelation.IsComposition then
  begin
    {        }
    with DBObject[ARelation.Name].AsObjects do
      for J := Count - 1 downto 0 do
        if AObjectRelation.FindObjectByID(Objects[J].ObjectId) = nil
          then Delete(J);

    {       }
    with AObjectRelation do
      for J := 0 to ObjectCount - 1 do
      begin
        if ObjectIDs[J] > 0 then {     }
          (Objects[J] as TovMemoryObject).SaveObjectToDB(nil)
        else
        begin
          AChildObject := DBObject[ARelation.Name].Add;
          (Objects[J] as TovMemoryObject).SaveObjectToDB(AChildObject);
        end;
      end;
  end
  else raise Exception.Create(' ');
end;

{ TovMemoryObjectManager }

constructor TovMemoryObjectManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FObjects := TObjectList.Create;
  FObjectClasses := TStringList.Create;
  FObjectClasses.Duplicates := dupIgnore;
  FObjectClasses.Sorted := True;
  DefaultObjectClass := TovMemoryObject;
  Options := [moValidateObjectOnSave];
end;

destructor TovMemoryObjectManager.Destroy;
begin
  FreeAndNil(FObjectClasses);
  FreeAndNil(FObjects);
  inherited Destroy;
end;

{***************************************************************************** }
{                                 }
{ **************************************************************************** }

function TovMemoryObjectManager.GetObjectCount: Integer;
begin
  Result := FObjects.Count;
end;

function TovMemoryObjectManager.GetObjects(Index: Integer): TovMemoryObject;
begin
  Result := TovMemoryObject(FObjects[Index]);
end;

procedure TovMemoryObjectManager.ReleaseAllObjects;
begin
  FObjects.Clear;
end;

{{
   AObjectClassName    AObjectId.
   ,    nil.   
   .
}
function TovMemoryObjectManager.FindObject(AObjectClassName: String; AObjectId:
  TOID): TovMemoryObject;
var
  I: Integer;
begin
  for I := 0 to ObjectCount - 1 do
    if (Objects[I].ObjectId = AObjectId) and (Objects[I].ObjectClassName = AObjectClassName) then
    begin
      Result := Objects[I];
      Exit;
    end;
  Result := nil;
end;

{***************************************************************************** }
{                   /                       }
{ **************************************************************************** }

{{
         
   TMemoryObjectRelation.Objects.   
      FreeObjectInstance.

      .
}
function TovMemoryObjectManager.CreateObjectInstance(AObjectClassName: String;
  AObjectId: TOID; LoadFromDBImmediately: Boolean; AParentObject:
  TovMemoryObject = nil): TovMemoryObject;
begin
  {   }
  Result := GetObjectClass(AObjectClassName, AObjectId).Create(Self, AObjectClassName, AObjectId, AParentObject);
  try

    {    }
    if LoadFromDBImmediately then Result.LoadObjectFromDB;

    {      }
    FObjects.Add(Result);

  except
    Result.Free;
    raise;
  end;
end;

{{
     ,  
   ADataSet.
}
function TovMemoryObjectManager.CreateObjectInstance(ADataSet: TDataSet):
  TovMemoryObject;
var
  AObjectClassName: String;
  AObjectId: TOID;
begin
  {   }
  AObjectId := ADataSet.Fields[0].AsInteger;
  AObjectClassName := Model.DomainModel.GetObject(ADataSet.Fields[1].AsInteger).FullName;
  Result := GetObjectClass(AObjectClassName, AObjectId).Create(Self, AObjectClassName, AObjectId);
  try

    {    }
    Result.LoadObjectFromDB(ADataSet);

    {      }
    FObjects.Add(Result);

  except
    Result.Free;
    raise;
  end;
end;

{{
         .
}
function TovMemoryObjectManager.GetObjectFromDB(ADBObject: IovObject):
  TovMemoryObject;
begin
  Result := CreateObjectInstance(ADBObject.ClassName, ADBObject.ObjectId, True)
end;

{{
         .
}
function TovMemoryObjectManager.GetObjectFromDB(AObjectClassName: String;
  AObjectId: TOID): TovMemoryObject;
begin
  Result := CreateObjectInstance(AObjectClassName, AObjectId, True)
end;

{{
         .
    AObjectClassName,  
AFilter      TovMemoryObjectList.
}
function TovMemoryObjectManager.GetObjectsFromDB(AObjectClassName: String;
  AFilter: String = ''): TovMemoryObjectList;
var
  AObject: TovMemoryObject;
  ADataSet: TDataSet;
begin
  Result := TovMemoryObjectList.Create;
  try
    ADataSet := ObjectSpace.OpenQuery(
      TovMemoryObject.GetSelectQuery(Model.DomainModel.GetObject(AObjectClassName), 0, AFilter));
    with ADataSet do
    try
      while Not Eof do
      begin
        AObject := CreateObjectInstance(ADataSet);
        Result.Add(AObject);
        Next;
      end;
    finally
      Free
    end;
  except
    Result.Free;
    raise;
  end;
end;

{{
         .
}
function TovMemoryObjectManager.CreateNewObject(AObjectClassName: String):
  TovMemoryObject;
begin
  Result := CreateObjectInstance(AObjectClassName, GenerateObjectID(AObjectClassName), False);
  try
    Result.NewObject;
  except
    FreeObjectInstance(Result);
    raise;
  end;
end;

procedure TovMemoryObjectManager.FreeObjectInstance(AMemoryObject:
  TovMemoryObject);
var
  I: Integer;
begin
  I := FObjects.IndexOf(AMemoryObject);
  if I >= 0 then
    FObjects.Delete(I)
  else
    AMemoryObject.Free;
end;

{***************************************************************************** }
{                                                            }
{ **************************************************************************** }

{{
           
 AObjectClassName    AObjectID.
}
function TovMemoryObjectManager.GetObjectClass(AObjectClassName: String;
  AObjectID: TOID): TovMemoryObjectClass;
var
  I: Integer;
begin
  I := FObjectClasses.IndexOf(AObjectClassName);
  if I >= 0 then
    Result := TovMemoryObjectClass(FObjectClasses.Objects[I])
  else
    Result := DefaultObjectClass;
end;

{{
         
   ,   AObjectClassName.
}
procedure TovMemoryObjectManager.RegisterObjectClass(AObjectClassName: String;
  AObjectClass: TovMemoryObjectClass);
begin
  FObjectClasses.AddObject(AObjectClassName, TObject(AObjectClass))
end;

{{
       
      ,   AObjectClassName.  
    AObjectClassName     
 DefaultObjectClass.
}
procedure TovMemoryObjectManager.UnregisterObjectClass(AObjectClassName:
  String);
var
  I: Integer;
begin
  I := FObjectClasses.IndexOf(AObjectClassName);
  if I >= 0 then FObjectClasses.Delete(I);
end;

{***************************************************************************** }
{                                                                  }
{ **************************************************************************** }

{{
       
  .    
 0,    ,   .
}
function TovMemoryObjectManager.GenerateObjectID(AObjectClassName: String):
  TOID;
begin
  Result := 0;
  repeat
    Dec(Result);
  until FindObject(AObjectClassName, Result) = nil;
end;

procedure TovMemoryObjectManager.ValidateObject(AMemoryObject: TovMemoryObject);
begin
  if Assigned(FOnValidateObject) then FOnValidateObject(Self, AMemoryObject);
  AMemoryObject.Validate;
end;

end.
