unit ovMemObjects;

interface

uses
  SysUtils, Windows, Messages, Classes, ovTypes, Contnrs;

type
  TMemoryObjectAttribute = class;
  TMemoryObjectRelation = class;
  TMemoryObject = class(TObject)
  private
    FAttributes: TStringList;
    FRelations: TStringList;
    FObjectClassName: String;
    FAutoCreateAttributes: Boolean;
    function GetAttributeCount: Integer;
    function GetAttributes(Index: Integer): TMemoryObjectAttribute;
    function GetRelationByName(RelationName: String): TMemoryObjectRelation;
    function GetRelationCount: Integer;
    function GetRelations(Index: Integer): TMemoryObjectRelation;
    procedure SetAttributeValues(AttributeName: String; Value: Variant);
    function GetAttributeValues(AttributeName: String): Variant;
    function GetAttributeByName(AttributeName: String): TMemoryObjectAttribute;
  protected
    FObjectId: TOID;
    function GetObject(AObjectClassName: String; AObjectId: TOID): TMemoryObject;
      virtual;
    procedure InitRelation(ARelation: TMemoryObjectRelation); virtual;
    procedure InitAttribute(Attribute: TMemoryObjectAttribute); virtual;
  public
    procedure AddAttribute(AttributeName: String); overload;
    procedure AddRelation(ARelationName: String; const ADefaultObjectClassName:
      String = '');
    procedure DeleteAttribute(AttributeName: String);
    procedure DeleteRelation(RelationName: String);
    constructor Create(const AObjectClassName: String = ''; AObjectId: TOID = 0);
    destructor Destroy; override;
    procedure Clear;
    function IsAttributeExists(AttributeName: String): Boolean;
    procedure DeleteAllAttributes;
    procedure DeleteAllRelations;
    procedure AddAttribute(AttributeName: String; Value: Variant); overload;
    function IsRelationExists(RelationName: String): Boolean;
    function FindRelationByName(RelationName: String): TMemoryObjectRelation;
    function FindAttributeByName(AttributeName: String): TMemoryObjectAttribute;
    function IsRelationInitialized(RelationName: String): Boolean;

    property AttributeValues[AttributeName: String]: Variant read
      GetAttributeValues write SetAttributeValues;
    property AttributeCount: Integer read GetAttributeCount;
    property Attributes[Index: Integer]: TMemoryObjectAttribute read GetAttributes;
    property RelationByName[RelationName: String]: TMemoryObjectRelation read
      GetRelationByName;
    property RelationCount: Integer read GetRelationCount;
    property Relations[Index: Integer]: TMemoryObjectRelation read
      GetRelations;
    property ObjectClassName: String read FObjectClassName;
    property ObjectId: TOID read FObjectId;

    {{
        ( ),   
       AttributeValues   
        .
    }
    property AutoCreateAttributes: Boolean read FAutoCreateAttributes write
      FAutoCreateAttributes;
    property AttributeByName[AttributeName: String]: TMemoryObjectAttribute read
      GetAttributeByName; default;
  published
  end;

  TMemoryObjectItem = class
  private
    FMemoryObject: TMemoryObject;
    FName: String;
  public
    constructor Create(AMemoryObject: TMemoryObject; const AName: String);
    property MemoryObject: TMemoryObject read FMemoryObject;
    property Name: String read FName;
  end;

  TMemoryObjectAttribute = class(TMemoryObjectItem)
  private
    FValueIsInitialized: Boolean;
    FValue: Variant;
    function GetAsBoolean: Boolean;
    function GetAsDateTime: TDateTime;
    function GetAsFloat: Double;
    function GetAsInteger: Integer;
    function GetAsString: String;
    procedure SetAsBoolean(const Value: Boolean);
    procedure SetAsDateTime(const Value: TDateTime);
    procedure SetAsFloat(const Value: Double);
    procedure SetAsInteger(const Value: Integer);
    procedure SetAsString(const Value: String);
    function GetIsNull: Boolean;
    procedure InitValue;
    function GetValue: Variant;
    procedure SetValue(const Value: Variant);
  public
    constructor Create(AMemoryObject: TMemoryObject; AttributeName: String; AValue:
      Variant);
    property AsBoolean: Boolean read GetAsBoolean write SetAsBoolean;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property AsFloat: Double read GetAsFloat write SetAsFloat;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsString: String read GetAsString write SetAsString;
    property Value: Variant read GetValue write SetValue;
    property IsNull: Boolean read GetIsNull;
  end;

  TMemoryObjectRelation = class(TMemoryObjectItem)
  private
    FObjects: TObjectList;
    FRelationIsInitialized: Boolean;
    FDefaultObjectClassName: String;
    function GetObjectClassNames(Index: Integer): String;
    function GetObjectCount: Integer;
    function GetObjectIDs(Index: Integer): TOID;
    function GetObjects(Index: Integer): TMemoryObject;
    function GetAsObjectID: TOID;
    procedure SetAsObjectID(const Value: TOID);
    function GetAsObject: TMemoryObject;
    procedure SetAsObject(const Value: TMemoryObject);
    procedure SetObjectIDs(Index: Integer; const Value: TOID);
    procedure SetObjects(Index: Integer; const Value: TMemoryObject);
    procedure InitRelation;
  public
    constructor Create(AMemoryObject: TMemoryObject; const ARelationName: String;
      const ADefaultObjectClassName: String = '');
    destructor Destroy; override;
    function AddObject(AObjectID: TOID; AObjectClassName: String = ''): Integer;
      overload;
    function AddObject(AObject: TMemoryObject): Integer; overload;
    procedure DeleteObject(Index: Integer);
    function GetObjectByID(AObjectId: TOID): TMemoryObject;
    function FindObjectByAttributeValue(AttributeName: String; AttributeValue:
      Variant): TMemoryObject;
    function IndexOfObject(AObject: TMemoryObject): Integer; overload;
    procedure Clear;
    function FindObjectByID(AObjectId: TOID): TMemoryObject;
    function IndexOfObject(AObjectID: TOID; AObjectClassName: String): Integer;
      overload;
    property ObjectClassNames[Index: Integer]: String read GetObjectClassNames;
    property ObjectCount: Integer read GetObjectCount;
    property ObjectIDs[Index: Integer]: TOID read GetObjectIDs write SetObjectIDs;
    property Objects[Index: Integer]: TMemoryObject read GetObjects write
      SetObjects;
    property AsObjectID: TOID read GetAsObjectID write SetAsObjectID;
    property AsObject: TMemoryObject read GetAsObject write SetAsObject;
  published
    property DefaultObjectClassName: String read FDefaultObjectClassName;
  end;

  EMemoryObjectException = class(Exception);
  EAttributeNotDefined = class(EMemoryObjectException);
  EAttributeAlreadyExists = class(EMemoryObjectException);
  ERelationNotDefined = class(EMemoryObjectException);
  ERelationAlreadyExists = class(EMemoryObjectException);
  EObjectWithIDNotFound = class(EMemoryObjectException);

implementation

uses Variants;

resourcestring
  SAttributeNotDefined = '   : ';
  SAttributeAlreadyExists = '     : ';
  SRelationNotDefined = '   : ';
  SRelationAlreadyExists = '     : ';

type
  TObjectReference = class
  private
    FObjectClassName: String;
    FObjectID: TOID;
    FObjectRef: TMemoryObject;
  public
    constructor Create(const AObjectClassName: String; AObjectID: TOID); overload;
    constructor Create(AObjectRef: TMemoryObject); overload;
    property ObjectClassName: String read FObjectClassName write FObjectClassName;
    property ObjectID: TOID read FObjectID write FObjectID;
    property ObjectRef: TMemoryObject read FObjectRef write FObjectRef;
  end;

procedure FreeObjects(List: TStringList);
var
  I: Integer;
begin
  for I := 0 to List.Count - 1 do
  begin
    List.Objects[I].Free;
    List.Objects[I] := nil;
  end;
end;

{ TMemoryObjectItem }

constructor TMemoryObjectItem.Create(AMemoryObject: TMemoryObject; const AName:
  String);
begin
  inherited Create;
  FMemoryObject := AMemoryObject;
  FName := AName;
end;

{ TObjectReference }

constructor TObjectReference.Create(const AObjectClassName: String; AObjectID:
  TOID);
begin
  inherited Create;
  FObjectID := AObjectID;
  FObjectClassName := AObjectClassName;
end;

constructor TObjectReference.Create(AObjectRef: TMemoryObject);
begin
  inherited Create;
  FObjectRef := AObjectRef;
  FObjectID := AObjectRef.ObjectId;
  FObjectClassName := AObjectRef.ObjectClassName;
end;

{ TMemoryObject }

constructor TMemoryObject.Create(const AObjectClassName: String = '';
  AObjectId: TOID = 0);
begin
  inherited Create;
  FAutoCreateAttributes := True;

  FAttributes := TStringList.Create;
  FAttributes.Sorted := True;
  FAttributes.Duplicates := dupError;

  FRelations := TStringList.Create;
  FRelations.Sorted := True;
  FRelations.Duplicates := dupError;

  FObjectClassName := AObjectClassName;
  FObjectId := AObjectId;
end;

destructor TMemoryObject.Destroy;
begin
  Clear;
  FreeAndNil(FRelations);
  FreeAndNil(FAttributes);
  inherited Destroy;
end;

procedure TMemoryObject.Clear;
begin
  FreeObjects(FAttributes);
  FAttributes.Clear;
  FreeObjects(FRelations);
  FRelations.Clear;
end;

{{
     .  
      .
}
procedure TMemoryObject.InitAttribute(Attribute: TMemoryObjectAttribute);
begin
end;

{{
       . 
   ()   .
}
procedure TMemoryObject.InitRelation(ARelation: TMemoryObjectRelation);
begin
end;


{{
        
 TMemoryObjectRelation.Objects.    
  !
}
function TMemoryObject.GetObject(AObjectClassName: String; AObjectId: TOID):
  TMemoryObject;
begin
  Result := nil;
end;

{***************************************************************************** }
{                          Attribute management                                }
{ **************************************************************************** }

function TMemoryObject.GetAttributeCount: Integer;
begin
  Result := FAttributes.Count
end;

function TMemoryObject.GetAttributes(Index: Integer): TMemoryObjectAttribute;
begin
  Result := TMemoryObjectAttribute(FAttributes.Objects[Index])
end;

function TMemoryObject.GetAttributeValues(AttributeName: String): Variant;
begin
  Result := AttributeByName[AttributeName].Value
end;

procedure TMemoryObject.SetAttributeValues(AttributeName: String; Value:
  Variant);
begin
  AttributeByName[AttributeName].Value := Value
end;

{   }

procedure TMemoryObject.AddAttribute(AttributeName: String);
begin
  if Not IsAttributeExists(AttributeName) then
    FAttributes.AddObject(AttributeName, TMemoryObjectAttribute.Create(Self, AttributeName, Unassigned))
  else
    raise EAttributeAlreadyExists.Create(SAttributeAlreadyExists+AttributeName);
end;

{{
  ,   AddAttribute(AttributeName),   
  .
}
procedure TMemoryObject.AddAttribute(AttributeName: String; Value: Variant);
begin
  if Not IsAttributeExists(AttributeName) then
    FAttributes.AddObject(AttributeName, TMemoryObjectAttribute.Create(Self, AttributeName, Value))
  else
    raise EAttributeAlreadyExists.Create(SAttributeAlreadyExists+AttributeName);
end;

{   }

procedure TMemoryObject.DeleteAttribute(AttributeName: String);
var
  I: Integer;
begin
  I := FAttributes.IndexOf(AttributeName);
  if I < 0 then
    raise EAttributeNotDefined.Create(SAttributeNotDefined+AttributeName);
  FAttributes.Objects[I].Free;
  FAttributes.Delete(I);
end;

procedure TMemoryObject.DeleteAllAttributes;
begin
  FreeObjects(FAttributes);
  FAttributes.Clear;
end;

{   }

{{
       .
}
function TMemoryObject.IsAttributeExists(AttributeName: String): Boolean;
var
  I: Integer;
begin
  I := FAttributes.IndexOf(AttributeName);
  Result := I >= 0
end;

function TMemoryObject.FindAttributeByName(AttributeName: String):
  TMemoryObjectAttribute;
var
  I: Integer;
begin
  I := FAttributes.IndexOf(AttributeName);
  if I < 0 then
    Result := nil
  else
    Result := TMemoryObjectAttribute(FAttributes.Objects[I])
end;

function TMemoryObject.GetAttributeByName(AttributeName: String):
  TMemoryObjectAttribute;
begin
  Result := FindAttributeByName(AttributeName);
  if Result = nil then
    if AutoCreateAttributes then
    begin
      Result := TMemoryObjectAttribute.Create(Self, AttributeName, Unassigned);
      FAttributes.AddObject(AttributeName, Result);
    end else
      raise EAttributeNotDefined.Create(SAttributeNotDefined+AttributeName)
end;

{***************************************************************************** }
{                          Relation management                                 }
{ **************************************************************************** }

function TMemoryObject.GetRelationCount: Integer;
begin
  Result := FRelations.Count
end;

function TMemoryObject.GetRelations(Index: Integer): TMemoryObjectRelation;
begin
  Result := TMemoryObjectRelation(FRelations.Objects[Index]);
  Result.InitRelation;
end;

{   }

procedure TMemoryObject.AddRelation(ARelationName: String; const
  ADefaultObjectClassName: String = '');
var
  NewItem: TMemoryObjectRelation;
begin
  NewItem := TMemoryObjectRelation.Create(Self, ARelationName, ADefaultObjectClassName);
  try
    FRelations.AddObject(ARelationName, NewItem);
  except
    on E: EStringListError do
    begin
      NewItem.Free;
      raise ERelationAlreadyExists.Create(SRelationAlreadyExists+ARelationName);
    end;
  end;
end;

{   }

procedure TMemoryObject.DeleteRelation(RelationName: String);
var
  I: Integer;
begin
  I := FRelations.IndexOf(RelationName);
  if I < 0 then raise ERelationNotDefined.Create(SRelationNotDefined+RelationName);
  FRelations.Objects[I].Free;
  FRelations.Delete(I);
end;

procedure TMemoryObject.DeleteAllRelations;
begin
  FreeObjects(FRelations);
  FRelations.Clear;
end;

{   }

{{
       .
}
function TMemoryObject.IsRelationExists(RelationName: String): Boolean;
var
  I: Integer;
begin
  I := FRelations.IndexOf(RelationName);
  Result := I >= 0
end;

function TMemoryObject.FindRelationByName(RelationName: String):
  TMemoryObjectRelation;
var
  I: Integer;
begin
  I := FRelations.IndexOf(RelationName);
  if I < 0 then
    Result := nil
  else
  begin
    Result := TMemoryObjectRelation(FRelations.Objects[I]);
    Result.InitRelation;
  end;
end;

function TMemoryObject.GetRelationByName(RelationName: String):
  TMemoryObjectRelation;
begin
  Result := TMemoryObjectRelation(FindRelationByName(RelationName));
  if Result = nil then
    raise ERelationNotDefined.Create(SRelationNotDefined+RelationName);
end;

{{
     , ..    , 
  .
}
function TMemoryObject.IsRelationInitialized(RelationName: String): Boolean;
var
  I: Integer;
begin
  I := FRelations.IndexOf(RelationName);
  if I >= 0 then
    Result := TMemoryObjectRelation(FRelations.Objects[I]).FRelationIsInitialized
  else
    raise ERelationNotDefined.Create(SRelationNotDefined+RelationName);
end;

{ TMemoryObjectAttribute }

constructor TMemoryObjectAttribute.Create(AMemoryObject: TMemoryObject;
  AttributeName: String; AValue: Variant);
begin
  inherited Create(AMemoryObject, AttributeName);
  FValue := AValue;
end;

{{
         
.
}
procedure TMemoryObjectAttribute.InitValue;
begin
  if Not FValueIsInitialized then
  begin
    FMemoryObject.InitAttribute(Self);
    FValueIsInitialized := True;
  end;
end;

function TMemoryObjectAttribute.GetAsBoolean: Boolean;
begin
  InitValue;
  if VarIsNull(Value) then Result := False else Result := Value;
end;

function TMemoryObjectAttribute.GetAsDateTime: TDateTime;
begin
  InitValue;
  if VarIsNull(Value) then Result := 0 else Result := Value;
end;

function TMemoryObjectAttribute.GetAsFloat: Double;
begin
  InitValue;
  if VarIsNull(Value) then Result := 0 else Result := Value;
end;

function TMemoryObjectAttribute.GetAsInteger: Integer;
begin
  InitValue;
  if VarIsNull(Value) then Result := 0 else Result := Value;
end;

function TMemoryObjectAttribute.GetAsString: String;
begin
  InitValue;
  if VarIsNull(Value) then Result := '' else Result := Value;
end;

function TMemoryObjectAttribute.GetValue: Variant;
begin
  InitValue;
  Result := FValue;
end;

function TMemoryObjectAttribute.GetIsNull: Boolean;
begin
  InitValue;
  Result := VarIsNull(Value);
end;

procedure TMemoryObjectAttribute.SetAsBoolean(const Value: Boolean);
begin
  FValue := Value;
  FValueIsInitialized := True;
end;

procedure TMemoryObjectAttribute.SetAsDateTime(const Value: TDateTime);
begin
  FValue := Value;
  FValueIsInitialized := True;
end;

procedure TMemoryObjectAttribute.SetAsFloat(const Value: Double);
begin
  FValue := Value;
  FValueIsInitialized := True;
end;

procedure TMemoryObjectAttribute.SetAsInteger(const Value: Integer);
begin
  FValue := Value;
  FValueIsInitialized := True;
end;

procedure TMemoryObjectAttribute.SetAsString(const Value: String);
begin
  FValue := Value;
  FValueIsInitialized := True;
end;

procedure TMemoryObjectAttribute.SetValue(const Value: Variant);
begin
  FValue := Value;
  FValueIsInitialized := True;
end;

{ TMemoryObjectRelation }

constructor TMemoryObjectRelation.Create(AMemoryObject: TMemoryObject; const
  ARelationName: String; const ADefaultObjectClassName: String = '');
begin
  inherited Create(AMemoryObject, ARelationName);
  FObjects := TObjectList.Create;
  FDefaultObjectClassName := ADefaultObjectClassName;
end;

destructor TMemoryObjectRelation.Destroy;
begin
  FreeAndNil(FObjects);
  inherited Destroy;
end;

{{
     .    
     .
}
procedure TMemoryObjectRelation.InitRelation;
begin
  if Not FRelationIsInitialized then
  begin
    FMemoryObject.InitRelation(Self);
    FRelationIsInitialized := True;
  end;
end;

function TMemoryObjectRelation.GetObjectCount: Integer;
begin
  Result := FObjects.Count;
end;

function TMemoryObjectRelation.GetObjects(Index: Integer): TMemoryObject;
begin
  Result := TObjectReference(FObjects[Index]).ObjectRef;
  if Not Assigned(Result) then
    with TObjectReference(FObjects[Index]) do
    begin
      Result := FMemoryObject.GetObject(ObjectClassName, ObjectID);
      ObjectRef := Result
    end;
end;

procedure TMemoryObjectRelation.SetObjects(Index: Integer; const Value:
  TMemoryObject);
begin
  with TObjectReference(FObjects[Index]) do
  if ObjectRef <> Value then
  begin
    ObjectRef := Value;
    if Value <> nil then
    begin
      ObjectID := Value.ObjectId;
      ObjectClassName := Value.ObjectClassName;
    end else
    begin
      ObjectID := 0;
      ObjectClassName := '';
    end
  end;
end;

function TMemoryObjectRelation.GetObjectClassNames(Index: Integer): String;
begin
  Result := TObjectReference(FObjects[Index]).ObjectClassName
end;

function TMemoryObjectRelation.GetObjectIDs(Index: Integer): TOID;
begin
  Result := TObjectReference(FObjects[Index]).ObjectID
end;

procedure TMemoryObjectRelation.SetObjectIDs(Index: Integer; const Value: TOID);
begin
  with TObjectReference(FObjects[Index]) do
  if Value <> ObjectID then
  begin
    ObjectID := Value;
    ObjectRef := nil;
    if Value = 0 then ObjectClassName := '';
  end;
end;

function TMemoryObjectRelation.GetAsObjectID: TOID;
begin
  if FObjects.Count > 0 then
    Result := ObjectIDs[0]
  else
    Result := 0
end;

procedure TMemoryObjectRelation.SetAsObjectID(const Value: TOID);
begin
  Clear;
  if Value <> 0 then AddObject(Value, DefaultObjectClassName)
end;

function TMemoryObjectRelation.GetAsObject: TMemoryObject;
begin
  if FObjects.Count > 0 then
    Result := Objects[0]
  else
    Result := nil
end;

procedure TMemoryObjectRelation.SetAsObject(const Value: TMemoryObject);
begin
  Clear;
  if Value <> nil then AddObject(Value)
end;

function TMemoryObjectRelation.AddObject(AObjectID: TOID; AObjectClassName:
  String = ''): Integer;
begin
  if AObjectClassName = '' then AObjectClassName := DefaultObjectClassName;
  Result := FObjects.Add(TObjectReference.Create(AObjectClassName, AObjectID));
end;

function TMemoryObjectRelation.AddObject(AObject: TMemoryObject): Integer;
begin
  Result := FObjects.Add(TObjectReference.Create(AObject))
end;

procedure TMemoryObjectRelation.DeleteObject(Index: Integer);
begin
  FObjects.Delete(Index);
end;

procedure TMemoryObjectRelation.Clear;
begin
  FObjects.Clear;
end;

{{
    .    ,  
 .
}
function TMemoryObjectRelation.GetObjectByID(AObjectId: TOID): TMemoryObject;
begin
  Result := FindObjectByID(AObjectId);
  if Result = nil then
    raise EObjectWithIDNotFound.CreateFmt(
      '    (%d)  ', [AObjectId]);
end;

{{
    .    ,  
 nil.
}
function TMemoryObjectRelation.FindObjectByID(AObjectId: TOID): TMemoryObject;
var
  I: Integer;
begin
  for I := 0 to ObjectCount - 1 do
    if ObjectIDs[I] = AObjectId then
    begin
      Result := Objects[I];
      Exit;
    end;
  Result := nil;
end;

{{
       AttributeName.   
,    nil.
}
function TMemoryObjectRelation.FindObjectByAttributeValue(AttributeName:
  String; AttributeValue: Variant): TMemoryObject;
var
  I: Integer;
begin
  for I := 0 to ObjectCount - 1 do
    if Objects[I].AttributeValues[AttributeName] = AttributeValue then
    begin
      Result := Objects[I];
      Exit;
    end;
  Result := nil;
end;

function TMemoryObjectRelation.IndexOfObject(AObject: TMemoryObject): Integer;
var
  I: Integer;
begin
  for I := 0 to ObjectCount - 1 do
    if TObjectReference(FObjects[I]).ObjectRef = AObject then
    begin
      Result := I;
      Exit;
    end;
  Result := -1;
end;

function TMemoryObjectRelation.IndexOfObject(AObjectID: TOID; AObjectClassName:
  String): Integer;
var
  I: Integer;
begin
  for I := 0 to ObjectCount - 1 do
    if (TObjectReference(FObjects[I]).ObjectID = AObjectID) and
       (TObjectReference(FObjects[I]).ObjectClassName = AObjectClassName) then
    begin
      Result := I;
      Exit;
    end;
  Result := -1;
end;

end.
