unit ovModelStorage;

interface

uses SysUtils, Classes, DB, ovComponents, ovIObjects, ovModel;

type
  TModelStorageItem = class(TCollectionItem)
  end;

  TDataTypeStorageItem = class(TModelStorageItem)
  private
    FBaseType: String;
    FDescription: String;
    FName: String;
    FSize: Integer;
  published
    property Name: String read FName write FName;
    property BaseType: String read FBaseType write FBaseType;
    property Size: Integer read FSize write FSize default 0;
    property Description: String read FDescription write FDescription;
  end;

  TIndexStorageItem = class(TModelStorageItem)
  private
    FName: String;
    FAttributes: String;
    FOptions: TIndexOptions;
  public
  published
    property Name: String read FName write FName;
    property Attributes: String read FAttributes write FAttributes;
    property Options: TIndexOptions read FOptions write FOptions default [];
  end;

  TObjectStorageItem = class(TModelStorageItem)
  private
    FAttributes: TCollection;
    FIndexes: TCollection;
    FName: String;
    FTableName: String;
    FIsAbstract: Boolean;
    FId: Integer;
    FDescription: String;
    FVisibleForUser: Boolean;
    procedure ReadAttributes(Reader: TReader);
    procedure WriteAttributes(Writer: TWriter);
    procedure ReadIndexes(Reader: TReader);
    procedure WriteIndexes(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
  published
    property Name: String read FName write FName;
    property TableName: String read FTableName write FTableName;
    property Id: Integer read FId write FId;
    property IsAbstract: Boolean read FIsAbstract write FIsAbstract default False;
    property Description: String read FDescription write FDescription;
    property VisibleForUser: Boolean read FVisibleForUser write FVisibleForUser
      default True;
  end;

  TAttributeStorageItem = class(TModelStorageItem)
  private
    FFieldName: String;
    FName: String;
    FSize: Integer;
    FTypeName: String;
    FCalculated: Boolean;
    FRequired: Boolean;
    FRequiredMsg: String;
    FDescription: String;
    FPersistent: Boolean;
    FVisibleForUser: Boolean;
  public
    constructor Create(Collection: TCollection); override;
  published
    property Name: String read FName write FName;
    property FieldName: String read FFieldName write FFieldName;
    property Size: Integer read FSize write FSize default 0;
    property TypeName: String read FTypeName write FTypeName;
    property Persistent: Boolean read FPersistent write FPersistent default True;
    property Calculated: Boolean read FCalculated write FCalculated default False;
    property RequiredMsg: String read FRequiredMsg write FRequiredMsg;
    property Required: Boolean read FRequired write FRequired default False;
    property Description: String read FDescription write FDescription;
    property VisibleForUser: Boolean read FVisibleForUser write FVisibleForUser
      default True;
  end;

  TRelationStorageItem = class(TModelStorageItem)
  private
    FFieldName: String;
    FId: Integer;
    FMaxCount: Integer;
    FMinCount: Integer;
    FName: String;
    FRelationType: TRelationType;
    FSourceObject: String;
    FTargetObject: String;
    FRelationTableName: String;
    FTargetNameSpace: String;
    FRequired: Boolean;
    FRequiredMsg: String;
    FDescription: String;
    FDeleteMode: TDeleteMode;
    FVisibleForUser: Boolean;
    FBiDirectional: Boolean;
    FDerived: Boolean;
    function IsCountStored: Boolean;
  public
    constructor Create(Collection: TCollection); override;
  published
    property Name: String read FName write FName;
    property FieldName: String read FFieldName write FFieldName;
    property Id: Integer read FId write FId;
    property RelationType: TRelationType read FRelationType write FRelationType;
    property SourceObject: String read FSourceObject write FSourceObject;
    property TargetObject: String read FTargetObject write FTargetObject;
    property TargetNameSpace: String read FTargetNameSpace write FTargetNameSpace;
    property RelationTableName: String read FRelationTableName write
      FRelationTableName;
    property MaxCount: Integer read FMaxCount write FMaxCount stored
      IsCountStored;
    property MinCount: Integer read FMinCount write FMinCount stored
      IsCountStored;
    property Required: Boolean read FRequired write FRequired default False;
    property RequiredMsg: String read FRequiredMsg write FRequiredMsg;
    property Description: String read FDescription write FDescription;
    property DeleteMode: TDeleteMode read FDeleteMode write FDeleteMode default
      dmBlockDelete;
    property VisibleForUser: Boolean read FVisibleForUser write FVisibleForUser
      default True;
    property BiDirectional: Boolean read FBiDirectional write FBiDirectional
      default False;
    property Derived: Boolean read FDerived write FDerived default False;
  end;

  TovModelStorage = class(TovCustomModelStorage)
  private
    FDataTypes: TCollection;
    FObjects: TCollection;
    FRelations: TCollection;
    FFileName: String;
    procedure ReadObjects(Reader: TReader);
    procedure ReadRelations(Reader: TReader);
    procedure WriteObjects(Writer: TWriter);
    procedure WriteRelations(Writer: TWriter);
    procedure ReadDataTypes(Reader: TReader);
    procedure WriteDataTypes(Writer: TWriter);
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure WriteAttribute(AModelObject: TModelObject; Attribute:
      TModelAttribute); override;
    procedure WriteObject(AObject: TModelObject); override;
    procedure WriteRelation(Relation: TModelRelation); override;
    function GetAttributeCount(Index: Integer): Integer; override;
    function GetObjectCount: Integer; override;
    procedure ReadAttribute(AObject: TModelObject; ObjectIndex, AttributeIndex:
      Integer); override;
    function ReadObject(Model: TDomainModel; Index: Integer): TModelObject;
      override;
    function GetRelationCount: Integer; override;
    procedure ReadRelation(Model: TDomainModel; Index: Integer); override;
    procedure Clear;
    procedure LoadObjects(Model: TDomainModel); override;
    procedure LoadRelations(Model: TDomainModel); override;
    function GetDataTypeCount: Integer; override;
    function ReadDataType(Model: TDomainModel; Index: Integer): TModelDataType;
      override;
    procedure WriteDataType(AModelDataType: TModelDataType); override;
    procedure WriteIndex(AModelObject: TModelObject; Index: TModelObjectIndex);
      override;
    function GetIndexCount(Index: Integer): Integer; override;
    procedure ReadIndex(AObject: TModelObject; ObjectIndex, Index: Integer);
      override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SaveModel(Model: TDomainModel; SaveAllObjects: Boolean = False); 
      override;
  published
    property FileName: String read FFileName write FFileName;
  end;

  TovDBModelStorage = class(TovCustomModelStorage)
  private
    FDatabaseAdapter: TovCustomDatabaseAdapter;
    FObjectTableName: String;
    FRelationTableName: String;
    FAttributeTableName: String;
    FNameFieldSize: Integer;
    FObjectTable: TDataSet;
    FAttributeTable: TDataSet;
    FRelationTable: TDataSet;
    procedure SetDatabaseAdapter(Value: TovCustomDatabaseAdapter);
    function GetAdapter: IDatabaseAdapter;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CreateTables;
    function GetAttributeCount(Index: Integer): Integer; override;
    function GetObjectCount: Integer; override;
    function GetRelationCount: Integer; override;
    procedure ReadAttribute(AObject: TModelObject; ObjectIndex, AttributeIndex: 
      Integer); override;
    function ReadObject(Model: TDomainModel; Index: Integer): TModelObject; 
      override;
    procedure ReadRelation(Model: TDomainModel; Index: Integer); override;
    procedure WriteAttribute(AModelObject: TModelObject; Attribute:
      TModelAttribute); override;
    procedure WriteObject(AObject: TModelObject); override;
    procedure WriteRelation(Relation: TModelRelation); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure LoadModel(Model: TDomainModel); override;
    procedure SaveModel(Model: TDomainModel; SaveAllObjects: Boolean = False); 
      override;
    property Adapter: IDatabaseAdapter read GetAdapter;
  published
    property DatabaseAdapter: TovCustomDatabaseAdapter read FDatabaseAdapter write
      SetDatabaseAdapter;
    property ObjectTableName: String read FObjectTableName write FObjectTableName;
    property RelationTableName: String read FRelationTableName write
      FRelationTableName;
    property AttributeTableName: String read FAttributeTableName write
      FAttributeTableName;
    property NameFieldSize: Integer read FNameFieldSize write FNameFieldSize
      default 50;
  end;

  TovXMLModelStorage = class(TovCustomModelStorage)
  end;

implementation

uses TypInfo, ovUtils;

{
****************************** TObjectStorageItem ******************************
}

constructor TObjectStorageItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FVisibleForUser := True;
  FAttributes := TCollection.Create(TAttributeStorageItem);
  FIndexes := TCollection.Create(TIndexStorageItem);
end;

destructor TObjectStorageItem.Destroy;
begin
  FIndexes.Free;
  FAttributes.Free;
  inherited Destroy;
end;

procedure TObjectStorageItem.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('Attributes', ReadAttributes, WriteAttributes, FAttributes.Count > 0);
  Filer.DefineProperty('Indexes', ReadIndexes, WriteIndexes, FIndexes.Count > 0);
end;

procedure TObjectStorageItem.ReadAttributes(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FAttributes)
end;

procedure TObjectStorageItem.WriteAttributes(Writer: TWriter);
begin
  Writer.WriteCollection(FAttributes)
end;

procedure TObjectStorageItem.ReadIndexes(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FIndexes)
end;

procedure TObjectStorageItem.WriteIndexes(Writer: TWriter);
begin
  Writer.WriteCollection(FIndexes)
end;

{
***************************** TAttributeStorageItem ****************************
}

constructor TAttributeStorageItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FPersistent := True;
  FVisibleForUser := True;
end;

{
***************************** TRelationStorageItem *****************************
}

constructor TRelationStorageItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FDeleteMode := dmBlockDelete;
  FVisibleForUser := True;
end;

function TRelationStorageItem.IsCountStored: Boolean;
begin
  Result := RelationType <> rtGeneralization;
end;

{
******************************* TovModelStorage ********************************
}
{<<Category>>: No category}
constructor TovModelStorage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FObjects := TCollection.Create(TObjectStorageItem);
  FRelations := TCollection.Create(TRelationStorageItem);
  FDataTypes := TCollection.Create(TDataTypeStorageItem);
end;

destructor TovModelStorage.Destroy;
begin
  FreeAndNil(FDataTypes);
  FreeAndNil(FObjects);
  FreeAndNil(FRelations);
  inherited Destroy;
end;

procedure TovModelStorage.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('DataTypes', ReadDataTypes, WriteDataTypes, FDataTypes.Count > 0);
  Filer.DefineProperty('Objects', ReadObjects, WriteObjects, FObjects.Count > 0);
  Filer.DefineProperty('Relations', ReadRelations, WriteRelations, FRelations.Count > 0);
end;

procedure TovModelStorage.ReadDataTypes(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FDataTypes)
end;

procedure TovModelStorage.ReadObjects(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FObjects)
end;

procedure TovModelStorage.ReadRelations(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(FRelations)
end;

procedure TovModelStorage.WriteDataTypes(Writer: TWriter);
begin
  Writer.WriteCollection(FDataTypes)
end;

procedure TovModelStorage.WriteObjects(Writer: TWriter);
begin
  Writer.WriteCollection(FObjects)
end;

procedure TovModelStorage.WriteRelations(Writer: TWriter);
begin
  Writer.WriteCollection(FRelations)
end;

procedure TovModelStorage.Clear;
begin
  FDataTypes.Clear;
  FObjects.Clear;
  FRelations.Clear;
end;

{***************************************************************************** }
{                                                              }
{ **************************************************************************** }

procedure TovModelStorage.SaveModel(Model: TDomainModel; SaveAllObjects:
  Boolean = False);
begin
  Clear;
  inherited;
  if FileName <> '' then
  begin
    SaveComponentToFile(Self, FileName);
    Clear;
  end;
end;

procedure TovModelStorage.WriteDataType(AModelDataType: TModelDataType);
var
  Item: TDataTypeStorageItem;
begin
  Item := FDataTypes.Add as TDataTypeStorageItem;
  with AModelDataType do begin
    Item.Name := Name;
    Item.BaseType := BaseType;
    Item.Size := Size;
    Item.Description := Description;
  end
end;

procedure TovModelStorage.WriteObject(AObject: TModelObject);
var
  Item: TObjectStorageItem;
begin
  Item := FObjects.Add as TObjectStorageItem;
  with AObject do begin
    Item.Id := Id;
    Item.Name := Name;
    Item.TableName := TableName;
    Item.IsAbstract := IsAbstract;
    Item.Description := Description;
    Item.VisibleForUser := VisibleForUser;
  end
end;

procedure TovModelStorage.WriteAttribute(AModelObject: TModelObject; Attribute:
  TModelAttribute);
var
  Item: TAttributeStorageItem;
begin
  Item := TObjectStorageItem(FObjects.Items[FObjects.Count - 1]).FAttributes.Add as TAttributeStorageItem;
  with Attribute do
  begin
    Item.Name := Name;
    Item.TypeName := TypeName;
    Item.Size := Size;
    Item.FieldName := FieldName;
    Item.Persistent := Persistent;
    Item.Calculated := Calculated;
    Item.Required := Required;
    Item.RequiredMsg := RequiredMsg;
    Item.Description := Description;
    Item.VisibleForUser := VisibleForUser;
  end;
end;

procedure TovModelStorage.WriteRelation(Relation: TModelRelation);
var
  Item: TRelationStorageItem;
begin
  Item := FRelations.Add as TRelationStorageItem;
  with Relation do begin
    Item.Name := Name;
    Item.FieldName := FieldName;
    Item.Id := Id;
    Item.RelationType := RelationType;
    Item.MinCount := TargetMinCount;
    Item.MaxCount := TargetMaxCount;
    Item.SourceObject := SourceObject.Name;
    Item.TargetObject := TargetObjectName;
    if TargetObjectNameSpace <> NameSpace then
      Item.TargetNameSpace := TargetObjectNameSpace;
    Item.RelationTableName := TableName;
    Item.Required := Required;
    Item.RequiredMsg := RequiredMsg;
    Item.Description := Description;
    Item.DeleteMode := DeleteMode;
    Item.VisibleForUser := VisibleForUser;
    Item.BiDirectional := BiDirectional;
    Item.Derived := Derived;
  end
end;

procedure TovModelStorage.WriteIndex(AModelObject: TModelObject; Index:
  TModelObjectIndex);
var
  Item: TIndexStorageItem;
begin
  Item := TObjectStorageItem(FObjects.Items[FObjects.Count - 1]).FIndexes.Add as TIndexStorageItem;
  with Index do
  begin
    Item.Name := Name;
    Item.Attributes := Attributes;
    Item.Options := Options;
  end
end;

{***************************************************************************** }
{                                                                }
{ **************************************************************************** }

procedure TovModelStorage.LoadObjects(Model: TDomainModel);
begin
  if FileName <> '' then LoadComponentFromFile(Self, FileName);
  inherited;
end;

procedure TovModelStorage.LoadRelations(Model: TDomainModel);
begin
  inherited;
  if FileName <> '' then Clear;
end;

function TovModelStorage.GetDataTypeCount: Integer;
begin
  Result := FDataTypes.Count
end;

function TovModelStorage.ReadDataType(Model: TDomainModel; Index: Integer):
  TModelDataType;
begin
  with TDataTypeStorageItem(FDataTypes.Items[Index]) do begin
    Result := Model.AddDataType(Name, BaseType, Size);
    Result.Namespace := Namespace;
    Result.Description := Description;
  end
end;

function TovModelStorage.GetObjectCount: Integer;
begin
  Result := FObjects.Count
end;

function TovModelStorage.ReadObject(Model: TDomainModel; Index: Integer):
  TModelObject;
begin
  with TObjectStorageItem(FObjects.Items[Index]) do
  begin
    Result := Model.AddObject(Name, NameSpace, Id);
    Result.TableName := TableName;
    Result.IsAbstract := IsAbstract;
    Result.Description := Description;
    Result.VisibleForUser := VisibleForUser;
  end
end;

function TovModelStorage.GetAttributeCount(Index: Integer): Integer;
begin
  Result := TObjectStorageItem(FObjects.Items[Index]).FAttributes.Count
end;

procedure TovModelStorage.ReadAttribute(AObject: TModelObject; ObjectIndex,
  AttributeIndex: Integer);
var
  Item: TObjectStorageItem;
  Attribute: TModelAttribute;
begin
  Item := TObjectStorageItem(FObjects.Items[ObjectIndex]);
  with TAttributeStorageItem(Item.FAttributes.Items[AttributeIndex]) do
  begin
    Attribute := AObject.AddAttribute(Name, TypeName, Size);
    Attribute.FieldName := FieldName;
    Attribute.Calculated := Calculated;
    Attribute.Persistent := Persistent;
    Attribute.Required := Required;
    Attribute.RequiredMsg := RequiredMsg;
    Attribute.Description := Description;
    Attribute.VisibleForUser := VisibleForUser;
  end;
end;

function TovModelStorage.GetRelationCount: Integer;
begin
  Result := FRelations.Count
end;

procedure TovModelStorage.ReadRelation(Model: TDomainModel; Index: Integer);
var
  Relation: TModelRelation;
  ATargetNameSpace: String;
  Source: TModelObject;
begin
  with TRelationStorageItem(FRelations.Items[Index]) do
  begin
    Source := Model.GetObject(MakeFullObjectName(SourceObject, NameSpace));
    if TargetNameSpace = '' then
      ATargetNameSpace := NameSpace
    else
      ATargetNameSpace := TargetNameSpace;

    if RelationType = rtGeneralization then
    begin
      Relation := Source.AddGeneralization(MakeFullObjectName(TargetObject, ATargetNameSpace));
      Relation.FieldName := FieldName;
    end else begin
      Relation := Source.AddRelation(Name, MakeFullObjectName(TargetObject, ATargetNameSpace),
             RelationType, Id);
      Relation.FieldName := FieldName;
      Relation.TableName := RelationTableName;
      Relation.TargetMinCount := MinCount;
      Relation.TargetMaxCount := MaxCount;
      Relation.Required := Required;
      Relation.RequiredMsg := RequiredMsg;
      Relation.Description := Description;
      Relation.DeleteMode := DeleteMode;
      Relation.VisibleForUser := VisibleForUser;
      Relation.BiDirectional := BiDirectional;
      Relation.Derived := Derived;
    end
  end
end;

function TovModelStorage.GetIndexCount(Index: Integer): Integer;
begin
  Result := TObjectStorageItem(FObjects.Items[Index]).FIndexes.Count
end;

procedure TovModelStorage.ReadIndex(AObject: TModelObject; ObjectIndex, Index:
  Integer);
var
  Item: TObjectStorageItem;
begin
  Item := TObjectStorageItem(FObjects.Items[ObjectIndex]);
  with TIndexStorageItem(Item.FIndexes.Items[Index]) do
    AObject.AddIndex(Name, Attributes, Options);
end;

{
****************************** TovDBModelStorage *******************************
}

constructor TovDBModelStorage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FNameFieldSize := 50;
  FObjectTableName := 'Objects';
  FRelationTableName := 'Relations';
  FAttributeTableName := 'Attributes';
end;

procedure TovDBModelStorage.SetDatabaseAdapter(Value: TovCustomDatabaseAdapter);
begin
  if FDatabaseAdapter <> Value then
  begin
    FDatabaseAdapter := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovDBModelStorage.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if AComponent = DatabaseAdapter then DatabaseAdapter := nil
end;

function TovDBModelStorage.GetAdapter: IDatabaseAdapter;
begin
  Result := DatabaseAdapter;
end;

{ ************************   ********************************* }

procedure TovDBModelStorage.SaveModel(Model: TDomainModel; SaveAllObjects:
  Boolean = False);
begin
  if Not Assigned(DatabaseAdapter) then
    raise Exception.Create('   DatabaseAdapter');

  CreateTables;

  FObjectTable := nil;
  FAttributeTable := nil;
  FRelationTable := nil;
  try
    FObjectTable := Adapter.OpenTable(ObjectTableName, False);
    FAttributeTable := Adapter.OpenTable(AttributeTableName, False);
    FRelationTable := Adapter.OpenTable(RelationTableName, False);

    with FObjectTable do while Not IsEmpty do Delete;
    with FAttributeTable do while Not IsEmpty do Delete;
    with FRelationTable do while Not IsEmpty do Delete;

    inherited;

  finally
    FreeAndNil(FObjectTable);
    FreeAndNil(FAttributeTable);
    FreeAndNil(FRelationTable)
  end;
end;

{{
   ObjectTableName, RelationTableName, AttributeTableName
    DatabaseAdapter.
}
procedure TovDBModelStorage.CreateTables;
var
  ADataSet: TDataSet;
  IndexDefs: TIndexDefs;
begin
  ADataSet := nil;
  IndexDefs := nil;
  try
    ADataSet := TDataSet.Create(nil);
    IndexDefs := TIndexDefs.Create(ADataSet);

    with ADataSet.FieldDefs do
    begin
      //   Objects
      if Not Adapter.TableExists(ObjectTableName) then
      begin
        ADataSet.FieldDefs.Clear;
        Add('Id', ftInteger);
        Add('Name', ftString, NameFieldSize);
        Add('TableName', ftString, NameFieldSize);
        Add('IsAbstract', ftBoolean);

        //  
        IndexDefs.Clear;
        IndexDefs.Add('Primary', 'Id;Name', [ixPrimary]);
        Adapter.CreateTable(ObjectTableName, ADataSet.FieldDefs, IndexDefs);
      end;

      //   Attributes
      if Not Adapter.TableExists(ObjectTableName) then
      begin
        ADataSet.FieldDefs.Clear;
        Add('Name', ftString, NameFieldSize);
        Add('ObjectId', ftInteger);
        Add('ObjectName', ftString, NameFieldSize);
        Add('TypeName', ftString, NameFieldSize);
        Add('Size', ftInteger);
        Add('FieldName', ftString, NameFieldSize);

        //  
        IndexDefs.Clear;
        IndexDefs.Add('Primary', 'AttributeId;Name', [ixPrimary]);
        IndexDefs.Add('Secondary', 'ObjectId', []);
        Adapter.CreateTable(AttributeTableName, ADataSet.FieldDefs, IndexDefs);
      end;

      //   Relations
      if Not Adapter.TableExists(ObjectTableName) then
      begin
        ADataSet.FieldDefs.Clear;
        Add('RelationId', ftInteger);
        Add('Name', ftString, NameFieldSize);
        Add('RelationType', ftString, 20);
        Add('MinCount', ftInteger);
        Add('MaxCount', ftInteger);
        Add('SourceObject', ftString, NameFieldSize);
        Add('TargetObject', ftString, NameFieldSize);
        Add('TargetObjectModelName', ftString, NameFieldSize);
        Add('FieldName', ftString, NameFieldSize);
        Add('TableName', ftString, NameFieldSize);

        //  
        IndexDefs.Clear;
        IndexDefs.Add('Primary', 'AttributeId;Name', [ixPrimary]);
        Adapter.CreateTable(RelationTableName, ADataSet.FieldDefs, IndexDefs);
      end

    end
  finally
    IndexDefs.Free;
    ADataSet.Free
  end;
end;

procedure TovDBModelStorage.WriteObject(AObject: TModelObject);
begin
  with FObjectTable, AObject do
  begin
    Append;
    FieldByName('Id').AsInteger := Id;
    FieldByName('Name').AsString := Name;
    FieldByName('TableName').AsString := TableName;
    FieldByName('IsAbstract').AsBoolean := IsAbstract;
    Post;
  end
end;

procedure TovDBModelStorage.WriteAttribute(AModelObject: TModelObject;
  Attribute: TModelAttribute);
begin
  with FAttributeTable, Attribute do
  begin
    Append;
    FieldByName('Name').AsString := Name;
    FieldByName('ObjectId').AsInteger := FObjectTable.FieldByName('Id').AsInteger;
    FieldByName('ObjectName').AsString := FObjectTable.FieldByName('Name').AsString;
    FieldByName('TypeName').AsString := TypeName;
    FieldByName('Size').AsInteger := Size;
    FieldByName('FieldName').AsString := FieldName;
    Post;
  end;
end;

procedure TovDBModelStorage.WriteRelation(Relation: TModelRelation);
begin
  with FRelationTable, Relation do
  begin
    Append;
    FieldByName('Name').AsString := Name;
    FieldByName('FieldName').AsString := FieldName; //   
    FieldByName('RelationId').AsInteger := Id;
    FieldByName('RelationType').AsString := GetEnumName(TypeInfo(TRelationType), Integer(RelationType));
    FieldByName('MinCount').AsInteger := TargetMinCount;
    FieldByName('MaxCount').AsInteger := TargetMaxCount;
    FieldByName('SourceObject').AsString := SourceObject.Name;
    FieldByName('TargetObject').AsString := TargetObjectName;
    if TargetObjectNameSpace <> NameSpace then
      FieldByName('TargetNameSpace').AsString := TargetObjectNameSpace;
    FieldByName('TableName').AsString := RelationTableName;
    Post;
  end
end;

{ **************************   ********************************* }

procedure TovDBModelStorage.LoadModel(Model: TDomainModel);
begin
  if Not Assigned(DatabaseAdapter) then
    raise Exception.Create('   DatabaseAdapter');

  FObjectTable := nil;
  FAttributeTable := nil;
  FRelationTable := nil;
  try
    FObjectTable := Adapter.OpenTable(ObjectTableName, True);
    FRelationTable := Adapter.OpenTable(RelationTableName, True);
    inherited;
  finally
    FreeAndNil(FObjectTable);
    FreeAndNil(FAttributeTable);
    FreeAndNil(FRelationTable)
  end;
end;

function TovDBModelStorage.GetObjectCount: Integer;
begin
  Result := FObjectTable.RecordCount
end;

function TovDBModelStorage.ReadObject(Model: TDomainModel; Index: Integer):
  TModelObject;
begin
  with FObjectTable do begin
    FreeAndNil(FAttributeTable);
    FAttributeTable := Adapter.OpenQuery(Format('SELECT * FROM %s WHERE ObjectId = %d',
      [AttributeTableName, FieldByName('Id').AsInteger]), nil, False);

    Result := Model.AddObject(FieldByName('Name').AsString, NameSpace,
      FieldByName('Id').AsInteger);
    Result.TableName := FieldByName('TableName').AsString;
    Result.IsAbstract := FieldByName('IsAbstract').AsBoolean;
    Next;
  end
end;

function TovDBModelStorage.GetAttributeCount(Index: Integer): Integer;
begin
  Result := FAttributeTable.RecordCount
end;

procedure TovDBModelStorage.ReadAttribute(AObject: TModelObject; ObjectIndex,
  AttributeIndex: Integer);
begin
  with FAttributeTable do
  begin
    AObject.AddAttribute(FieldByName('Name').AsString,
      FieldByName('TypeName').AsString, FieldByName('Size').AsInteger).FieldName := FieldByName('FieldName').AsString;
    Next;
  end
end;

function TovDBModelStorage.GetRelationCount: Integer;
begin
  Result := FRelationTable.RecordCount
end;

procedure TovDBModelStorage.ReadRelation(Model: TDomainModel; Index: Integer);
var
  Relation: TModelRelation;
  Source: TModelObject;
  RelType: TRelationType;
begin
  with FRelationTable do
  begin
    //  ,    
    Source := Model.GetObject(FieldByName('SourceObject').AsString);
    RelType := TRelationType(GetEnumValue(TypeInfo(TRelationType),
       FieldByName('RelationType').AsString));
    if RelType = rtGeneralization then
    begin
      Relation := Source.AddGeneralization(FieldByName('TargetObject').AsString);
      Relation.FieldName := FieldByName('FieldName').AsString
    end else
    with Source.AddRelation(FieldByName('Name').AsString,
      FieldByName('TargetObject').AsString,
      RelType, FieldByName('RelationId').AsInteger) do
    begin
      FieldName := FieldByName('FieldName').AsString;
      RelationTableName := FieldByName('TableName').AsString;
      TargetMinCount := FieldByName('MinCount').AsInteger;
      TargetMaxCount := FieldByName('MaxCount').AsInteger;
    end;
    Next;
  end
end;

end.
