unit ovObjectInspector;

interface

uses Classes, Controls, ComCtrls, ovIObjects, ovObjects, ovModel;

type
  TovCustomObjectInspector = class(TCustomTreeView)
  private
    FObjects: IovObjectList;
    FRelationImageIndex: Integer;
    FObjectImageIndex: Integer;
    FAttributeImageIndex: Integer;
    FChildObjectImageIndex: Integer;
    function GetObjects(Index: Integer): IovObject;
    function GetObjectCount: Integer;
    procedure ShowChildNodes(Node: TTreeNode);
    function WasExpanded(Node: TTreeNode): Boolean;
    procedure ShowChildObjectNodes(AParentNode: TTreeNode; AObject: IovObject;
      ARelationName: String);
    function GetSelectedObject: IovObject;
  protected
    procedure ShowObjectsInternal; dynamic;
    procedure ShowObjectInternal(ObjectIndex: Integer); virtual;
    function ObjectToText(AObject: IovObject): String; virtual;
    function CanExpand(Node: TTreeNode): Boolean; override;
    function ObjectAttributeToText(AObject: IovObject; AttributeName: String):
      String; virtual;
    procedure ShowObjectNodes(AParentNode: TTreeNode; AObject: IovObject); overload;
      virtual;
    function ObjectRelationToText(AObject: IovObject; RelationName: String): String;
      virtual;
    function ChildObjectToText(AObject: IovObject; RelationName: String;
      ChildIndex: Integer): String; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    procedure ShowObject(AObject: IovObject);
    procedure ShowObjects(AObjects: IovObjectList);
    property Objects[Index: Integer]: IovObject read GetObjects;
    property ObjectCount: Integer read GetObjectCount;
    property RelationImageIndex: Integer read FRelationImageIndex write
      FRelationImageIndex;
    property ObjectImageIndex: Integer read FObjectImageIndex write
      FObjectImageIndex;
    property AttributeImageIndex: Integer read FAttributeImageIndex write
      FAttributeImageIndex;
    property ChildObjectImageIndex: Integer read FChildObjectImageIndex write
      FChildObjectImageIndex;
    property SelectedObject: IovObject read GetSelectedObject;
  end;

  TovObjectInspector = class(TovCustomObjectInspector)
  published
    property Align;
    property Anchors;
    property AutoExpand;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind default bkNone;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property ChangeDelay;
    property Color;
    property Ctl3D;
    property Constraints;
    property DragKind;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property HotTrack;
    property Images;
    property Indent;
    property MultiSelect;
    property MultiSelectStyle;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RightClickSelect;
    property RowSelect;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property ShowRoot;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property ToolTips;
    property Visible;
    property OnAddition;
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnCompare;
    property OnContextPopup;
    property OnCreateNodeClass;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnDblClick;
    property OnDeletion;
    property OnDragDrop;
    property OnDragOver;
    property OnEdited;
    property OnEditing;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnExpanding;
    property OnExpanded;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;

    property ObjectImageIndex;
    property ChildObjectImageIndex;
    property AttributeImageIndex;
    property RelationImageIndex;
  end;

implementation

uses SysUtils;

type
  TNodeData = class
  private
    FObject: IovObject;
    FRelationName: String;
    FIsRootObject: Boolean;
  public
    constructor Create(AObject: IovObject); overload;
    constructor Create(AObject: IovObject; ARelationName: String); overload;
    class function CompareCode(Node: TTreeNode): Integer;
    property IsRootObject: Boolean read FIsRootObject;
  end;

{ TNodeData }

constructor TNodeData.Create(AObject: IovObject);
begin
  inherited Create;
  FObject := AObject;
  FIsRootObject := True;
end;

constructor TNodeData.Create(AObject: IovObject; ARelationName: String);
begin
  inherited Create;
  FObject := AObject;
  FRelationName := ARelationName;
end;

class function TNodeData.CompareCode(Node: TTreeNode): Integer;
begin
  if Assigned(Node.Data) then Result := 0
  else Result := 1
end;

{ TovCustomObjectInspector }

constructor TovCustomObjectInspector.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ReadOnly := True;
  FObjects := NewObjectList;
end;

destructor TovCustomObjectInspector.Destroy;
begin
  Clear;
  inherited Destroy;
end;

{ **************************   ****************************** }

function TovCustomObjectInspector.GetObjects(Index: Integer): IovObject;
begin
  Result := FObjects[Index]
end;

function TovCustomObjectInspector.GetObjectCount: Integer;
begin
  Result := FObjects.Count;
end;

procedure TovCustomObjectInspector.Clear;
var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    if Assigned(Items[I].Data) then begin
      TObject(Items[I].Data).Free;
      Items[I].Data := nil;
    end;
  Items.Clear;
  FObjects.Clear;
end;

procedure TovCustomObjectInspector.ShowObject(AObject: IovObject);
begin
  Clear;
  if Assigned(AObject) then
  begin
    FObjects.Add(AObject);
    ShowObjectsInternal;
  end
end;

procedure TovCustomObjectInspector.ShowObjects(AObjects: IovObjectList);
var
  I: Integer;
begin
  Clear;
  for I := 0 to AObjects.Count - 1 do
    FObjects.Add(AObjects[I]);
  ShowObjectsInternal;
end;

function TovCustomObjectInspector.GetSelectedObject: IovObject;
var
  NodeData: TNodeData;
begin
  NodeData := TObject(Selected.Data) as TNodeData;
  Result := NodeData.FObject;
end;

{ **************************   **************************** }

function InspectorTreeViewSort(Node1, Node2: TTreeNode; lParam: Integer): Integer; stdcall;
begin
  //     -  
  if TNodeData.CompareCode(Node1) = TNodeData.CompareCode(Node2) then Result := 0
  else if TNodeData.CompareCode(Node1) > TNodeData.CompareCode(Node2)
  then Result := -1 else Result := 1;

  if Result = 0 then  //    
    if Node1.Text = Node2.Text then Result := 0
    else if Node1.Text < Node2.Text
    then Result := -1 else Result := 1
end;

procedure TovCustomObjectInspector.ShowObjectsInternal;
var
  I: Integer;
begin
  Items.BeginUpdate;
  try
    Items.Clear;
    for I := 0 to ObjectCount - 1 do ShowObjectInternal(I);
    CustomSort(@InspectorTreeViewSort, 0);
    if Items.Count > 0 then Selected := Items[0];
  finally
    Items.EndUpdate
  end;
end;

procedure TovCustomObjectInspector.ShowObjectInternal(ObjectIndex: Integer);
var
  ObjectNode: TTreeNode;
begin
  ObjectNode := Items.Add(nil, ObjectToText(Objects[ObjectIndex]));
  with ObjectNode do begin
    Data := TNodeData.Create(Objects[ObjectIndex]);
    ImageIndex := ObjectImageIndex;
    SelectedIndex := ObjectImageIndex;
  end;
  Items.AddChild(ObjectNode, ''); //    ,    
                                  //  +
end;

function TovCustomObjectInspector.CanExpand(Node: TTreeNode): Boolean;
begin
  Result := inherited CanExpand(Node);
  if Result and Not WasExpanded(Node) then ShowChildNodes(Node)
end;

function TovCustomObjectInspector.WasExpanded(Node: TTreeNode): Boolean;
begin
  Result := (Node.Count > 0) and ((Node[0].Text <> '') or (Node[0].Data <> nil));
end;

{{
     Node.  Node   
,       .
}
procedure TovCustomObjectInspector.ShowChildNodes(Node: TTreeNode);
var
  NodeData: TNodeData;
begin
  Items.BeginUpdate;
  try
    Node[0].Delete; //   
    NodeData := TObject(Node.Data) as TNodeData;
    if NodeData.IsRootObject then
      ShowObjectNodes(Node, NodeData.FObject)
    else
      ShowChildObjectNodes(Node, NodeData.FObject, NodeData.FRelationName);
    CustomSort(@InspectorTreeViewSort, 0);
  finally
    Items.EndUpdate
  end;
end;

procedure TovCustomObjectInspector.ShowObjectNodes(AParentNode: TTreeNode; 
  AObject: IovObject);
var
  I: Integer;
  ModelObject: TModelObject;
  Node: TTreeNode;
begin
  ModelObject := AObject.ParentClass.ObjectSpace.Model.GetObject(AObject.ClassName);

  //     
  for I := 0 to ModelObject.AttributeCount - 1 do
  begin
    Node := Items.AddChild(AParentNode, ObjectAttributeToText(AObject, ModelObject.Attributes[I].Name));
    with Node do begin
      ImageIndex := AttributeImageIndex;
      SelectedIndex := AttributeImageIndex;
    end;
  end;

  //     , ..   
  for I := 0 to ModelObject.RelationCount - 1 do
  if AObject[ModelObject.Relations[I].Name].Count > 0 then
  begin
    Node := Items.AddChild(AParentNode,
      ObjectRelationToText(AObject, ModelObject.Relations[I].Name));
    with Node do begin
      Data := TNodeData.Create(AObject, ModelObject.Relations[I].Name);
      ImageIndex := RelationImageIndex;
      SelectedIndex := RelationImageIndex;
    end;
    Items.AddChild(Node, ''); //    ,    
                                    //  +
  end;
end;

procedure TovCustomObjectInspector.ShowChildObjectNodes(AParentNode: TTreeNode; 
  AObject: IovObject; ARelationName: String);
var
  I: Integer;
  Node: TTreeNode;
begin
  for I := 0 to AObject[ARelationName].Count - 1 do
  begin
    Node := Items.AddChild(AParentNode,
      ChildObjectToText(AObject, ARelationName, I));
    with Node do begin
      Data := TNodeData.Create(AObject[ARelationName].AsObjects[I]);
      ImageIndex := ChildObjectImageIndex;
      SelectedIndex := ChildObjectImageIndex;
    end;
    Items.AddChild(Node, ''); //    ,    
                                    //  +
  end;
end;

{ ******************     ********** }

function TovCustomObjectInspector.ObjectToText(AObject: IovObject): String;
begin
  with AObject do
    Result := ClassName + ':' + IntToStr(ObjectId)
end;

function TovCustomObjectInspector.ObjectAttributeToText(AObject: IovObject;
  AttributeName: String): String;
begin
  if AObject[AttributeName].IsBlob then
    Result := AttributeName+ ':(blob)'
  else
    Result := AttributeName+ ':' + AObject[AttributeName].AsString
end;

function TovCustomObjectInspector.ObjectRelationToText(AObject: IovObject; 
  RelationName: String): String;
begin
  Result := RelationName+ ' (' + IntToStr(AObject[RelationName].Count) + ')'
end;

function TovCustomObjectInspector.ChildObjectToText(AObject: IovObject;
  RelationName: String; ChildIndex: Integer): String;
begin
  with AObject[RelationName].AsObjects[ChildIndex] do
    Result := ClassName + ':' + IntToStr(ObjectId)
end;

end.
