unit ovQueryManager;

interface

uses SysUtils, Classes, Forms, Variants, DB, ovIObjects, ovTypes, Contnrs;

type
  TovQueryManager = class;

  TPendingChange = class(TObject)
  protected
  public
    procedure Apply(QueryManager: TovQueryManager); virtual; abstract;
    procedure Merge(APendingChanges: TObjectList); virtual; abstract;
  end;

  TovQueryManager = class(TObject)
  private
    FDatabaseAdapter: IDatabaseAdapter;
    FQuerySyntaxParams: TQuerySyntaxParams;
    FUpdateCount: Integer;
    FPendingChanges: TObjectList;
    procedure PrepareQuery(var Query: String);
    function GetIsUpdating: Boolean;
    procedure ApplyPendingChanges;
    procedure AddPendingChange(APendingChange: TPendingChange);
    procedure UpdateRecordInternal(TableName, KeyField: String; KeyValue: Integer;
      const AFieldValues: TFieldValues);
  protected
  public
    constructor Create(ADatabaseAdapter: IDatabaseAdapter);
    function OpenQuery(ASQLQuery: String; MasterDataSource: TDataSource = nil;
       LiveDataSet: Boolean = False): TDataSet;
    procedure DeleteRecords(TableName: String; const FieldValues: TFieldValues);
      overload;
    function SelectRecords(TableName: String; const SelectFieldNames: array of
      String; const WhereFieldValues: TFieldValues; Filter: String = ''): TDataSet;
    function CountRecords(TableName, FieldName: String; const WhereFieldValues:
      TFieldValues; Filter: String = ''): Integer;
    function SelectRecord(TableName, KeyField: String; KeyValue: Integer; ReadOnly:
      Boolean): TDataSet;
    function AddRecord(TableName, KeyField: String; const FieldValues:
      TFieldValues): Integer;
    procedure UpdateRecord(TableName, KeyField: String; KeyValue: Integer; const
      FieldValues: TFieldValues);
    {{
            KeyField, 
    KeyValue.
    }
    procedure DeleteRecord(TableName: String; KeyField: String; KeyValue: Integer);
      overload;
    property DatabaseAdapter: IDatabaseAdapter read FDatabaseAdapter;

    procedure StartUpdating;
    procedure EndUpdating;
    destructor Destroy; override;
    procedure CancelUpdating;
    procedure ExecQuery(ASQLQuery: String);
    procedure UpdateRecords(TableName: String; const NewFieldValues: TFieldValues;
      WhereExpression: String);
    procedure DeleteRecords(TableName, WhereExpression: String); overload;
    property IsUpdating: Boolean read GetIsUpdating;
  end;

procedure WriteToLog(Lines: String);
procedure WriteMsgToLog(Msg: String);

var
  WriteToLogEnabled: Boolean = False;
  LastSQL: String; //     SQL- (      - )

implementation

uses Windows, ovConsts, ovUtils;

type
  {{
       UpdateRecord.
  }
  TUpdateRecordPendingChange = class(TPendingChange)
  private
    FTableName: String;
    FKeyField: String;
    FKeyValue: Integer;
    FFieldValues: TFieldValues;
  public
    constructor Create(TableName, KeyField: String; KeyValue: Integer; const
      FieldValues: TFieldValues);
    procedure Apply(QueryManager: TovQueryManager); override;
    procedure Merge(APendingChanges: TObjectList); override;
    property TableName: String read FTableName;
    property KeyField: String read FKeyField;
    property KeyValue: Integer read FKeyValue;
    property FieldValues: TFieldValues read FFieldValues;
  end;

constructor TUpdateRecordPendingChange.Create(TableName, KeyField: String;
  KeyValue: Integer; const FieldValues: TFieldValues);
begin
  FTableName := TableName;
  FKeyField := KeyField;
  FKeyValue := KeyValue;
  FFieldValues := FieldValues;
end;

{{
         
       ,       
       .
}
procedure TUpdateRecordPendingChange.Merge(APendingChanges: TObjectList);

  {{
        , ..  
        ,     
  ,   , ..    
   (  )    .
  }
  procedure MergeFieldValue(const FieldValue: TFieldValue);
  var
    I: Integer;
    CurrCount: Integer;
  begin
    CurrCount := High(FieldValues) + 1;
    for I := 0 to CurrCount - 1 do
      if FieldValues[I].FieldName = FieldValue.FieldName then Exit;
    SetLength(FFieldValues, CurrCount+1);
    FFieldValues[CurrCount] := FieldValue;
  end;

var
  I, J: Integer;
  APendingChange: TUpdateRecordPendingChange;
begin
  //       
  for I := APendingChanges.Count - 1 downto 0 do
  if APendingChanges.Items[I] is TUpdateRecordPendingChange then
  begin
    APendingChange := TUpdateRecordPendingChange(APendingChanges.Items[I]);
    if (APendingChange.TableName = TableName) and
      (APendingChange.KeyField = KeyField) and (APendingChange.KeyValue = KeyValue)
    then begin
      //        
      with APendingChange do
        for J := Low(FieldValues) to High(FieldValues) do
          MergeFieldValue(FieldValues[J]);
      APendingChanges.Delete(I);
      Exit;
    end;
  end;
  //     ,     -  
  //     APendingChanges   
end;

{{
     TableName   KeyField,
  KeyValue.
}
procedure TUpdateRecordPendingChange.Apply(QueryManager: TovQueryManager);
begin
  QueryManager.UpdateRecordInternal(TableName, KeyField, KeyValue, FieldValues);
end;

{***************************************************************************** }
{                                                      }
{ **************************************************************************** }

procedure AddValuesExpression(ADatabaseAdapter: IDatabaseAdapter; var Query:
  String; const FieldValues: TFieldValues);
var
  I: Integer;
  Fields, Values: String;
begin
  if Length(FieldValues) > 0 then begin
    Fields := '';
    Values := '';
    for I := Low(FieldValues) to High(FieldValues) do
    with FieldValues[I] do
    begin
      Fields := Fields + '[' + FieldName + '], ';
      Values := Values + ADatabaseAdapter.ValueToSQLString(Value, FieldType, FieldSize) + ', ';
    end;
    Delete(Fields, Length(Fields) - 1, 2);
    Delete(Values, Length(Values) - 1, 2);
    Query := Query + Format(' (%s) VALUES (%s)', [Fields, Values]);
  end;
end;

procedure AddSetExpression(var Query: String; const
  SetExpression: String);
begin
  Query := Query + ' SET ' + SetExpression;
end;

function GetFieldList(const FieldNames: array of String): String;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Length(FieldNames) - 1 do
    Result := Result + ', [' + FieldNames[I] + ']';
  if Result <> '' then Delete(Result, 1, 2);
end;

procedure AddWhereExpression(var Query: String; const
  WhereExpression: String);
begin
  if WhereExpression <> '' then Query := Query + ' WHERE ' + WhereExpression;
end;

function MakeValue(const Value: Variant): String;
begin
  case VarType(Value) of
    varInteger : Result := IntToStr(Value);
    varString  : Result := '"' + Value + '"';
    varNull    : Result := 'Null';
    else raise Exception.Create('  ')
  end
end;

function MakeListExpression(ADatabaseAdapter: IDatabaseAdapter; const
  FieldValues: TFieldValues; const Separator, NullOperator: String;
  SkipBlobValues: Boolean = True): String;

  function MakeCondition(const AFieldValue: TFieldValue): String;
  begin
    with AFieldValue do
      if VarIsNull(Value)
        then Result := Format('[%s] %s Null', [FieldName, NullOperator])
        else Result := Format('[%s] = %s',
          [FieldName, ADatabaseAdapter.ValueToSQLString(Value, FieldType, FieldSize)])
  end;

var
  I: Integer;
begin
  Result := '';
  if Length(FieldValues) > 0 then
  begin
    for I := Low(FieldValues) to High(FieldValues) do
      if Not SkipBlobValues or
        ((FieldValues[I].FieldType <> ftBlob) and (FieldValues[I].FieldType <> ftMemo))
      then
        Result := Result + MakeCondition(FieldValues[I]) + Separator;
    Delete(Result, Length(Result) - Length(Separator) + 1,
      Length(Separator)); //   Separator
  end
end;

{***************************************************************************** }
{                                                    }
{ **************************************************************************** }

var _Clock: Cardinal;

procedure StartClock;
begin
  _Clock := GetTickCount;
end;

function StopClock(): Cardinal;
begin
  Result := GetTickCount - _Clock
end;

procedure WriteToLog(Lines: String);
var
  F: TextFile;
  FileName: String;
begin
  FileName := ChangeFileExt(Application.ExeName, '.ovl'); { Object View Log file }
  AssignFile(F, FileName);
  if FileExists(FileName)
    then Append(F)
    else Rewrite(F);
  try
    WriteLn(F, Lines);
  finally
    CloseFile(F);
  end;
end;

procedure WriteQueryToLog(Query: String; ElapsedTime: Cardinal);
var
  Msg: String;
begin
  Msg := #13#10 + Format('%s  : %d ms', [DateTimeToStr(Now), ElapsedTime]) + #13#10 + Query;
  WriteToLog(Msg);
end;

procedure WriteMsgToLog(Msg: String);
begin
  WriteToLog(#13#10 + DateTimeToStr(Now) + #13#10 + Msg);
end;

{ TovQueryManager }

constructor TovQueryManager.Create(ADatabaseAdapter: IDatabaseAdapter);
begin
  inherited Create;
  FDatabaseAdapter := ADatabaseAdapter;
  //       
  FQuerySyntaxParams := DatabaseAdapter.GetQuerySyntaxParams;
  FPendingChanges := TObjectList.Create;
end;

destructor TovQueryManager.Destroy;
begin
  FreeAndNil(FPendingChanges);
  inherited Destroy;
end;

{***************************************************************************** }
{                                                     }
{ **************************************************************************** }

{{
   ,   FieldValues,  
     KeyField  
(    ).
}
function TovQueryManager.AddRecord(TableName, KeyField: String; const
  FieldValues: TFieldValues): Integer;
var
  Query: String;
begin
  Query := Format('INSERT INTO {%s}', [TableName]);
  AddValuesExpression(DatabaseAdapter, Query, FieldValues);
  ExecQuery(Query);
  Result := DatabaseAdapter.GetLastKeyValue(TableName, KeyField);
end;

{{
        KeyField, 
KeyValue.
}
procedure TovQueryManager.DeleteRecord(TableName: String; KeyField: String;
  KeyValue: Integer);
var
  Query: String;
begin
  Query := Format('DELETE FROM {%s} WHERE [%s] = %d', [TableName, KeyField, KeyValue]);
  ExecQuery(Query);
end;

{{
        ,   
 FieldValues.      ,  
       ,     .
}
procedure TovQueryManager.DeleteRecords(TableName: String; const FieldValues:
  TFieldValues);
var
  Query: String;
begin
  Query := Format('DELETE FROM {%s}', [TableName]);
  AddWhereExpression(Query, MakeListExpression(DatabaseAdapter,
    FieldValues, ' AND ', 'is'));
  ExecQuery(Query);
end;

procedure TovQueryManager.DeleteRecords(TableName, WhereExpression: String);
var
  Query: String;
begin
  Query := Format('DELETE FROM {%s}', [TableName]);
  AddWhereExpression(Query, WhereExpression);
  ExecQuery(Query);
end;

{{
        ,   
 WhereFieldValues.      ,  
       .    ,
   SelectFieldNames.
}
function TovQueryManager.SelectRecords(TableName: String; const
  SelectFieldNames: array of String; const WhereFieldValues: TFieldValues;
  Filter: String = ''): TDataSet;
var
  Query: String;
begin
  Query := Format('SELECT %s FROM {%s}', [GetFieldList(SelectFieldNames), TableName]);
  AddWhereExpression(Query, CombineSubStr(
    MakeListExpression(DatabaseAdapter, WhereFieldValues, ' AND ', 'is'), ' AND ', Filter));
  PrepareQuery(Query);
  Result := DatabaseAdapter.OpenQuery(Query);
end;

{{
  ,      TableName 
 KeyField,  KeyValue.
}
function TovQueryManager.SelectRecord(TableName, KeyField: String; KeyValue:
  Integer; ReadOnly: Boolean): TDataSet;
var
  Query: String;
begin
  if WriteToLogEnabled then StartClock;

  if ReadOnly then
  begin
    Query := Format('SELECT * FROM {%s} WHERE [%s] = %d', [TableName, KeyField, KeyValue]);
    PrepareQuery(Query);
    Result := DatabaseAdapter.OpenQuery(Query);
  end else
  begin
    Query := FQuerySyntaxParams.FilterLeftFieldDelimiter + KeyField +
      FQuerySyntaxParams.FilterRightFieldDelimiter + ' = ' + IntToStr(KeyValue);
    Result := DatabaseAdapter.OpenTable(TableName, False, Query);
  end;

  if WriteToLogEnabled then WriteQueryToLog(
    Format('   %s   %s  %d',
      [TableName, KeyField, KeyValue]), StopClock)
end;

{{
  ,   ,   
 WhereFieldValues.
}
function TovQueryManager.CountRecords(TableName, FieldName: String; const
  WhereFieldValues: TFieldValues; Filter: String = ''): Integer;
var
  Query: String;
begin
  Query := Format('SELECT Count([%s]) FROM {%s}', [FieldName, TableName]);
  AddWhereExpression(Query, CombineSubStr(
    MakeListExpression(DatabaseAdapter, WhereFieldValues, ' AND ', 'is'), ' AND ', Filter));
  PrepareQuery(Query);
  with DatabaseAdapter.OpenQuery(Query) do
  try
    Result := Fields[0].AsInteger
  finally
    Free
  end;
end;

{{
         
KeyField,  KeyValue.  FieldValues     ,
           .
   IsUpdating,     
   .
}
procedure TovQueryManager.UpdateRecord(TableName, KeyField: String; KeyValue:
  Integer; const FieldValues: TFieldValues);
begin
  if Not IsUpdating then
    UpdateRecordInternal(TableName, KeyField, KeyValue, FieldValues)
  else
    AddPendingChange(
      TUpdateRecordPendingChange.Create(TableName, KeyField, KeyValue, FieldValues))
end;

{
        
    KeyField,  KeyValue.
}
procedure TovQueryManager.UpdateRecordInternal(TableName, KeyField: String;
  KeyValue: Integer; const AFieldValues: TFieldValues);
var
  Query: String;
  I, BlobCount: Integer;
begin
  BlobCount := 0;
  for I := Low(AFieldValues) to High(AFieldValues) do
    if ((AFieldValues[I].FieldType = ftBlob) or (AFieldValues[I].FieldType = ftMemo)) then
      {   blob  memo  }
      with SelectRecord(TableName, KeyField, KeyValue, False) do
      try
        Inc(BlobCount);
        Edit;
        FieldByName(AFieldValues[I].FieldName).Value := AFieldValues[I].Value;
        Post;
      finally
        Free
      end;
  if Length(AFieldValues) = BlobCount then Exit; {     }

  {  UPDATE-      ( blob-) }
  Query := Format('UPDATE {%s}', [TableName]);
  AddSetExpression(Query, MakeListExpression(DatabaseAdapter, AFieldValues, ' , ', '='));
  AddWhereExpression(Query, Format('[%s] = %d', [KeyField, KeyValue]));
  ExecQuery(Query);
end;

{{
      TableName, 
   ,   NewFieldValues,  
  WhereExpression.
}
procedure TovQueryManager.UpdateRecords(TableName: String; const
  NewFieldValues: TFieldValues; WhereExpression: String);
var
  Query: String;
begin
  Query := Format('UPDATE {%s}', [TableName]);
  AddSetExpression(Query, MakeListExpression(DatabaseAdapter, NewFieldValues, ' , ', '='));
  AddWhereExpression(Query, WhereExpression);
  ExecQuery(Query);
end;

{***************************************************************************** }
{                                                  }
{ **************************************************************************** }

procedure TovQueryManager.ExecQuery(ASQLQuery: String);
begin
  PrepareQuery(ASQLQuery);

  if WriteToLogEnabled then StartClock;

  LastSQL := ASQLQuery; //    SQL-
  try
    DatabaseAdapter.ExecQuery(ASQLQuery);
  except
    on E: Exception do
    begin
      //           
      //     WriteToLogEnabled
      WriteQueryToLog(ASQLQuery + #13'  : '+ E.Message, StopClock);
      raise;
    end
  end;

  if WriteToLogEnabled then WriteQueryToLog(ASQLQuery, StopClock)
end;

function TovQueryManager.OpenQuery(ASQLQuery: String; MasterDataSource:
   TDataSource = nil; LiveDataSet: Boolean = False): TDataSet;
begin
  PrepareQuery(ASQLQuery);

  if WriteToLogEnabled then StartClock;

  LastSQL := ASQLQuery; //    SQL-
  try
    Result := DatabaseAdapter.OpenQuery(ASQLQuery, MasterDataSource, LiveDataSet);
  except
    on E: Exception do
    begin
      //           
      //     WriteToLogEnabled
      WriteQueryToLog(ASQLQuery + #13'  : '+ E.Message, StopClock);
      raise;
    end
  end;

  if WriteToLogEnabled then WriteQueryToLog(ASQLQuery, StopClock)
end;

procedure TovQueryManager.PrepareQuery(var Query: String);

  function ReplaceSubStr(var S: String; Position, DeleteNum:
    Integer; NewText: String): Integer;
  begin
    Delete(S, Position, DeleteNum);
    Insert(NewText, S, Position);
    Result := Length(NewText) - DeleteNum + 1; //   ,
                                               //    
  end;

var
  I, D, L: Integer;
begin
  I := 1;
  L := Length(Query);
  while I <= L do
  begin
    with FQuerySyntaxParams do
    case Query[I] of
      '{': D := ReplaceSubStr(Query, I, 1, LeftTableDelimiter);
      '}': D := ReplaceSubStr(Query, I, 1, RightTableDelimiter);
      '[': D := ReplaceSubStr(Query, I, 1, LeftFieldDelimiter);
      ']': D := ReplaceSubStr(Query, I, 1, RightFieldDelimiter);
      else D := 1;
    end;
    Inc(I, D); Inc(L, D-1);
  end;
end;

{***************************************************************************** }
{                                                       }
{ **************************************************************************** }

{{
   .
}
procedure TovQueryManager.StartUpdating;
begin
  Inc(FUpdateCount)
end;

{{
   .     StartUpdating
  ,       .
}
procedure TovQueryManager.EndUpdating;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then ApplyPendingChanges;
end;

procedure TovQueryManager.CancelUpdating;
begin
  FUpdateCount := 0;
  FPendingChanges.Clear;
end;

function TovQueryManager.GetIsUpdating: Boolean;
begin
  Result := FUpdateCount > 0;
end;

procedure TovQueryManager.AddPendingChange(APendingChange: TPendingChange);
begin
  APendingChange.Merge(FPendingChanges);
  FPendingChanges.Add(APendingChange);
end;

procedure TovQueryManager.ApplyPendingChanges;
var
  I: Integer;
begin
  try
    for I := 0 to FPendingChanges.Count - 1 do
      TPendingChange(FPendingChanges.Items[I]).Apply(Self);
  finally
    {    (  )    }
    FPendingChanges.Clear;
  end;
end;

end.
