unit ovSysOperations;

interface

uses SysUtils, Classes, Contnrs, Forms;

type
  TCloseFormAction = (caDefault, caHide, caFree);

  TSysOperation = class
  private
    FMethodAddr: Pointer;
    FFormClass: TFormClass;
    FFreeAfterUse: Boolean;
    FAutoCreate: Boolean;
    FForms: TComponentList;
    FSingleInstance: Boolean;
    function GetActiveForm: TForm;
    function GetFormCount: Integer;
    function GetForms(Index: Integer): TForm;
  protected
    property FormCount: Integer read GetFormCount;
    property Forms[Index: Integer]: TForm read GetForms;
  public
    constructor Create(AFormClass: TFormClass; AMethodProc: Pointer; AutoCreate:
      Boolean = False; FreeAfterUse: Boolean = True; SingleInstance: Boolean =
      False);
    destructor Destroy; override;
    procedure CreateForm;
    procedure ReleaseForm(CloseAction: TCloseFormAction);
    function GetMethod: TMethod;
    property ActiveForm: TForm read GetActiveForm;
    property FormClass: TFormClass read FFormClass;
    {{
     ,      
      .
    }
    property FreeAfterUse: Boolean read FFreeAfterUse;
    {{
     ,        
      AutoCreateForms.
    }
    property AutoCreate: Boolean read FAutoCreate;
    {{
     ,      .
    }
    property SingleInstance: Boolean read FSingleInstance;
  end;

procedure RegisterSysOperation(AOperationType: SmallInt; AOperationName: String;
  AOperationArgCount: SmallInt; AOperation: TSysOperation);

procedure UnregisterSysOperation(AOperationType: SmallInt; AOperationName:
  String; AOperationArgCount: SmallInt);

function GetSysOperation(AOperationType: SmallInt; AOperationName: String;
  AOperationArgCount: SmallInt): TSysOperation;

procedure AutoCreateForms;

implementation

{ TSysOperation }

constructor TSysOperation.Create(AFormClass: TFormClass; AMethodProc: Pointer;
  AutoCreate: Boolean = False; FreeAfterUse: Boolean = True; SingleInstance:
  Boolean = False);
begin
  inherited Create;
  FMethodAddr := AMethodProc;
  FFormClass := AFormClass;
  FAutoCreate := AutoCreate;
  FFreeAfterUse := FreeAfterUse;
  FSingleInstance := SingleInstance;
  FForms := TComponentList.Create(True);
end;

destructor TSysOperation.Destroy;
begin
  FreeAndNil(FForms);
  inherited Destroy;
end;

function TSysOperation.GetForms(Index: Integer): TForm;
begin
  Result := TForm(FForms[Index]);
end;

function TSysOperation.GetFormCount: Integer;
begin
  Result := FForms.Count;
end;

function TSysOperation.GetActiveForm: TForm;
begin
  if FormCount > 0 then
    Result := Forms[FormCount - 1]
  else
    Result := nil;
end;

procedure TSysOperation.CreateForm;
begin
  if Assigned(FormClass) then
  begin
    if Not Assigned(ActiveForm) or Not SingleInstance then
      FForms.Add(FormClass.Create(Application))
  end;
end;

procedure TSysOperation.ReleaseForm(CloseAction: TCloseFormAction);
begin
  if (CloseAction = caFree) or (CloseAction = caDefault) and
     FreeAfterUse and Assigned(ActiveForm)
  then ActiveForm.Free
end;

function TSysOperation.GetMethod: TMethod;
begin
  Result.Data := ActiveForm;
  Result.Code := FMethodAddr;
end;

{ TRegisteredOperation }

type
  TRegisteredOperation = class
    OperationType: SmallInt;
    OperationName: String;
    ArgCount: SmallInt;
    Operation: TSysOperation;
  end;

var
  RegisteredOperations: TObjectList; //     TRegisteredOperation

function FindSysOperation(AOperationType: SmallInt; AOperationName: String; 
  AOperationArgCount: SmallInt; ExactMatch: Boolean): Integer;
var
  I: Integer;
begin
  //     
  for I := 0 to RegisteredOperations.Count - 1 do
    with TRegisteredOperation(RegisteredOperations[I]) do
      if (OperationType = AOperationType) and (OperationName = AOperationName) and
         (ArgCount = AOperationArgCount) then
      begin
        Result := I;
        Exit;
      end;

  //   
  if Not ExactMatch then
  begin
    //     
    for I := 0 to RegisteredOperations.Count - 1 do
      with TRegisteredOperation(RegisteredOperations[I]) do
        if (OperationType = AOperationType) and (OperationName = AOperationName) and (ArgCount = -1) then
        begin
          Result := I;
          Exit;
        end;

    //     
    for I := 0 to RegisteredOperations.Count - 1 do
      with TRegisteredOperation(RegisteredOperations[I]) do
        if (OperationType = AOperationType) and (OperationName = '') then
        begin
          Result := I;
          Exit;
        end;
  end;

  Result := -1;
end;

procedure RegisterSysOperation(AOperationType: SmallInt; AOperationName: String;
  AOperationArgCount: SmallInt; AOperation: TSysOperation);
var
  I: Integer;
  Item: TRegisteredOperation;
begin
  try
    I := FindSysOperation(AOperationType, AOperationName, AOperationArgCount, True);
    if I >= 0 then raise Exception.CreateFmt(' "%s"     ', [AOperationName]);

    Item := TRegisteredOperation.Create;
    with Item do
    begin
      OperationType := AOperationType;
      OperationName := AOperationName;
      ArgCount := AOperationArgCount;
      Operation := AOperation;
    end;
    RegisteredOperations.Add(Item);
  except
    AOperation.Free; raise;
  end;
end;

procedure UnregisterSysOperation(AOperationType: SmallInt; AOperationName:
  String; AOperationArgCount: SmallInt);
var
  I: Integer;
begin
  I := FindSysOperation(AOperationType, AOperationName, AOperationArgCount, True);
  if I >= 0 then RegisteredOperations.Delete(I);
end;

function GetSysOperation(AOperationType: SmallInt; AOperationName: String;
  AOperationArgCount: SmallInt): TSysOperation;
var
  I: Integer;
begin
  I := FindSysOperation(AOperationType, AOperationName, AOperationArgCount, False);
  if I < 0 then raise Exception.CreateFmt(' "%s"     ', [AOperationName]);

  Result := TRegisteredOperation(RegisteredOperations[I]).Operation
end;

{{
   ,   AutoCreate.
}
procedure AutoCreateForms;
var
  I: Integer;
begin
  for I := 0 to RegisteredOperations.Count - 1 do
    with TRegisteredOperation(RegisteredOperations[I]) do
      if Operation.AutoCreate then Operation.CreateForm
end;

initialization
  RegisteredOperations := TObjectList.Create(True);
finalization
  RegisteredOperations.Free;
end.
