{$INCLUDE oView.inc}
unit ovXMLSchema;

interface

uses XMLIntf, ovXMLImportExport, ovModel;

type
  TovXMLSchemaGenerator = class(TovCustomXMLSchemaGenerator)
  private
    FCurrentNamespace: String;
    FSchemaFileName: String;
    procedure WriteRootElement;
    procedure WriteObjectAttributes(AModelObject: TModelObject; AParentNode: 
      IXMLNode);
    procedure WriteObjectRelations(AModelObject: TModelObject; AParentNode: 
      IXMLNode);
    procedure WriteObject(AModelObject: TModelObject);
    procedure WriteDomainTypes;
    procedure GenerateNamespaceSchema(Namespace, FileName: String);
    function GetNamespaceSchemaFileName(Namespace: String): String;
    function GetXMLSchemaType(TypeName: String): String;
  protected
    procedure DoExecute; override;
    procedure BeforeExecute; override;
    property SchemaFileName: String read FSchemaFileName write FSchemaFileName;
  public
    procedure Reset; override;
    procedure GenerateSchema(FileName: String); override;
  published
  end;

implementation

uses SysUtils, TypInfo, DB, Classes, ovUtils, ovComponents;

function GetXMLSchemaCardinality(Cardinality: Integer): String;
begin
  if Cardinality = Unlimited then
    Result := 'unbounded'
  else
    Result := IntToStr(Cardinality);
end;

function HasBlobAttributes(AModelObject: TModelObject): Boolean;
var
  I: Integer;
begin
  for I := 0 to AModelObject.AttributeCount - 1 do
    if AModelObject.Attributes[I].IsBlob then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

{ TovXMLSchemaGenerator }

{{
      FileName.    
   .   FileName  
          .
}
procedure TovXMLSchemaGenerator.GenerateSchema(FileName: String);
var
  I: Integer;
begin
  FSchemaFileName := FileName;
  with Model.DomainModel do
  begin
    for I := 0 to NameSpaces.Count - 1 do
      GenerateNamespaceSchema(NameSpaces[I], GetNamespaceSchemaFileName(NameSpaces[I]));
    //      ,       
    if NameSpaces.IndexOf('') < 0 then
      GenerateNamespaceSchema('', FileName);
  end
end;

function TovXMLSchemaGenerator.GetNamespaceSchemaFileName(Namespace: String):
  String;
begin
  if NameSpace <> '' then
    { TODO :        .  . }
    Result := ExtractFilePath(SchemaFileName) + NameSpace + '.xsd'
  else
    Result := SchemaFileName
end;

procedure TovXMLSchemaGenerator.GenerateNamespaceSchema(Namespace, FileName:
  String);
begin
  FCurrentNamespace := Namespace;
  Reset;
  Execute;
  SaveToFile(FileName);
end;

procedure TovXMLSchemaGenerator.BeforeExecute;
begin
  inherited;
  if Not Assigned(XMLDocument) then Reset;
end;

procedure TovXMLSchemaGenerator.Reset;
var
  I: Integer;
  SchemaNode, ImportNode: IXMLNode;
begin
  inherited;
  SchemaNode := XMLDocument.AddChild('xs:schema');
  SchemaNode.Attributes['xmlns:xs'] := 'http://www.w3.org/2001/XMLSchema';

  if FCurrentNamespace <> '' then
  begin
    SchemaNode.Attributes['targetNamespace'] := FCurrentNamespace;
    SchemaNode.Attributes['elementFormDefault'] := 'qualified';
    SchemaNode.Attributes['xmlns:'+GetValidXMLName(FCurrentNamespace)] := FCurrentNamespace;
  end;

  //     
  with Model.DomainModel do
    for I := 0 to NameSpaces.Count - 1 do
      if (NameSpaces[I] <> '') and (NameSpaces[I] <> FCurrentNamespace) then
      begin
        ImportNode := SchemaNode.AddChild('xs:import');
        ImportNode.Attributes['namespace'] := NameSpaces[I];
        ImportNode.Attributes['schemaLocation'] := GetNamespaceSchemaFileName(NameSpaces[I]);

        SchemaNode.Attributes['xmlns:'+GetValidXMLName(NameSpaces[I])] := NameSpaces[I];
      end
end;

procedure TovXMLSchemaGenerator.DoExecute;
var
  I: Integer;
begin
  WriteDomainTypes;
  if FCurrentNamespace = '' then WriteRootElement;
  with DomainModel do
  begin
    for I := 0 to ObjectCount - 1 do
    if Objects[I].NameSpace = FCurrentNamespace then
    begin
      WriteObject(Objects[I]);
      DoProgress(psMain, I, ObjectCount);
    end;
  end
end;

{{
    .
}
procedure TovXMLSchemaGenerator.WriteDomainTypes;
var
  TypeSize: Integer;
  TypeName: String;
  I: Integer;
  Node: IXMLNode;
begin
  with Model.DomainModel do
  for I := 0 to DataTypeCount - 1 do
  begin
    Node := XMLDocument.DocumentElement.AddChild('xs:simpleType');
    Node.Attributes['name'] := DataTypes[I].Name;
    Node := Node.AddChild('xs:restriction');
    Node.Attributes['base'] := GetXMLSchemaType(DataTypes[I].BaseType);
    if TypeSize > 0 then
    begin
      Node := Node.AddChild('xs:maxLength');
      Node.Attributes['value'] := IntToStr(DataTypes[I].Size);
    end
  end;

  //     XML- Memo-
  Node := XMLDocument.DocumentElement.AddChild('xs:complexType');
  Node.Attributes['name'] := 'XMLAttributeMemoType';
  Node := Node.AddChild('xs:simpleContent');
  Node := Node.AddChild('xs:extension');
  Node.Attributes['base'] := 'xs:string';
  Node := Node.AddChild('xs:attribute');
  Node.Attributes['name'] := 'type';
  Node.Attributes['fixed'] := 'memo';
end;

{{
    ,  
    .
}
procedure TovXMLSchemaGenerator.WriteRootElement;
var
  I: Integer;
  Node: IXMLNode;
begin
  Node := XMLDocument.DocumentElement.AddChild('xs:element');
  Node.Attributes['name'] := RootObjectsTagName;
  Node := Node.AddChild('xs:complexType');
  Node := Node.AddChild('xs:choice');
  Node.Attributes['maxOccurs'] := 'unbounded';
  with DomainModel do
  for I := 0 to ObjectCount - 1 do
  with Node.AddChild('xs:element') do
  begin
    Attributes['ref'] := GetValidXMLName(Objects[I].FullName);
    Attributes['minOccurs'] := '0';
  end
end;

procedure TovXMLSchemaGenerator.WriteObject(AModelObject: TModelObject);
var
  ObjectNode, TypeNode: IXMLNode;
begin
  ObjectNode := XMLDocument.DocumentElement.AddChild('xs:element');
  ObjectNode.Attributes['name'] := GetValidXMLName(AModelObject.Name);

  TypeNode := ObjectNode.AddChild('xs:complexType');
  WriteObjectRelations(AModelObject, TypeNode);
  WriteObjectAttributes(AModelObject, TypeNode);
end;

{{
      Blob- (  
XML-      )
}
procedure TovXMLSchemaGenerator.WriteObjectRelations(AModelObject: TModelObject;
  AParentNode: IXMLNode);
var
  Relation: TModelRelation;
  RelationNode, Node: IXMLNode;
  Attribute: TModelAttribute;
  AttributeNode: IXMLNode;
  I: Integer;
begin
  if (AModelObject.RelationCount > 0) or HasBlobAttributes(AModelObject) then
  begin
    AParentNode := AParentNode.AddChild('xs:all');
    //   
    for I := 0 to AModelObject.RelationCount - 1 do
    begin
      Relation := AModelObject.Relations[I];
      RelationNode := AParentNode.AddChild('xs:element');
      RelationNode.Attributes['name'] := GetValidXMLName(Relation.Name);
      RelationNode.Attributes['minOccurs'] := '0';
      RelationNode.Attributes['maxOccurs'] := '1';
      Node := RelationNode.AddChild('xs:complexType');
      Node := Node.AddChild('xs:sequence');
      Node.Attributes['minOccurs'] :=
        GetXMLSchemaCardinality(Relation.TargetMinCount);
      Node.Attributes['maxOccurs'] :=
        GetXMLSchemaCardinality(Relation.TargetMaxCount);
      Node := Node.AddChild('xs:element');
      Node.Attributes['ref'] := GetValidXMLName(Relation.TargetObject.FullName);
    end;

    //   Blob-
    for I := 0 to AModelObject.AttributeCount - 1 do
    begin
      Attribute := AModelObject.Attributes[I];
      if Attribute.IsBlob then
      begin
        AttributeNode := AParentNode.AddChild('xs:element');
        AttributeNode.Attributes['name'] := GetValidXMLName(Attribute.Name);
        AttributeNode.Attributes['type'] :=
          GetValidXMLName(MakeFullObjectName('XMLAttributeMemoType', AModelObject.NameSpace));
        AttributeNode.Attributes['minOccurs'] := '0';
      end
    end

  end;
end;

procedure TovXMLSchemaGenerator.WriteObjectAttributes(AModelObject:
  TModelObject; AParentNode: IXMLNode);
var
  Attribute: TModelAttribute;
  AttributeNode: IXMLNode;
  I: Integer;
begin
  AttributeNode := AParentNode.AddChild('xs:attribute');
  AttributeNode.Attributes['name'] := ObjectIdAttributeName;
  AttributeNode.Attributes['type'] := 'xs:integer';
  AttributeNode.Attributes['use'] := 'optional';

  for I := 0 to AModelObject.AttributeCount - 1 do
  begin
    Attribute := AModelObject.Attributes[I];
    if Not Attribute.IsBlob then
    begin
      AttributeNode := AParentNode.AddChild('xs:attribute');
      AttributeNode.Attributes['name'] := GetValidXMLName(Attribute.Name);
      AttributeNode.Attributes['type'] := GetXMLSchemaType(Attribute.TypeName);
      AttributeNode.Attributes['use'] := 'optional';
    end
  end
end;

{{
     XML-.  TypeName 
     TFieldType.
}
function TovXMLSchemaGenerator.GetXMLSchemaType(TypeName: String): String;
const
{$IFNDEF Delphi2006}
  XMLDataTypes: array [TFieldType] of String =
  ('unknown', 'xs:string', 'xs:int', 'xs:int', 'xs:nonNegativeInteger', 'xs:boolean',
   'xs:double', 'xs:double', 'xs:double', 'xs:string', 'xs:string', 'xs:string',
   'Bytes', 'VarBytes', 'xs:integer', 'xs:hexBinary', 'xs:string', 'xs:hexBinary',
   'Memo', 'Blob', 'Blob', 'Blob', 'Blob',
   'xs:string', 'xs:string', 'xs:integer', 'Blob', 'Blob',
   'Blob', 'Blob', 'Blob', 'Blob', 'Blob',
   'Blob', 'Blob', 'xs:string', 'xs:string', 'xs:string');
{$ELSE}
  XMLDataTypes: array [TFieldType] of String =
  ('unknown', 'xs:string', 'xs:int', 'xs:int', 'xs:nonNegativeInteger', 'xs:boolean',
   'xs:double', 'xs:double', 'xs:double', 'xs:string', 'xs:string', 'xs:string',
   'Bytes', 'VarBytes', 'xs:integer', 'xs:hexBinary', 'xs:string', 'xs:hexBinary',
   'Memo', 'Blob', 'Blob', 'Blob', 'Blob',
   'xs:string', 'xs:string', 'xs:integer', 'Blob', 'Blob',
   'Blob', 'Blob', 'Blob', 'Blob', 'Blob',
   'Blob', 'Blob', 'xs:string', 'xs:string', 'xs:string',
   //  Delphi 2006  4 
   'xs:string', 'xs:string', 'xs:string', 'xs:string');
{$ENDIF}
var
  EnumValue: Integer;
begin
  EnumValue := GetEnumValue(TypeInfo(TFieldType), 'ft' + TypeName);
  if EnumValue < 0 then
    Result := GetValidXMLName(FCurrentNamespace) + ':' + TypeName
  else
    Result := XMLDataTypes[TFieldType(EnumValue)]
end;

end.
