unit ovModelSync;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ovModel, ovComponents;

type
  TovUpdatableModelStorage = class(TovCustomModelStorage)
  public
    procedure AddClass(SourceModel: TDomainModel; ClassName: String); virtual;
      abstract;
    procedure AddClassRelations(SourceModel: TDomainModel; ClassName: String); virtual;
      abstract;
    procedure AddAttribute(SourceModel: TDomainModel; ClassName, AttributeName: String); virtual;
      abstract;
    procedure AddRelation(SourceModel: TDomainModel; ClassName, RelationName: String); virtual;
      abstract;
    procedure AddGeneralization(SourceModel: TDomainModel; ClassName, ParentClassName: String); virtual;
      abstract;

    procedure UpdateAttribute(SourceModel: TDomainModel; ClassName, AttributeName: String); virtual;
      abstract;
    procedure UpdateRelation(SourceModel: TDomainModel; ClassName, RelationName: String); virtual;
      abstract;

    procedure DeleteClass(ClassName: String); virtual; abstract;
    procedure DeleteAttribute(ClassName, AttributeName: String); virtual;
      abstract;
    procedure DeleteRelation(ClassName, RelationName: String); virtual;
      abstract;
    procedure DeleteGeneralization(ClassName, ParentClassName: String); virtual;
      abstract;
  end;

  TovModelSynchronizer = class(TComponent)
  private
    FSource: TovCustomModelStorage;
    FTarget: TovCustomModelStorage;
    FSourceList: TStrings;
    FTargetList: TStrings;
    FSourceModel: TDomainModel;
    FTargetModel: TDomainModel;
    FDlgCaption: String;
    procedure SetSource(Value: TovCustomModelStorage);
    procedure SetTarget(Value: TovCustomModelStorage);
    procedure GetClasses(Model: TDomainModel; Classes: TStrings);
    procedure GetAttributes(ModelObject: TModelObject; AttributeList: TStrings);
    procedure GetRelations(ModelObject: TModelObject; RelationList: TStrings);
    procedure GetGeneralizations(ModelObject: TModelObject; GeneralizationList:
      TStrings);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function ShowSyncDialog: Boolean; dynamic;
    function Execute: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {{
         Source  Target.
    }
    procedure AddAttribute(ClassName, AttributeName: String);
    {{
       Source  Target.
    }
    procedure AddClass(ClassName: String);
    {{
          
      TargetClassName  Source  Target.
    }
    procedure AddGeneralization(ClassName, ParentClassName: String);
    {{
         Source  Target.
    }
    procedure AddRelation(ClassName, RelationName: String);
    {{
         Target.
    }
    procedure DeleteAttribute(ClassName, AttributeName: String);
    {{
       Target.
    }
    procedure DeleteClass(ClassName: String);
    {{
          Target.
    }
    procedure DeleteGeneralization(ClassName, ParentClassName: String);
    {{
         Target.
    }
    procedure DeleteRelation(ClassName, RelationName: String);
    procedure CompareAttributes(ClassName: String; AddedAttributes, 
      DeletedAttributes, ExistAttributes: TStrings);
    procedure CompareClasses(AddedClasses, DeletedClasses, ExistClasses: TStrings);
    procedure CompareRelations(ClassName: String; AddedRelations, DeletedRelations, 
      ExistRelations: TStrings);
    {{
          Source  
    Target.
    }
    procedure UpdateAttribute(ClassName, AttributeName: String);
    {{
          Source 
    Target.
    }
    procedure UpdateRelation(ClassName, RelationName: String);
    procedure AddClassRelations(ClassName: String);
    function ImportModel: Boolean;
    function ExportModel: Boolean;
    property SourceModel: TDomainModel read FSourceModel;
    property TargetModel: TDomainModel read FTargetModel;
  published
    procedure CompareGeneralizations(ClassName: String; AddedGeneralizations,
      DeletedGeneralizations, ExistGeneralizations: TStrings);
    property Target: TovCustomModelStorage read FTarget write SetTarget;
    property Source: TovCustomModelStorage read FSource write SetSource;
  end;

implementation

uses ovModelSyncDlg;

{{
      :
List1OnlyItems,    List1,     List2,
List2OnlyItems,    List2,     List1,
CommonItems,      List1  List2.
}
procedure CompareLists(List1, List2,
  List1OnlyItems, List2OnlyItems, CommonItems: TStrings);
var
  I, J: Integer;
begin
  List1OnlyItems.Assign(List1);
  List2OnlyItems.Assign(List2);
  CommonItems.Clear;

  for I := List1OnlyItems.Count - 1 downto 0 do
  begin
    J := List2OnlyItems.IndexOf(List1OnlyItems[I]);
    if J >= 0 then
    begin
      CommonItems.Add(List1OnlyItems[I]);
      List1OnlyItems.Delete(I);
      List2OnlyItems.Delete(J);
    end;
  end;

end;

{
***************************** TovModelSynchronizer *****************************
}
constructor TovModelSynchronizer.Create(AOwner: TComponent);
begin
  inherited;
  FSourceList := TStringList.Create;
  FTargetList := TStringList.Create;
end;

destructor TovModelSynchronizer.Destroy;
begin
  FTargetList.Free;
  FSourceList.Free;
  inherited Destroy;
end;

procedure TovModelSynchronizer.SetSource(Value: TovCustomModelStorage);
begin
  if FSource <> Value then
  begin
    FSource := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovModelSynchronizer.SetTarget(Value: TovCustomModelStorage);
begin
  if FTarget <> Value then
  begin
    FTarget := Value;
    if Value <> nil then Value.FreeNotification(Self);
  end;
end;

procedure TovModelSynchronizer.Notification(AComponent: TComponent; Operation:
  TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if AComponent = Source then Source := nil else
    if AComponent = Target then Target := nil
end;

function TovModelSynchronizer.Execute: Boolean;
begin
  if Not Assigned(Source) then
    raise Exception.Create('   Source');
  if Not Assigned(Target) then
    raise Exception.Create('   Target');

  FSourceModel := TDomainModel.Create;
  FTargetModel := TDomainModel.Create;
  try
    Source.LoadModel(FSourceModel);
    Target.LoadModel(FTargetModel);
    Result := ShowSyncDialog;
    if Result then Target.SaveModel(FTargetModel);
  finally
    FreeAndNil(FTargetModel);
    FreeAndNil(FSourceModel);
  end;
end;

function TovModelSynchronizer.ShowSyncDialog: Boolean;
begin
  Result := SynchronizeModels(Self, FDlgCaption);
end;

function TovModelSynchronizer.ImportModel: Boolean;
var
  SaveSourceNamespace, SaveTargetNamespace: String;
begin
  FDlgCaption := ' ';

  SaveSourceNamespace := Source.NameSpace;
  SaveTargetNamespace := Target.NameSpace;
  //       
  Source.NameSpace := '';
  Target.NameSpace := '';
  try
    Result := Execute;
  finally
    Source.NameSpace := SaveSourceNamespace;
    Target.NameSpace := SaveTargetNamespace;
  end;
end;

function TovModelSynchronizer.ExportModel: Boolean;
var
  SaveSource, SaveTarget: TovCustomModelStorage;
begin
  FDlgCaption := ' ';

  SaveSource := FSource;
  SaveTarget:= FTarget;
  try
    FTarget := FSource;
    FSource := SaveTarget;
    Result := Execute
  finally
    FSource := SaveSource;
    FTarget:= SaveTarget;
  end;
end;

//***********************     ***********************

{{
     
}
procedure TovModelSynchronizer.GetClasses(Model: TDomainModel; Classes:
  TStrings);
var
  I: Integer;
begin
  Classes.Clear;
  for I := 0 to Model.ObjectCount - 1 do
    Classes.Add(Model.Objects[I].Name)
end;

{{
      
}
procedure TovModelSynchronizer.GetAttributes(ModelObject: TModelObject;
  AttributeList: TStrings);
var
  I: Integer;
begin
  AttributeList.Clear;
  with ModelObject do
  for I := 0 to AttributeCount - 1 do
    with Attributes[I] do
      if Not IsInherited then AttributeList.Add(Name)
end;

{{
       ( 
)
}
procedure TovModelSynchronizer.GetRelations(ModelObject: TModelObject;
  RelationList: TStrings);
var
  I: Integer;
begin
  RelationList.Clear;
  with ModelObject do
  for I := 0 to RelationCount - 1 do
    with Relations[I] do
      if Not IsInherited then RelationList.Add(Name)
end;

{{
       
(     ).
}
procedure TovModelSynchronizer.GetGeneralizations(ModelObject: TModelObject;
  GeneralizationList: TStrings);
var
  I: Integer;
begin
  GeneralizationList.Clear;
  with ModelObject do
  for I := 0 to ParentCount - 1 do
    GeneralizationList.Add(Parents[I].Name)
end;

//*****************************   ******************************

{{
 AddedClasses     SourceModel,   
TargetModel.  DeletedClasses     TargetModel,
   SourceModel.  ExistClasses   
   SourceModel   TargetModel.
}
procedure TovModelSynchronizer.CompareClasses(AddedClasses, DeletedClasses,
  ExistClasses: TStrings);
begin
  GetClasses(SourceModel, FSourceList);
  GetClasses(TargetModel, FTargetList);
  CompareLists(FSourceList, FTargetList,
    AddedClasses, DeletedClasses, ExistClasses);
end;

{{
  AddedAttributes    ClassName  SourceModel,
   TargetModel.   DeletedAttributes  
 ClassName  TargetModel,    SourceModel.  
ExistAttributes    ClassName,   
SourceModel   TargetModel.  ClassName      .
}
procedure TovModelSynchronizer.CompareAttributes(ClassName: String;
  AddedAttributes, DeletedAttributes, ExistAttributes: TStrings);
begin
  GetAttributes(SourceModel.GetObject(ClassName), FSourceList);
  GetAttributes(TargetModel.GetObject(ClassName), FTargetList);
  CompareLists(FSourceList, FTargetList,
    AddedAttributes, DeletedAttributes, ExistAttributes);
end;

{{
  AddedRelations    ClassName  SourceModel,
   TargetModel.   DeletedRelations  
 ClassName  TargetModel,    SourceModel.  
ExistRelations    ClassName,   
SourceModel   TargetModel.  ClassName      .
}
procedure TovModelSynchronizer.CompareRelations(ClassName: String;
  AddedRelations, DeletedRelations, ExistRelations: TStrings);
begin
  GetRelations(SourceModel.GetObject(ClassName), FSourceList);
  GetRelations(TargetModel.GetObject(ClassName), FTargetList);
  CompareLists(FSourceList, FTargetList,
    AddedRelations, DeletedRelations, ExistRelations);
end;

{{
  AddedGeneralizations    
ClassName  SourceModel,    TargetModel.  
DeletedGeneralizations     ClassName 
TargetModel,    SourceModel.   ExistGeneralizations
    ClassName, 
  SourceModel   TargetModel.  ClassName  
   .
}
procedure TovModelSynchronizer.CompareGeneralizations(ClassName: String;
  AddedGeneralizations, DeletedGeneralizations, ExistGeneralizations: TStrings);
begin
  GetGeneralizations(SourceModel.GetObject(ClassName), FSourceList);
  GetGeneralizations(TargetModel.GetObject(ClassName), FTargetList);
  CompareLists(FSourceList, FTargetList,
    AddedGeneralizations, DeletedGeneralizations, ExistGeneralizations);
end;

//****************************   ******************************

{{
       SourceModel  TargetModel.
}
procedure TovModelSynchronizer.AddClass(ClassName: String);
var
  TargetObject: TModelObject;
  I: Integer;
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).AddClass(SourceModel, ClassName)

  else
    with SourceModel.GetObject(ClassName) do
    begin
      TargetObject := TargetModel.AddObject(Name, '', Id);
      TargetObject.TableName := TableName;
      TargetObject.Description := Description;
      for I := 0 to AttributeCount - 1 do
        with Attributes[I] do
          if Not IsInherited then
            TargetObject.AddAttribute(Name, TypeName, Size).FieldName := FieldName;
    end;

end;

{{
    ClassName  SourceModel  TargetModel.
}
procedure TovModelSynchronizer.AddClassRelations(ClassName: String);
var
  I: Integer;
  ASourceObject, ATargetObject: TModelObject;
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).AddClassRelations(SourceModel, ClassName)

  else begin
    ASourceObject := SourceModel.GetObject(ClassName);
    ATargetObject := TargetModel.GetObject(ClassName);

    //   
    with ASourceObject do
    for I := 0 to RelationCount - 1 do
      with Relations[I] do
        if Not IsInherited then
          //  ,     
          if TargetModel.FindObject(TargetObjectName) <> nil then
            with ATargetObject.AddRelation(Name,
              TargetObjectName, RelationType, Id) do
            begin
              TargetMinCount := Relations[I].TargetMinCount;
              TargetMaxCount := Relations[I].TargetMaxCount;
            end;

    //   
    with ASourceObject do
    for I := 0 to ParentCount - 1 do
      //  ,     
      if TargetModel.FindObject(ASourceObject.Parents[I].Name) <> nil
        then ATargetObject.AddGeneralization(ASourceObject.Parents[I].Name)
  end
end;

{{
     SourceModel  TargetModel.
}
procedure TovModelSynchronizer.AddAttribute(ClassName, AttributeName: String);
var
  SourceObject, TargetObject: TModelObject;
  TargetAttribute: TModelAttribute;
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).AddAttribute(SourceModel, ClassName, AttributeName)

  else begin
    SourceObject := SourceModel.GetObject(ClassName);
    TargetObject := TargetModel.GetObject(ClassName);
    with SourceObject.GetAttribute(AttributeName) do
    begin
      TargetAttribute := TargetObject.AddAttribute(Name, TypeName, Size);
      TargetAttribute.FieldName := FieldName;
      TargetAttribute.Description := Description;
    end;
  end
end;

{{
     SourceModel  TargetModel.
}
procedure TovModelSynchronizer.AddRelation(ClassName, RelationName: String);
var
  Relation: TModelRelation;
  ASourceObject, ATargetObject: TModelObject;
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).AddRelation(SourceModel, ClassName, RelationName)

  else begin
    ASourceObject := SourceModel.GetObject(ClassName);
    ATargetObject := TargetModel.GetObject(ClassName);
    with ASourceObject.GetRelation(RelationName) do
    begin
      Relation := ATargetObject.AddRelation(Name, TargetObjectName, RelationType, Id);
      Relation.TargetMinCount := TargetMinCount;
      Relation.TargetMaxCount := TargetMaxCount;
      Relation.TableName := TableName;
      Relation.FieldName := FieldName;
      Relation.Description := Description;
    end;
  end
end;

{{
     
  ParentClassName  SourceModel  TargetModel.
}
procedure TovModelSynchronizer.AddGeneralization(ClassName, ParentClassName:
  String);
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).AddGeneralization(SourceModel, ClassName, ParentClassName)

  else
    TargetModel.GetObject(ClassName).AddGeneralization(ParentClassName);
end;

{{
   TargetModel.
}
procedure TovModelSynchronizer.DeleteClass(ClassName: String);
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).DeleteClass(ClassName)
  else
    TargetModel.GetObject(ClassName).Delete
end;

{{
     TargetModel.
}
procedure TovModelSynchronizer.DeleteAttribute(ClassName, AttributeName: String);
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).DeleteAttribute(ClassName, AttributeName)
  else
    TargetModel.GetObject(ClassName).DeleteAttribute(AttributeName);
end;

{{
     TargetModel.
}
procedure TovModelSynchronizer.DeleteRelation(ClassName, RelationName: String);
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).DeleteRelation(ClassName, RelationName)
  else
    TargetModel.GetObject(ClassName).DeleteRelation(RelationName);
end;

{{
      TargetModel.
}
procedure TovModelSynchronizer.DeleteGeneralization(ClassName, ParentClassName: 
  String);
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).DeleteGeneralization(ClassName, ParentClassName)
  else
    TargetModel.GetObject(ClassName).DeleteGeneralization(ParentClassName);
end;

{{
      SourceModel 
TargetModel.
}
procedure TovModelSynchronizer.UpdateAttribute(ClassName, AttributeName: String);
var
  SourceAttribute: TModelAttribute;
  SourceObject, TargetObject: TModelObject;
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).UpdateAttribute(SourceModel, ClassName, AttributeName)

  else begin
    SourceObject := SourceModel.GetObject(ClassName);
    TargetObject := TargetModel.GetObject(ClassName);
    SourceAttribute := SourceObject.GetAttribute(AttributeName);
    with TargetObject.GetAttribute(AttributeName) do
    begin
      TypeName := SourceAttribute.TypeName;
      Size := SourceAttribute.Size;
      FieldName := SourceAttribute.FieldName;
      Description := SourceAttribute.Description;
    end
  end
end;

{{
      SourceModel 
TargetModel.
}
procedure TovModelSynchronizer.UpdateRelation(ClassName, RelationName: String);
var
  SourceRelation: TModelRelation;
  SourceObject, TargetObject: TModelObject;
begin
  if Target is TovUpdatableModelStorage then
     TovUpdatableModelStorage(Target).UpdateRelation(SourceModel, ClassName, RelationName)

  else begin
    SourceObject := SourceModel.GetObject(ClassName);
    TargetObject := TargetModel.GetObject(ClassName);
    SourceRelation := SourceObject.GetRelation(RelationName);
    with TargetObject.GetRelation(RelationName) do
    begin
      TargetMinCount := SourceRelation.TargetMinCount;
      TargetMaxCount := SourceRelation.TargetMaxCount;
      FieldName := SourceRelation.FieldName;
      TableName := SourceRelation.TableName;
      Description := SourceRelation.Description;
    end;
  end
end;


end.
