unit ovXMLModelStorage;

interface

uses SysUtils, Classes, XMLDoc, XMLIntf, ovComponents, ovModel;

type
  TovXMLModelStorageOption = (msoSaveModelSpace, msoSaveMapping, msoLoadModelSpace, msoLoadMapping);
  TovXMLModelStorageOptions = set of TovXMLModelStorageOption;

  TModelNamespace = String;
  TModelAssociation = TModelRelation;
  TModelGeneralization = TModelRelation;
  TModelAssociationEnd = TModelRelation;

  TovXMLModelStorage = class(TovCustomModelStorage)
  private
    FIsSavingMapping: Boolean;
    FOptions: TovXMLModelStorageOptions;
    FXMLDocument: IXMLDocument;
    FDataTypesNode: IXMLNode;
    procedure SaveNamespace(AParentNode: IXMLNode; AModelStorage:
      TovCustomModelStorage);
    procedure SaveDataType(AParentNode: IXMLNode; ADataType: TModelDataType);
    function AddNode(AParentNode: IXMLNode; AName: String): IXMLNode; overload;
    function AddNode(AParentNode: IXMLNode; AName: String; AValue: OleVariant):
      IXMLNode; overload;
    procedure SaveClass(AParentNode: IXMLNode; AModelClass: TModelObject);
    procedure SaveIndex(AParentNode: IXMLNode; AIndex: TModelObjectIndex);
    procedure SaveAttribute(AParentNode: IXMLNode; AModelAttribute:
      TModelAttribute);
    procedure SaveAssociation(AParentNode: IXMLNode; Association:
      TModelAssociation);
    procedure SaveGeneralization(AParentNode: IXMLNode; Generalization:
      TModelGeneralization);
    procedure SaveModelInternal(ARootNode: IXMLNode; AModel: TovModel);
    procedure SaveAssociationEnd(AParentNode: IXMLNode; AModelRelation:
      TModelRelation; IsSourceEnd: Boolean);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    procedure SaveModel(AModel: TovModel);
    procedure Clear;
    property XMLDocument: IXMLDocument read FXMLDocument;
  published
    property Options: TovXMLModelStorageOptions read FOptions write FOptions
      default [msoSaveModelSpace, msoSaveMapping, msoLoadModelSpace,
      msoLoadMapping];
  end;

  ELoadModelError = class(Exception)
  end;

procedure SaveModelToFile(AModel: TovModel; const AFileName: String;
  SaveModelSpace, SaveMapping: Boolean);
implementation

uses DB, Variants, TypInfo;

resourcestring
  SMappingNodeName = 'Mapping';

procedure SaveModelToFile(AModel: TovModel; const AFileName: String;
  SaveModelSpace, SaveMapping: Boolean);
begin
  with TovXMLModelStorage.Create(nil) do
  try
    if SaveModelSpace then
      Options := Options + [msoSaveModelSpace]
    else
      Options := Options - [msoSaveModelSpace];

    if SaveMapping then
      Options := Options + [msoSaveMapping]
    else
      Options := Options - [msoSaveMapping];

    SaveModel(AModel);
    XMLDocument.SaveToFile(AFileName);
  finally
    Free
  end;
end;

{ TovXMLModelStorage }

constructor TovXMLModelStorage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOptions := [msoSaveModelSpace, msoSaveMapping, msoLoadModelSpace, msoLoadMapping];
  Clear;
end;

procedure TovXMLModelStorage.Clear;
begin
  FXMLDocument := NewXMLDocument;
  FXMLDocument.Encoding := 'UTF-8';
  FXMLDocument.Options := FXMLDocument.Options - [doAttrNull];
  FXMLDocument.AddChild('Model');
  FXMLDocument.DocumentElement.Attributes['SchemaVersion'] := '1.0';
end;

{***************************************************************************** }
{                                                              }
{ **************************************************************************** }

procedure TovXMLModelStorage.SaveModel(AModel: TovModel);
var
  ModelNode: IXMLNode;
begin
  Clear;

  if msoSaveModelSpace in Options then
  begin
    FIsSavingMapping := False;
    ModelNode := AddNode(XMLDocument.DocumentElement, 'ModelSpace');
    SaveModelInternal(ModelNode, AModel);
  end;

  if msoSaveMapping in Options then
  begin
    FIsSavingMapping := True;
    ModelNode := AddNode(XMLDocument.DocumentElement, 'ModelDBMapping');
    SaveModelInternal(ModelNode, AModel);
  end;
end;

procedure TovXMLModelStorage.SaveModelInternal(ARootNode: IXMLNode; AModel:
  TovModel);
var
  I: Integer;
  NamespacesNode, RelationsNode: IXMLNode;
begin
  NamespacesNode := AddNode(ARootNode, 'Namespaces');

  with AModel do
  for I := 0 to StorageCount - 1 do
    SaveNamespace(NamespacesNode, Storages[I]);

  RelationsNode := AddNode(ARootNode, 'Relations');

  with AModel.DomainModel do
  begin
    for I := 0 to RelationCount - 1 do
      if Relations[I].IsGeneralization then
        SaveGeneralization(RelationsNode, Relations[I])
      else
        if Not Relations[I].IsInherited then
          SaveAssociation(RelationsNode, Relations[I]);
  end;
end;

procedure TovXMLModelStorage.SaveNamespace(AParentNode: IXMLNode;
  AModelStorage: TovCustomModelStorage);
var
  I: Integer;
  NamespaceNode, ClassesNode: IXMLNode;
begin
  NamespaceNode := AddNode(AParentNode, 'Namespace');
  with AModelStorage do
  begin
    if NameSpace <> '' then
      AddNode(NamespaceNode, 'Name', NameSpace)
    else
      AddNode(NamespaceNode, 'Name', ' ');

    if Not FIsSavingMapping then
    begin
      AddNode(NamespaceNode, 'IsDefault', Namespace = '');
      AddNode(NamespaceNode, 'Description', '');
    end;

    with AModelStorage.DomainModel do
    begin
      if DataTypeCount > 0 then
      begin
        FDataTypesNode := AddNode(NamespaceNode, 'DataTypes');
        for I := 0 to DataTypeCount - 1 do SaveDataType(FDataTypesNode, DataTypes[I]);
      end;
      if ObjectCount > 0 then
      begin
        ClassesNode := AddNode(NamespaceNode, 'Classes');
        for I := 0 to ObjectCount - 1 do SaveClass(ClassesNode, Objects[I]);
      end;
    end
  end
end;

procedure TovXMLModelStorage.SaveDataType(AParentNode: IXMLNode; ADataType:
  TModelDataType);
var
  DataTypeNode, MappingNode, ValuesNode, ValueNode: IXMLNode;
  I: Integer;
  TypeName: String;
begin
  DataTypeNode := AddNode(AParentNode, 'DataType');

  AddNode(DataTypeNode, 'Name', ADataType.Name);
  if Not FIsSavingMapping then
  begin
    AddNode(DataTypeNode, 'Description', '');
  end else
  begin
    MappingNode := AddNode(DataTypeNode, SMappingNodeName);
    TypeName := ADataType.BaseType;
    AddNode(MappingNode, 'BaseType', TypeName);
    if ADataType.Size > 0 then AddNode(MappingNode, 'Size', ADataType.Size);
  end;
end;

procedure TovXMLModelStorage.SaveClass(AParentNode: IXMLNode; AModelClass:
  TModelObject);
var
  I: Integer;
  ClassNode, AttributesNode, MappingNode, IndexesNode: IXMLNode;
begin
  ClassNode := AddNode(AParentNode, 'Class');

  AddNode(ClassNode, 'Name', AModelClass.Name);

  if Not FIsSavingMapping then
  begin
    AddNode(ClassNode, 'Description', AModelClass.Description);
    AddNode(ClassNode, 'IsAbstract', AModelClass.IsAbstract);
  end else
  begin
    MappingNode := AddNode(ClassNode, SMappingNodeName);
    AddNode(MappingNode, 'Id', AModelClass.ID);
    AddNode(MappingNode, 'TableName', AModelClass.TableName);
    if AModelClass.IndexCount > 0 then
    begin
      IndexesNode := AddNode(MappingNode, 'Indexes');
      for I := 0 to AModelClass.IndexCount - 1 do SaveIndex(IndexesNode, AModelClass.Indexes[I]);
    end;
  end;

  with AModelClass do
  if AttributeCount > 0 then
  begin
    AttributesNode := AddNode(ClassNode, 'Attributes');
    for I := 0 to AttributeCount - 1 do
      if Not Attributes[I].IsInherited then
        SaveAttribute(AttributesNode, Attributes[I]);
  end;
end;

procedure TovXMLModelStorage.SaveAttribute(AParentNode: IXMLNode;
  AModelAttribute: TModelAttribute);
var
  AttributeNode, MappingNode: IXMLNode;

  function GetDataType(TypeName: String): String;
  var
    I: Integer;
  begin
    Result := TypeName;
    if AModelAttribute.Owner.Model.FindDataType(TypeName) = nil then
    begin
      {  ""         }
      SaveDataType(FDataTypesNode, AModelAttribute.Owner.Model.AddDataType(TypeName, TypeName))
    end;
  end;

begin
  AttributeNode := AddNode(AParentNode, 'Attribute');
  AddNode(AttributeNode, 'Name', AModelAttribute.Name);
  if Not FIsSavingMapping then
  begin
    AddNode(AttributeNode, 'Description', AModelAttribute.Description);
    AddNode(AttributeNode, 'DataType', GetDataType(AModelAttribute.TypeName));
    AddNode(AttributeNode, 'Size', AModelAttribute.Size);
    AddNode(AttributeNode, 'DefaultValue', Null);
    AddNode(AttributeNode, 'Persistent', AModelAttribute.Persistent);
    AddNode(AttributeNode, 'Calculated', AModelAttribute.Calculated);
    AddNode(AttributeNode, 'Validated', False);
    AddNode(AttributeNode, 'Required', AModelAttribute.Required);
    AddNode(AttributeNode, 'RequiredMsg', AModelAttribute.RequiredMsg);
  end else
  begin
    MappingNode := AddNode(AttributeNode, SMappingNodeName);
    AddNode(MappingNode, 'FieldName', AModelAttribute.FieldName);
  end;
end;

procedure TovXMLModelStorage.SaveAssociation(AParentNode: IXMLNode;
  Association: TModelAssociation);
var
  AssociationNode: IXMLNode;
begin
  AssociationNode := AddNode(AParentNode, 'Association');
  SaveAssociationEnd(AddNode(AssociationNode, 'SourceClass'), Association, True);
  SaveAssociationEnd(AddNode(AssociationNode, 'TargetClass'), Association, False);
  if Not FIsSavingMapping then
  begin
    if Association.IsReference then
      AddNode(AssociationNode, 'RelationType', 'rtReference')
    else
      AddNode(AssociationNode, 'RelationType',
        GetEnumName(TypeInfo(TRelationType), Integer(Association.RelationType)));
  end;
end;

procedure TovXMLModelStorage.SaveAssociationEnd(AParentNode: IXMLNode;
  AModelRelation: TModelRelation; IsSourceEnd: Boolean);

  procedure SaveRelationMappingProperties(AMappingNode: IXMLNode;
    ARelation: TModelRelation);
  begin
    if ARelation.IsReference then
    begin
      AddNode(AMappingNode, 'TargetObjectIdField', ARelation.FieldName);
      AddNode(AMappingNode, 'TargetClassIdField', ARelation.FieldName+'_ClassId');
    end else
    if ARelation.IsComposition then
    begin
      AddNode(AMappingNode, 'OwnerObjectIdField', 'OwnerId');
    end else
    if ARelation.IsAggregation then
    begin
      AddNode(AMappingNode, 'OwnerObjectIdField', ARelation.FieldName);
      AddNode(AMappingNode, 'OwnerClassIdField', ARelation.FieldName+'_ClassId');
    end else
    if ARelation.IsAssociation then
    begin
      AddNode(AMappingNode, 'RelationTable', ARelation.TableName);
      AddNode(AMappingNode, 'SourceObjectIdField', 'OwnerId');
      AddNode(AMappingNode, 'SourceClassIdField', 'SClassId');
      AddNode(AMappingNode, 'TargetObjectIdField', 'ObjectId');
      AddNode(AMappingNode, 'TargetClassIdField', 'TClassId');
    end else
      raise Exception.Create('   ' + ARelation.Name);
  end;

var
  MappingNode: IXMLNode;
begin
  if IsSourceEnd then
  begin
    AddNode(AParentNode, 'ClassName', AModelRelation.SourceObjectName);
    AddNode(AParentNode, 'RoleName', AModelRelation.Name);
  end else
  begin
    AddNode(AParentNode, 'ClassName', AModelRelation.TargetObjectName);
    //AddNode(AParentNode, 'RoleName', AModelRelation.SourceObject.Name);
  end;

  if IsSourceEnd then
  begin
    if Not FIsSavingMapping then
    begin
      AddNode(AParentNode, 'Description', AModelRelation.Description);
      AddNode(AParentNode, 'MinCount', CardinalityToString(AModelRelation.TargetMinCount));
      AddNode(AParentNode, 'MaxCount', CardinalityToString(AModelRelation.TargetMaxCount));
      AddNode(AParentNode, 'Required', AModelRelation.Required);
      AddNode(AParentNode, 'RequiredMsg', AModelRelation.RequiredMsg);
    end else
    begin
      MappingNode := AddNode(AParentNode, SMappingNodeName);
      AddNode(MappingNode, 'Id', AModelRelation.Id);
      SaveRelationMappingProperties(MappingNode, AModelRelation);
    end;
  end;
end;

procedure TovXMLModelStorage.SaveGeneralization(AParentNode: IXMLNode;
  Generalization: TModelGeneralization);
var
  GeneralizationNode: IXMLNode;
begin
  GeneralizationNode := AddNode(AParentNode, 'Generalization');
  AddNode(GeneralizationNode, 'SourceClass', Generalization.SourceObjectName);
  AddNode(GeneralizationNode, 'TargetClass', Generalization.TargetObjectName);
end;

procedure TovXMLModelStorage.SaveIndex(AParentNode: IXMLNode; AIndex:
  TModelObjectIndex);
var
  IndexNode: IXMLNode;
  OptionsNode: IXMLNode;
begin
  IndexNode := AddNode(AParentNode, 'Index');

  AddNode(IndexNode, 'Name', AIndex.Name);
  AddNode(IndexNode, 'Attributes', AIndex.Attributes);
  OptionsNode := AddNode(IndexNode, 'Options');
  OptionsNode.Attributes['Primary'] := ixPrimary in AIndex.Options;
  OptionsNode.Attributes['Unique'] := ixUnique in AIndex.Options;
  OptionsNode.Attributes['Descending'] := ixDescending in AIndex.Options;
  OptionsNode.Attributes['CaseInsensitive'] := ixCaseInsensitive in AIndex.Options;
end;

{***************************************************************************** }
{                                                         }
{ **************************************************************************** }

function TovXMLModelStorage.AddNode(AParentNode: IXMLNode; AName: String):
  IXMLNode;
begin
  Result := AParentNode.AddChild(AName);
end;

function TovXMLModelStorage.AddNode(AParentNode: IXMLNode; AName: String;
  AValue: OleVariant): IXMLNode;
begin
  Result := AParentNode.AddChild(AName);
  Result.NodeValue := AValue
end;

end.

